/* access.c */
/* code to do authentication*/
/* Sat Jan 30 20:01:19 EST 1993 wbardwel : created*/
/*
 * HISTORY
 * $Log:	access.c,v $
 * Revision 2.14  94/07/12  17:26:11  wbardwel
 * 	update copyright, fix for svr4
 * 	[94/07/12  17:19:24  wbardwel]
 * 
 * Revision 2.13  94/06/14  18:01:37  wbardwel
 * 	remove redundant inclusion of in.h
 * 	[94/06/14  17:55:25  wbardwel]
 * 
 * Revision 2.12  94/04/19  16:45:49  wbardwel
 * 	kill some multiple inclusions, cause ultrix breaks on them
 * 	[94/04/19  16:45:17  wbardwel]
 * 
 * Revision 2.11  94/03/29  17:10:40  wbardwel
 * 	fix minor bug in localhost lookups
 * 	[94/03/29  17:09:53  wbardwel]
 * 
 * Revision 2.10  94/03/29  17:07:13  wbardwel
 * 	make gethostbyname on localhost only happen if needed...
 * 	[94/03/29  17:04:10  wbardwel]
 * 
 * Revision 2.9  94/02/15  22:54:49  wbardwel
 * 	lookup localhost name too cause some systems don't have FQDNs as hostnames
 * 	[94/02/15  22:54:11  wbardwel]
 * 
 * Revision 2.8  93/08/13  16:33:01  wbardwel
 * 	twiddle what we do with servername
 * 	[93/08/13  16:24:01  wbardwel]
 * 
 * Revision 2.7  93/08/10  19:19:12  wbardwel
 * 	support command line specification of access file
 * 	[93/08/10  19:12:00  wbardwel]
 * 
 * Revision 2.6  93/08/05  00:26:00  wbardwel
 * 	buglette with hosts with no reverse lookup.
 * 	[93/08/05  00:23:03  wbardwel]
 * 
 * Revision 2.5  93/08/04  16:36:18  wbardwel
 * 	make servername global (but not advertised in dnntp.h)
 * 	[93/08/04  16:29:20  wbardwel]
 * 
 * Revision 2.4  93/07/21  22:40:45  wbardwel
 * 	Added kludge to always send XREALLY's so we can enable cmu.andrew
 * 	if someone uses forwarder.
 * 	Fix for unixi that will say a connection came from 127.0.0.1 when you
 * 	go from localhost to localhost.
 * 	[93/07/21  22:31:22  wbardwel]
 * 
 * Revision 2.3  93/06/11  20:29:48  wbardwel
 * 	added code to allow use of reversed(on nibbles) hex ip addresses in the
 * 	access file.
 * 	[93/06/11  20:27:54  wbardwel]
 * 
 * Revision 2.2  93/06/02  14:51:48  wbardwel
 * 	added support code for access file for forwarding from non-local hosts
 * 	[93/05/24            wbardwel]
 * 
 * 	just return minor error if connection not from localhost. (and disable
 * 	XREALLY tries if localhost)
 * 	[93/05/24            wbardwel]
 * 
 * 	Created.
 * 	[93/01/30            wbardwel]
 * 
*/
/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

/*#include <sys/socket.h> done in dnntp.h*/
/*#include <netinet/in.h> done in dnntp.h*/
#include <arpa/inet.h>
/*#include <netdb.h> done in dnntp.h*/
/*#include <ctype.h> done in dnntp.h*/
extern int gethostname();

char servername[MAXHOSTNAMELEN+1];	/*size limit kinda bogus*/
char clientname[MAXHOSTNAMELEN+1];	/*size limit kinda bogus*/
char clientaddrstr[4*3+3+1];	/*3digits and a period per octet*/
static int hostmatch();
static void make_revaddr_str();

/* remote_auth -- make sure person connecting is someone we trust
    parameters
	sockt: file destriptor of client connected socket
    return
	0 or higher if trusted, less than 0 if not */
int remote_auth(sockt)
int sockt;
{
#if !LOCAL_ONLY
struct sockaddr_in clientaddr;
int addrlen = sizeof(struct sockaddr_in);
struct hostent *clientent;
struct hostent *serverent = NULL;
extern int h_errno;
#if DNNTP_ALLOW_OUTSIDE
	/*for reverse of address (access checks)*/
char clientrevaddrstr[8+1]; /*8 hex digits*/
#endif /*DNNTP_ALLOW_OUTSIDE*/
	if(gethostname(servername, sizeof(servername)) < 0) {
#if SYSLOG
		syslog(LOG_ERR, "remote_auth: gethostname %m");
#endif /*SYSLOG*/
		return(DNNTP_REMOTE_GONE_ERR);
	}
	if (!index(servername, '.')) {
		serverent = gethostbyname(servername);
		if (serverent)
			strncpy(servername, serverent->h_name,
				sizeof(servername));
		else
#if LOG
#if SYSLOG
			syslog(LOG_ERR, "remote_auth: unable to lookup \
servername");
#endif /*SYSLOG*/
#endif /*LOG*/
	}
	STRTOLOWER(servername);
	if (getpeername(sockt, (struct sockaddr *)&clientaddr, &addrlen) < 0){
#if SYSLOG
		syslog(LOG_ERR, "remote_auth: getpeername: unable to check \
remote host %m");
#endif /*SYSLOG*/
		return(DNNTP_REMOTE_GONE_ERR);
	}
	if(addrlen <= sizeof(struct in_addr)){
#if SYSLOG
		syslog(LOG_ERR, "remote_auth: (addrlen <= %d): unable to check \
remote host %m", sizeof(struct sockaddr_in));
#endif /*SYSLOG*/
		return(DNNTP_REMOTE_GONE_ERR);
	}
	strncpy(clientaddrstr, inet_ntoa( clientaddr.sin_addr),
		sizeof(clientaddrstr));
	do {
		if (strcmp(clientaddrstr, DNNTP_LOCALHOST_STR) == 0) {
			if (!serverent)
				clientent = gethostbyname(servername);
			else 
				clientent = serverent;
		} else {
			clientent = gethostbyaddr( (char *)&clientaddr.sin_addr,
				sizeof(clientaddr.sin_addr), AF_INET);
		}
	} while ((clientent == NULL) && (h_errno == TRY_AGAIN) && (sleep(5),1));
	if ((clientent == NULL) || (clientent->h_length !=
		sizeof(struct in_addr)) || (clientent->h_addrtype != AF_INET)){
#if DNNTP_ALLOW_OUTSIDE
		strcpy(clientname, clientaddrstr);
		make_revaddr_str(&(clientaddr.sin_addr.s_addr), clientrevaddrstr);
#else /*!DNNTP_ALLOW_OUTSIDE*/
#if SYSLOG
		syslog(LOG_ERR, "remote_auth: unable to lookup remote host \
name '%s' %m", clientaddrstr);

#endif /*SYSLOG*/
		return(DNNTP_REMOTE_GONE_ERR);
#endif /*!DNNTP_ALLOW_OUTSIDE*/
	} else {
#if DNNTP_ALLOW_OUTSIDE
		bcopy( clientent->h_addr_list[0], &(clientaddr.sin_addr.s_addr), sizeof(struct in_addr));
		make_revaddr_str(&(clientaddr.sin_addr.s_addr), clientrevaddrstr);
#endif /*DNNTP_ALLOW_OUTSIDE*/
		strncpy(clientaddrstr, inet_ntoa( clientaddr.sin_addr),
			sizeof(struct sockaddr_in));
		strncpy(clientname, clientent->h_name, sizeof(clientname));
	}
	if (!clientent || strncasecmp(servername, clientent->h_name,
		sizeof(servername))) {
		if (clientent)
			for(; *clientent->h_aliases && strncasecmp(servername, 
				*clientent->h_aliases, sizeof(servername));
				clientent->h_aliases++)
					;
		if (!clientent || !*clientent->h_aliases) {
#if DNNTP_ALLOW_OUTSIDE
		FILE *oafp;
		char buf[256], *tmpch;
		int permited = 0;
				/* check if we like outside connections*/
			if ((oafp = fopen((dnntp_access_file?
				dnntp_access_file:DNNTP_OUTSIDE_ALLOWED_FILE),
				"r")) == NULL) {
				/*no! we don't allow super forwarding...*/
#if LOG
#if SYSLOG
				syslog(LOG_INFO, "remote_auth: unauthorized \
outside connection from: %s(%s)", clientaddrstr, clientname);
#endif /*SYSLOG*/
#endif /*LOG*/
				return (DNNTP_SIMPLE_ERR);	/*no setup for this*/
			}
			/* do search and check on hostname*/
			while (!feof(oafp)){
				clear_spaces(oafp);
				if (fgets(buf, sizeof(buf), oafp) == NULL)
					continue;
				for (tmpch = buf; !isspace(*tmpch) &&
					(*tmpch != '#'); tmpch++);
				*tmpch = '\0';
				if (*buf == '!') {
					tmpch = buf;
					tmpch++;
					permited = (permited &&
						!hostmatch(clientname,
							clientrevaddrstr,
							tmpch));
				} else if (*buf) {
					permited = (permited ||
						hostmatch(clientname,
							clientrevaddrstr,
							buf));
				}
			}
			fclose(oafp);
#if LOG
#if SYSLOG
			if (!permited)
				syslog(LOG_INFO, "remote_auth: unauthorized \
outside connection from: %s(%s)", clientaddrstr, clientname);
			else
				syslog(LOG_INFO, "remote_auth: authorized \
outside connection from: %s(%s)", clientaddrstr, clientname);
#endif /*SYSLOG*/
#endif /*LOG*/
			return (permited?DNNTP_SUCCESS:DNNTP_SIMPLE_ERR);
#else /*!DNNTP_ALLOW_OUTSIDE*/
#if LOG
#if SYSLOG
			syslog(LOG_INFO, "remote_auth: unauthorized \
connection from: %s(%s)", clientaddrstr, clientname);
#endif /*SYSLOG*/
#endif /*LOG*/
			return (DNNTP_SIMPLE_ERR);
#endif /*!DNNTP_ALLOW_OUTSIDE*/
		}
	}
		/*if we got this far is succeded*/
#if LOG
#if SYSLOG
			syslog(LOG_INFO, "local connection from: %s(%s)",
				clientaddrstr, clientname);
#endif /*SYSLOG*/
#endif /*LOG*/
#endif /*!LOCAL_ONLY*/
		/* if this is local, go and suppress all xreally commands*/
#if DO_XREALLY && !DNNTP_CMU_ANDREW_KLUDGE
	ServersSuppressXREALLY();
#endif /*DO_XREALLY&& !DNNTP_CMU_ANDREW_KLUDGE*/
	return(DNNTP_SUCCESS);
}

#if DNNTP_ALLOW_OUTSIDE
/* hostmatch -- compares a hostname or * followed by and end of a hostname,
	with a hostname. (eg. patern to match against, orion.fac.cs.cmu.edu or
	*.fac.cs.cmu.edu or even *n.fac.cs.cmu.edu) also will do
	netaddress matching with the netaddress nibble-swaped in hex (eg.
	11ed2008 (thats 128.2.222.17) or *2008 (thats like 128.2.*))
    parameters
	host	hostname to check if matches
	revaddrstr	reverse of address in hex in a string
	match	pattern to match against
    return
	0 if doesn't match, 1 if does.
*/
static int hostmatch(host, revaddrstr, match)
char *host, *revaddrstr, *match;
{
	if (*match == '*') {
		long matchlen = strlen(++match);
		long hostlen = strlen(host);
		long raddrlen = strlen(revaddrstr);
				/* make sure hostname is atleast long enough*/
			if( hostlen >= matchlen ) {
				if (strcasecmp( host+(hostlen-matchlen), match)
				== 0)
					return (1);
				else if (raddrlen >= matchlen)
					return (strcasecmp( revaddrstr
						+(raddrlen-matchlen), match)
						== 0);
				else
					return (0);
			} else
				return (0);
	} else
		return ((strcasecmp(host, match) == 0) ||
			(strcasecmp(revaddrstr, match) == 0));
}

/* make_hostname_up -- composes a hostname based on the address...(it is
	the address in hex (rendered to a string) reversed (on the nibble
	(4bits) level (ie the characters))
    parameters
	addr	address to make up name with
	str	buffer to render too (must be 8+1 long (8 hexdigits and a NULL))
    return
	none
*/
static void make_revaddr_str(addr, str)
struct in_addr *addr;
char *str;
{
int i;
char ch;
	sprintf(str, "%08x", ntohl(*((u_long*)addr)));
		/*reverse the string...*/
	for(i = 0; i < 4; i++) {
		ch = str[i];
		str[i] = str[7-i];
		str[7-i] = ch;
	}
}
#endif /*DNNTP_ALLOW_OUTSIDE*/
