/* command.c */
/* code to actually execute (well distribute) commands*/
/* Mon Feb  8 16:12:59 EST 1993 wbardwel : created*/
/*
 * HISTORY
 * $Log:	command.c,v $
 * Revision 2.23  95/01/05  13:57:01  wbardwel
 * 	make NEWGROUPS responses include full active line, no matter what
 * 	it takes...
 * 	[95/01/05  13:24:41  wbardwel]
 * 
 * Revision 2.22  94/08/24  12:47:15  wbardwel
 * 	don't give up searchs by message-id if a server is down
 * 	[94/08/24  12:44:38  wbardwel]
 * 
 * Revision 2.21  94/07/30  01:47:03  wbardwel
 * 	add sys/time.h (for select) cause Linux wants it...
 * 	[94/07/30  01:41:08  wbardwel]
 * 
 * Revision 2.20  94/07/12  17:26:17  wbardwel
 * 	update copyright
 * 	[94/07/12  17:20:34  wbardwel]
 * 
 * Revision 2.19  94/05/18  22:32:02  wbardwel
 * 	don't make it so picky about what NEWGROUPS returns...
 * 	[94/05/18  22:31:32  wbardwel]
 * 
 * Revision 2.18  94/05/17  04:36:38  wbardwel
 * 	update to match new do_post
 * 	[94/05/17  04:32:49  wbardwel]
 * 
 * Revision 2.17  94/03/02  16:09:51  wbardwel
 * 	fix error message
 * 	[94/03/02  16:05:39  wbardwel]
 * 
 * Revision 2.16  94/01/06  04:38:53  wbardwel
 * 	Allow turnning off XREF filtering
 * 	[94/01/06  04:28:46  wbardwel]
 * 
 * Revision 2.15  93/12/21  01:12:54  wbardwel
 * 	STAT <message-id> forgot to reply on success
 * 	(reported by clamen@cs.cmu.edu)
 * 	[93/12/21  01:12:29  wbardwel]
 * 
 * Revision 2.14  93/11/02  22:34:12  wbardwel
 * 	include fixes
 * 	[93/11/02  22:01:08  wbardwel]
 * 
 * Revision 2.13  93/10/27  16:02:26  wbardwel
 * 	moved list_cmd() to list_cmd.c
 * 	[93/10/27  15:19:13  wbardwel]
 * 
 * Revision 2.12  93/08/18  19:41:30  wbardwel
 * 	bug fixes for when a server is down, bug fix for HEAD,ARTICLE with msg-ids
 * 	[93/08/18  19:31:09  wbardwel]
 * 
 * Revision 2.11  93/08/13  21:18:35  wbardwel
 * 	killed bogus variable
 * 	[93/08/13  21:16:18  wbardwel]
 * 
 * Revision 2.10  93/08/13  16:33:12  wbardwel
 * 	bug fixes, mostly code to make calls to filter routines to patch xrefs
 * 	[93/08/13  16:25:29  wbardwel]
 * 
 * Revision 2.9  93/08/10  19:19:16  wbardwel
 * 	reorder how it talks to servers for message-id based commands
 * 	(hack alert) to do last server in list first...
 * 	fix typo in list usage message
 * 	[93/08/10  19:12:55  wbardwel]
 * 
 * Revision 2.8  93/08/05  01:23:17  wbardwel
 * 	fixed list_cmd to match new Groups2ServerEndAdd
 * 	[93/08/05  01:21:31  wbardwel]
 * 
 * Revision 2.7  93/08/04  16:36:39  wbardwel
 * 	bug fixlette, convert use of AddGroup2Server to new style
 * 	[93/08/04  00:08:39  wbardwel]
 * 
 * Revision 2.6  93/07/28  13:25:10  wbardwel
 * 	"Parallelized" list_cmd, newnews_cmd, newgroups_cmd, mode_cmd
 * 	(so they try to keep all the servers busy while doing a command).
 * 	bug fixes...
 * 	[93/07/28  12:59:13  wbardwel]
 * 
 * Revision 2.5  93/07/21  22:40:52  wbardwel
 * 	bug fixes (mostly for when a server is down)
 * 	[93/07/21  22:31:52  wbardwel]
 * 
 * Revision 2.4  93/07/16  16:21:05  wbardwel
 * 	bug fixes and added DATE, MODE, and LISTGROUP commands.
 * 	[93/07/16  16:10:30  wbardwel]
 * 
 * Revision 2.3  93/06/23  00:50:16  wbardwel
 * 	added code for disk list cache so list, newgroups and group
 * 	commands work even with servers down.
 * 	[93/06/23  00:24:04  wbardwel]
 * 
 * Revision 2.2  93/06/02  14:51:54  wbardwel
 * 	Added code to use LIST cache (with Group2Server stuff) to avoid
 * 	"extrainious" LISTs.
 * 	[93/05/27            wbardwel]
 * 
 * 	changed to single server sending of posts (to match do_post)
 * 	[93/05/21            wbardwel]
 * 
 * 	removed command_no code
 * 	[93/05/12            wbardwel]
 * 
 * 	Created.
 * 	[93/02/08            wbardwel]
 * 
 *
*/
/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

#if DNNTP_USE_GROUP_INFO
#ifndef __STDC__
#include <varargs.h>	/*for concat*/
#else /*__STDC__*/
#include <stdarg.h>
#endif /*__STDC__*/
#endif /*DNNTP_USE_GROUP_INFO*/
#include <sys/time.h> /*for select*/

int currentserver = -1;
void print_couldnt_send();

/* art_cmd -- do the HEAD, BODY, ARTICLE, STAT commands
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int art_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
int status;
		/*is it a numbered request?*/
	if ((no_args == 1) || ((no_args == 2) && isdigit(args[1][0]))) {
		Group2ServerPrep();
		/*send it to currentserver, and return result*/
			/*but first, make sure a current server with a group*/
		if ((currentserver >=0) &&
			(serv_array[currentserver].curgroup[0])) {
				/*remeber this only opens if needed*/
			OpenServer(currentserver);
			if (SendGroup(currentserver, NULL) < 0){
					print_couldnt_send(currentserver);
					return(DNNTP_REMOTE_GONE_ERR);
			}
			if (no_args == 1) {
				if (SendToServer(currentserver, "%s",
					args[0]) < 0) {
					print_couldnt_send(currentserver);
					return(DNNTP_REMOTE_GONE_ERR);
				}
			} else {	/* no_args == 2*/
				if (SendToServer(currentserver, "%s %s",
					args[0], args[1]) < 0) {
					print_couldnt_send(currentserver);
					return(DNNTP_REMOTE_GONE_ERR);
				}
			}
			status = ReadFromServer(currentserver, str);
			if (status >= 0)
				fputs(str, stdout);
			else {
				print_couldnt_send(currentserver);
				return (DNNTP_REMOTE_GONE_ERR);
			}
				/*if command also returns other junk, do that too*/
#if DNNTP_POSSIBLE_XREF_ODDNESS
 			if (status == OK_BODY)
#else /*!DNNTP_POSSIBLE_XREF_ODDNESS*/
 			if ((status == OK_BODY) || (status == OK_HEAD) ||
				(status == OK_ARTICLE))
#endif /*!DNNTP_POSSIBLE_XREF_ODDNESS*/
			{
				while (ContReadFromServer(currentserver, str) > 0)
					fputs(str, stdout);
					/*ContReadFromServer returned 0 on
					the . so it wasn't sent...*/
				fputs(".\r\n", stdout);
			}
#if DNNTP_POSSIBLE_XREF_ODDNESS
			else if ((status == OK_HEAD) ||
				(status == OK_ARTICLE)) {
				send_art_filtered(currentserver);
				fputs(".\r\n", stdout);
			}
#endif /*DNNTP_POSSIBLE_XREF_ODDNESS*/
			(void) fflush(stdout);
		} else {
			printf("%d no newsgroup selected\r\n", ERR_NCING);
			(void) fflush(stdout);
			return(DNNTP_NO_SETUP_ERR);
		}
	} else if ((no_args == 2) && (args[1][0] == '<')) {
		/*or a message-id request*/
	int i;
		if (currentserver < 0) {
			currentserver = i = serv_array_len-1;
		} else
			i = currentserver;
		do {
			OpenServer(i);
			if (serv_array[i].curgroup[0])
				SendGroup(i, NULL);
			if (SendToServer(i, "%s %s", args[0], args[1]) < 0) {
				CloseServer(i);
				goto next_itr;
			}
			status = ReadFromServer(i, str);
#if DNNTP_POSSIBLE_XREF_ODDNESS
			if (status == OK_BODY)
#else /*!DNNTP_POSSIBLE_XREF_ODDNESS*/
			if ((status == OK_BODY) || (status == OK_HEAD) ||
				(status == OK_ARTICLE))
#endif /*!DNNTP_POSSIBLE_XREF_ODDNESS*/
			{
				fputs(str, stdout);
				while (ContReadFromServer(i, str) > 0)
					fputs(str, stdout);
					/*ContReadFromServer returned 0 on
					the . so it wasn't sent...*/
				fputs(".\r\n", stdout);
				(void) fflush(stdout);
				return(DNNTP_SUCCESS);
			}
#if DNNTP_POSSIBLE_XREF_ODDNESS
			else if ((status == OK_HEAD) ||
				(status == OK_ARTICLE)) {
				fputs(str, stdout);
				send_art_filtered(i);
				fputs(".\r\n", stdout);
				(void) fflush(stdout);
				return(DNNTP_SUCCESS);
			}
#endif /*DNNTP_POSSIBLE_XREF_ODDNESS*/
			else if (status == OK_NOTEXT) {
				fputs(str, stdout);
				(void) fflush(stdout);
				return (DNNTP_SUCCESS);
			}
next_itr:
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
			/*we should have returned before here*/
		printf("%d no article by message-id %s found\r\n", ERR_NOART,
			args[1]);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	} else {
		/*or garbage, send error message*/
		printf("%d Usage: %s [<message-id>|article number].\r\n",
			ERR_CMDSYN, args[0]);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
		/*shouldn't reach here*/
	return(DNNTP_SUCCESS);
}

/* group_cmd -- do the GROUP command
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int group_cmd( args, no_args)
char *args[];
int no_args;
{
int server;
int status;
char str[NNTP_STRLEN+1];
	if (no_args == 2) {
		/*figure out which connection this should be on*/
		Group2ServerPrep();
		server = Group2Server(args[1]);
		if (server < 0) {
			printf("%d Invalid group (failed group to server lookup)\r\n", ERR_NOGROUP);
			(void) fflush(stdout);
			return(DNNTP_SIMPLE_ERR);
		}
		currentserver = server;
			/* ERR_ACCESS not really the correct error number */
		if ((status = OpenServer(server)) < 0) {
			printf("%d group not avalible since %s is down\r\n",
				ERR_ACCESS, ServerName(server));
			(void) fflush(stdout);
			return(status);
		}
		str[0] = '\0';
		if ((status = SetGroup(server, args[1], str, 1)) < 0) {
			if (status >= DNNTP_LOWEST_ERR) {
				if (serv_array[server].connection >= 0)
					printf("%d failure sending group \
command %s\r\n" ,ERR_FAULT, ServerName(server));
				else 
					printf("%d group not avalible since %s \
is down\r\n", ERR_ACCESS, ServerName(server));
				(void) fflush(stdout);
				return(status);
			}
		}
		fputs(str, stdout);
		(void) fflush(stdout);
	} else {
		printf("%d Usage: GROUP <groupname>.\r\n", ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	return(DNNTP_SUCCESS);
}

/* nextlast_cmd -- do the NEXT and LAST commands
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
	command_no: decoded command (type) (not used here)
    return
	>=0 on success <0 on failure
*/
int nextlast_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
	if (no_args == 1) {
		if ((currentserver >=0) &&
			(serv_array[currentserver].curgroup[0])) {
			OpenServer(currentserver);
			if (SendGroup(currentserver, NULL) < 0){
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			if (SendToServer(currentserver, "%s", args[0]) < 0) {
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			(void) ReadFromServer(currentserver, str);
			fputs(str, stdout);
			(void) fflush(stdout);
		} else {
			printf("%d no newsgroup selected\r\n", ERR_NCING);
			(void) fflush(stdout);
			return(DNNTP_NO_SETUP_ERR);
		}
	} else {
		printf("%d Usage: %s.\r\n", ERR_CMDSYN, args[0]);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	return(DNNTP_SUCCESS);
}

/* slave_cmd -- do the SLAVE command (currently ignored)
	really ought to save this status and send it to all open servers and
	when ever it opens servers.
    parameters
	args: argv style broken up command (not used here)
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int slave_cmd( args, no_args)
char *args[];
int no_args;
{
		/*should we really do syntax checks on something we ignore?*/
	if (no_args == 1) {
		printf("%d whatever thrills you (I am ignoring this...)\r\n",
			OK_SLAVE);
		(void) fflush(stdout);
	} else {
		printf("%d Usage: %s.\r\n", ERR_CMDSYN, args[0]);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	return(DNNTP_SUCCESS);
}

#if XHDR
/* xhdr_cmd -- do the XHDR command
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int xhdr_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
int status;
		/*is is a ranged (or cur article) request*/
		/* more error checking would be good...*/
	if ((no_args == 2) || ((no_args == 3) && isdigit(args[2][0]))) {
		/*send it to currentserver, and return result*/
			/*but first, make sure a current server with a group*/
		if ((currentserver >= 0) &&
			(serv_array[currentserver].curgroup[0])) {
				/*remeber this only opens if needed*/
			OpenServer(currentserver);
			if (SendGroup(currentserver, NULL) < 0){
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			if (no_args == 2) {
				if (SendToServer(currentserver, "%s %s",
					args[0], args[1], args[2]) < 0){
					print_couldnt_send(currentserver);
					return(DNNTP_REMOTE_GONE_ERR);
				}
			} else {	/* no_args == 3*/
				if (SendToServer(currentserver, "%s %s %s",
					args[0], args[1], args[2]) < 0){
					print_couldnt_send(currentserver);
					return(DNNTP_REMOTE_GONE_ERR);
				}
			}
			status = ReadFromServer(currentserver, str);
			fputs(str, stdout);
			if (status == OK_HEAD) {
#if DNNTP_POSSIBLE_XREF_ODDNESS
				if (strcasecmp(args[1], "xref")) {
#endif /*DNNTP_POSSIBLE_XREF_ODDNESS*/
					while (ContReadFromServer(currentserver,
						str) > 0)
							fputs(str, stdout);
					/*ContReadFromServer returned 0 on
					the . so it wasn't sent...*/
#if DNNTP_POSSIBLE_XREF_ODDNESS
				} else {
					xhdr_xref_filtered(currentserver);
				}
#endif /*DNNTP_POSSIBLE_XREF_ODDNESS*/
				fputs(".\r\n", stdout);
			} else if (status < 0) {
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			(void) fflush(stdout);
		} else {
			printf("%d no newsgroup selected\r\n", ERR_NCING);
			(void) fflush(stdout);
			return(DNNTP_NO_SETUP_ERR);
		}
	} else if ((no_args == 3) && (args[2][0] == '<')) {
		/*or a message-id request*/
	int i;
		if (currentserver < 0) {
			currentserver = i = serv_array_len-1;
		} else
			i = currentserver;
		if (!strcasecmp(args[1], "xref"))
			Group2ServerPrep();
		do {
			OpenServer(i);
			if (serv_array[i].curgroup[0])
				SendGroup(i, NULL);
			if (SendToServer(i, "%s %s %s", args[0], args[1],
				args[2]) < 0) {
				print_couldnt_send(i);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			status = ReadFromServer(i, str);
			if (status == OK_HEAD) {
				fputs(str, stdout);
#if DNNTP_POSSIBLE_XREF_ODDNESS
				if (strcasecmp(args[1], "xref")) {
#endif /*DNNTP_POSSIBLE_XREF_ODDNESS*/
					while (ContReadFromServer(i, str) > 0)
						fputs(str, stdout);
					/*ContReadFromServer returned 0 on
					the . so it wasn't sent...*/
#if DNNTP_POSSIBLE_XREF_ODDNESS
				} else {
					xhdr_xref_filtered(i);
				}
#endif /*DNNTP_POSSIBLE_XREF_ODDNESS*/
				fputs(".\r\n", stdout);
				(void) fflush(stdout);
				return(DNNTP_SUCCESS);
			}
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
			/*we should have returned before here*/
		printf("%d no article by message-id %s found\r\n", ERR_NOART,
			args[2]);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	} else {
		/*or garbage, send error message*/
#if XHDR_EXT
		printf("%d Usage: XHDR headerfield[:headerfield\
[:headerfield...]] [article range|<message-id>].\r\n", ERR_CMDSYN);
#else /*!XHDR_EXT*/
		printf("%d Usage: XHDR headerfield [article range|<message-id>].\r\n",
			ERR_CMDSYN);
#endif /*!XHDR_EXT*/
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
		/*shouldn't reach here*/
	return(DNNTP_SUCCESS);
}
#endif /*XHDR*/

/* do Help command (ie print out supported commands)
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int help_cmd( args, no_args)
char *args[];
int no_args;
{
int i, j;
		/*note: not syntax check out of kindness...*/
	printf("%d This server accepts the following commands:\r\n", INF_HELP);
	for (i=0; i < NUM_COMMANDS; i+=3) {
		for (j=i; ((j < NUM_COMMANDS) && (j < (i+3))); j++)
			printf("%s\t\t", cmdtable[j].cmd_name);
		fputs("\r\n", stdout);
	}
	printf("\r\nBugs to %s\r\n.\r\n", DNNTP_TROUBLE_EMAIL);
	fflush(stdout);
	return (DNNTP_SUCCESS);
}


/* newgroups -- do NEWGROUPS command (ask every server about new groups)
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
/* note: this ought to figure out what servers it really needs to talk to...
instead of going to everyone...(trouble is a given group could be anywhere)*/
/*further note: this will not recognize newgroups since last list (which
may have been done internally when needed)*/
int newgroups_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
int i, status, first_time = 1;
char **read_args;
int no_read_args;
int ofunc;
	if ((no_args < 3)||(no_args > 5)) {
			/*send error message*/
		printf("%d Usage: NEWGROUPS yymmdd hhmmss [\"GMT\"] \
[<distributions>].\r\n", ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	if (currentserver < 0) {
		currentserver = i = 0;
	} else
		i = currentserver;
	Group2ServerPrep();
	do {
		serv_array[i].misc_state = serv_array[i].connection;
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
	for (ofunc = 0; ofunc < DNNTP_OPEN_STAGES; ofunc++) {
	do {
		if (serv_array[i].misc_state < 0)
			open_serv_func[ofunc](i);
		if (ofunc == (DNNTP_OPEN_STAGES-1)) {
			if (serv_array[i].curgroup[0])
				SendGroup(i, NULL);
			if (serv_array[i].connection >= 0) {
				/*send command*/
			if (no_args == 3)
				status = SendToServer(i, "NEWGROUPS %s %s",
					args[1], args[2]);
			else if (no_args == 4)
				status = SendToServer(i, "NEWGROUPS %s %s %s",
					args[1], args[2], args[3]);
			else /* (no_args == 5)*/
				status = SendToServer(i, "NEWGROUPS %s %s %s %s"
					, args[1], args[2], args[3], args[4]);
			}
		}
			/*ought to do something here....*/
		/*if ( status < 0) {
		}*/
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
	}
	/*i = currentserver;*/
	do {
	char contstr[NNTP_STRLEN+1];
	int item;
		if (serv_array[i].connection < 0) {
			i = (i+1)%serv_array_len;
			continue;
		}
		status = ReadFromServer(i, str);
		if (status == OK_NEWGROUPS) {
			while (ContReadFromServer(i, contstr) > 0) {
				/*ContReadFromServer returned 0 on
				the . so it wasn't sent...*/
			    read_args = break_args(contstr, &no_read_args);
			    if (first_time) {
/*printf("%d list of new groups since: %d%d follows.\r\n, OK_NEWGROUPS);*/
				fputs(str, stdout);
				first_time = 0;
			    }
			    if ((item = Group2ServerItem(read_args[0])) >= 0) {
				if (group_array[item].server_num == i) {
					if (no_read_args == 4)
					printf("%s %s %s %s\r\n", read_args[0],
						read_args[1], read_args[2],
						read_args[3]);
					else
					printf("%s %s\r\n", read_args[0],
						group_array[item].stuff);
				}
			    } else {
					if (no_read_args == 4)
					printf("%s %s %s %s\r\n", read_args[0],
						read_args[1], read_args[2],
						read_args[3]);
					else
					printf("%s 0 0 y\r\n", read_args[0],
						group_array[item].stuff);
			    }
			    free_arg(read_args);
			}
		}
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
		/* we got no non-error returns so say what servers thought*/
	if (first_time) {
			/*pretend it worked if no servers*/
		if (status < 0) {
			printf("%d list of new groups since: %s%s follows.\r\n",
				OK_NEWGROUPS, args[2], args[3]);
		} else {
			fputs(str, stdout);
			if (status != OK_NEWGROUPS) {
				(void) fflush(stdout);
				return(DNNTP_SUCCESS);
			}
		}
	}
	fputs(".\r\n", stdout);
	(void) fflush(stdout);
	return(DNNTP_SUCCESS);
}


/* newnews -- do NEWNEWS command (ask every server about new news)
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
/* note: this ought to figure out what servers it really needs to talk to...
instead of going to everyone...(trouble is a given group could be anywhere)*/
/* further note: odd similarity with newgroups routine no? */
/* more notes: this is were some sort of caching of message-ids would start*/
int newnews_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
int i, status, first_time = 1;
	if ((no_args < 4)||(no_args > 6)) {
			/*send error message*/
		printf("%d Usage: NEWNEWS newsgroups yymmdd hhmmss [\"GMT\"] \
[<distributions>].\r\n", ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	if (currentserver < 0) {
		currentserver = i = 0;
	} else
		i = currentserver;
	do {
		serv_array[i].misc_state = serv_array[i].connection;
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
	{
	int ofunc;
		for (ofunc = 0; ofunc < DNNTP_OPEN_STAGES; ofunc++)
		do {
			if (serv_array[i].misc_state < 0)
				open_serv_func[ofunc](i);
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
	}
	do {
		if (serv_array[i].curgroup[0])
			SendGroup(i, NULL);
			/*send command*/
#if DNNTP_USE_GROUP_INFO
		/*if overlaping servers, compose a groupfield query that
		takes care of that*/
		if (groupfile_array != NULL) {
			/*13 = strlen(NEWSNEWS+3spaces+\r+\n*/
		char groupbuf[NNTP_STRLEN+1-13];
		int j;
			if (!strcmp(args[1], "*")) {
				groupbuf[0] = '\0';
			for (j = groupfile_array_len-1; j >= 0; j--) {
			    if (groupfile_array[j].server_num == i) {
				if (groupbuf[0]) {
				    concat(groupbuf+strlen(groupbuf),
					sizeof(groupbuf)-strlen(groupbuf), ",",
 					groupfile_array[j].group, ".*", NULL);
				} else {
				    concat(groupbuf, sizeof(groupbuf),
 					groupfile_array[j].group, ".*", NULL);
				}
			    } else if (index(groupfile_array[j].group, '.')) {
				if (groupbuf[0]) {
				    concat(groupbuf+strlen(groupbuf),
					sizeof(groupbuf)-strlen(groupbuf), ",",
					"!", groupfile_array[j].group, ".*",
					NULL);
				} else { 
				    concat(groupbuf, sizeof(groupbuf), "!",
					groupfile_array[j].group, ".*", NULL);
				}
			    }
			}
			} else {
			    strncpy(groupbuf, args[1], sizeof(groupbuf));
			    groupbuf[sizeof(groupbuf)-1] = '\0';
			    for (j = 0; j < groupfile_array_len; j++) {
				if (groupfile_array[j].server_num != i) {
				    concat(groupbuf+strlen(groupbuf),
				    sizeof(groupbuf)-strlen(groupbuf), ",",
				    "!", groupfile_array[j].group, ".*",
				    NULL);
				}
			    }
			}
			if (groupbuf[0] == '\0')
				strcpy(groupbuf, "*");
			if (serv_array[i].connection >= 0){
				if (no_args == 4)
				status = SendToServer(i, "NEWNEWS %s %s %s",
					groupbuf, args[2], args[3]);
				else if (no_args == 5)
				status = SendToServer(i, "NEWNEWS %s %s %s %s",
					groupbuf, args[2], args[3], args[4]);
				else /* (no_args == 6)*/
				status = SendToServer(i, "NEWNEWS %s %s %s %s %s",
					groupbuf, args[2], args[3], args[4],
					args[5]);
			}
		} else {
#endif /*DNNTP_USE_GROUP_INFO*/
		if (serv_array[i].connection >= 0){
			if (no_args == 4) {
				status = SendToServer(i, "NEWNEWS %s %s %s",
					args[1], args[2], args[3]);
			} else if (no_args == 5) {
				status = SendToServer(i, "NEWNEWS %s %s %s %s",
					args[1], args[2], args[3], args[4]);
			} else {/* (no_args == 6)*/
				status = SendToServer(i, "NEWNEWS %s %s %s %s %s",
					args[1], args[2], args[3], args[4],
					args[5]);
			}
		}
#if DNNTP_USE_GROUP_INFO
		}
#endif /*DNNTP_USE_GROUP_INFO*/
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
	do {
		if ( serv_array[i].connection >= 0) {
			status = ReadFromServer(i, str);
			if (status == OK_NEWNEWS) {
				if (first_time) {
					fputs(str, stdout);
					first_time = 0;
				}
					/*note: we will return duplicates here*/
					/* note: we won't quit on errors...*/
				while (ContReadFromServer(i, str) > 0)
					/*ContReadFromServer returned 0 on
					the . so it wasn't sent...*/
					fputs(str, stdout);
			}
			
		} else if(first_time) {
			print_couldnt_send(i);
			{
			int ind;
				for(ind = i-1; ind >= 0; ind--)
					ClearServer(ind, OK_GROUPS);
			}
			return(DNNTP_REMOTE_GONE_ERR);
		}
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
	fputs(".\r\n", stdout);
	(void) fflush(stdout);
	return(DNNTP_SUCCESS);
}

#if XTHREAD
/* xthread_cmd -- do the XTHREAD command
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int xthread_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
int status;
	if ((no_args == 1) || (no_args == 2)) {
		if ((!strcasecmp(args[1], "DBINIT")) ||((currentserver >=0) &&
			(serv_array[currentserver].curgroup[0]))) {
			OpenServer(currentserver);
			if (strcasecmp(args[1], "DBINIT")) {
				if (SendGroup(currentserver, NULL) < 0){
					print_couldnt_send(currentserver);
					return(DNNTP_REMOTE_GONE_ERR);
				}
			} else {
				if (currentserver < 0)
					currentserver = 0;
				OpenServer(currentserver);
				if (serv_array[currentserver].curgroup[0])
					if (SendGroup(currentserver, NULL) < 0){
						print_couldnt_send(currentserver);
						return(DNNTP_REMOTE_GONE_ERR);
					}
			}
			if (no_args == 1)
				status = SendToServer(currentserver, "XTHREAD");
			else 
				status = SendToServer(currentserver, "XTHREAD \
%s", args[1]);
			if (status < 0) {
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			status = ReadFromServer(currentserver, str);
			fputs(str, stdout);
			if (status == OK_BIN) {
			char **tmp_arg;
			int tmp_no, i;
			char *tmp_buf;
				tmp_arg = break_args(str, &tmp_no);
					/* if this didn't work...*/
				if (tmp_no < 2) {
					printf("%d trouble getting real result \
\r\n", ERR_FAULT);
#if SYSLOG
					syslog(LOG_ERR, "xthread_cmd: wrong \
number of args in return string");
#endif /*SYSLOG*/
					return (DNNTP_REMOTE_DIS_ERR);
				}
					/* sizeof(\r\n.\r\n)=XTHREAD_PADDING*/
				tmp_no = atoi(tmp_arg[1])+XTHREAD_PADDING;
				free_arg(tmp_arg);
				tmp_buf = malloc(tmp_no);
				if (tmp_buf == NULL) {
					printf("%d server out of memory\r\n",
						ERR_GOODBYE);
#if SYSLOG
					syslog(LOG_CRIT, "xthread_cmd: ran out \
of memory");
#endif /*SYSLOG*/
					QuitAllServers();
					exit (1);
				}
			/*note: we are sending binary rubish this way, but it
			should work even so (ie terminating line still same*/
				if (ContBinReadFromServer(currentserver, tmp_no,
					tmp_buf) < 0) {
					printf("%d trouble getting real result \
d\r\n", ERR_FAULT);
					return (DNNTP_REMOTE_DIS_ERR);
				}
				while (tmp_no && ((i = fwrite( tmp_buf, 1,
					tmp_no, stdout)) < tmp_no)) {
					if (i <= 0)
						/*hum, client vanished...*/
						return (DNNTP_SIMPLE_ERR);
					tmp_no-=i;
					tmp_buf+=i;
				}
			} else if (status < 0) {
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
		} else {
			printf("%d no newsgroup selected\r\n", ERR_NCING);
			(void) fflush(stdout);
			return(DNNTP_NO_SETUP_ERR);
		}
	} else {
		printf("%d Usage: XTHREAD [DBINIT|THREAD].\r\n", ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	(void) fflush(stdout);
	return(DNNTP_SUCCESS);
}
#endif /*XTHREAD*/


#if AUTH
/* auth_cmd -- do the AUTHCAP, AUTHINFO, AUTHSYS commands
	current just refuse them...
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int auth_cmd( args, no_args)
char *args[];
int no_args;
{
	printf("%d sorry, no auth commands supported right now\r\n",
		ERR_COMMAND);
	(void)fflush(stdout);
	return (DNNTP_SUCCESS);
}
#endif /*AUTH*/


/* post_cmd -- do the POST command
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int post_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
int no_lines;
char **art;
int j, status, code = DNNTP_SUCCESS, serv_no;
	printf("%d send article to be posted. End with <CR-LF>.<CR-LF>\r\n",
		CONT_POST);
	(void) fflush(stdout);
	if (((serv_no = do_post(&no_lines, &art)) < 0) && no_lines) {
		printf("%d posting failed (no valid groups)\r\n",
			ERR_POSTFAIL);
		(void) fflush(stdout);
		return (DNNTP_SIMPLE_ERR);
	}
	if (serv_no < 0)
		return (DNNTP_SIMPLE_ERR);
	OpenServer(serv_no);
	if (serv_array[serv_no].curgroup[0])
		if (SendGroup(serv_no, NULL) < 0){
			print_couldnt_send(serv_no);
			code = DNNTP_REMOTE_GONE_ERR;
			goto quit;
		}
	status = SendToServer(serv_no, "POST");
	if (status < 0) {
		print_couldnt_send(serv_no);
		code = DNNTP_REMOTE_GONE_ERR;
		goto quit;
	}
	status = ReadFromServer(serv_no, str);
	if (status != CONT_POST) {
		printf("%d real server not accepting posts\r\n",
			ERR_POSTFAIL);
		code = DNNTP_REMOTE_DIS_ERR;
		goto quit;
	}
		/*send it*/
	for (j = 0; j < no_lines; j++) {
		SimpleSendToServer(serv_no, art[j]);
	}
	SimpleSendFlush(serv_no);
	status = ReadFromServer(serv_no, str);
	if (status < 0) {
		print_couldnt_send(serv_no);
		return(DNNTP_REMOTE_GONE_ERR);
	}
	fputs(str, stdout);
	if (status != OK_POSTED) {
		code = DNNTP_REMOTE_DIS_ERR;
		goto quit;
	}
quit:
	(void) fflush(stdout);
	do_post_free_ret(no_lines, art);
	return (code);
}

/* ihave_cmd -- do the IHAVE command
	current just refuse posts...
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int ihave_cmd( args, no_args)
char *args[];
int no_args;
{ 
/*no, I don't got it, I just don't want it....*/
	printf("%d sorry, I don't know how to do ihave's yet\r\n", ERR_GOTIT);
	(void) fflush(stdout);
	return (DNNTP_SUCCESS);
}

#ifdef XOVER
/* xover_cmd -- do the XOVER command
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int xover_cmd( args, no_args)
char *args[];
int no_args;
{ 
int status;
char str[NNTP_STRLEN+1];
	if ((no_args == 1) || ((no_args == 2) && isdigit(args[1][0]))) {
			/*but first, make sure a current server with a group*/
		/*Group2ServerPrep();*//*must be in a group, so not needed*/
		if ((currentserver >= 0) &&
			(serv_array[currentserver].curgroup[0])) {
				/*remeber this only opens if needed*/
			OpenServer(currentserver);
			if (SendGroup(currentserver, NULL) < 0){
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			if (no_args == 1) {
				if (SendToServer(currentserver, "XOVER") < 0){
					print_couldnt_send(currentserver);
					(void) fflush(stdout);
					return(DNNTP_REMOTE_GONE_ERR);
				}
			} else {	/* no_args == 2*/
				if (SendToServer(currentserver, "XOVER %s",
					args[1]) < 0){
					print_couldnt_send(currentserver);
					return(DNNTP_REMOTE_GONE_ERR);
				}
			}
			status = ReadFromServer(currentserver, str);
			if (status < 0) {
				print_couldnt_send(currentserver);
				return(DNNTP_REMOTE_GONE_ERR);
			}
			fputs(str, stdout);
			if (status == OK_OVER) {
#if DNNTP_POSSIBLE_XREF_ODDNESS
				xover_filtered(currentserver);
#else /*!DNNTP_POSSIBLE_XREF_ODDNESS*/
				while (ContReadFromServer(currentserver, str)
					> 0)
					/*ContReadFromServer returned 0 on
					the . so it wasn't sent...*/
					fputs(str, stdout);
#endif /*!DNNTP_POSSIBLE_XREF_ODDNESS*/
				fputs(".\r\n", stdout);
			}
			(void) fflush(stdout);
		} else {
			printf("%d no newsgroup selected\r\n", ERR_NCING);
			(void) fflush(stdout);
			return(DNNTP_NO_SETUP_ERR);
		}
	} else {
		/*or garbage, send error message*/
		printf("%d Usage: XOVER [article number|article range]\r\n",
			ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	return(DNNTP_SUCCESS);
}
#endif /*XOVER*/

/* date_cmd -- do the DATE command (just send it to current server)
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure

	(note: this is as simple as these routines can be...hum)
*/
int date_cmd( args, no_args)
char *args[];
int no_args;
{ 
char str[NNTP_STRLEN+1];
int status;
	if (no_args == 1) {
		if (currentserver < 0)
			currentserver = 0;
		(void) OpenServer(currentserver);
		if (serv_array[currentserver].curgroup[0])
			(void) SendGroup(currentserver, NULL);
		if (SendToServer(currentserver, "DATE") < 0){
			print_couldnt_send(currentserver);
			return(DNNTP_REMOTE_GONE_ERR);
		}
		status = ReadFromServer(currentserver, str);
		if (status < 0) {
			print_couldnt_send(currentserver);
			return(DNNTP_REMOTE_GONE_ERR);
		}
		fputs(str, stdout);
		(void) fflush(stdout);
	} else {
		printf("%d Usage: DATE\r\n", ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	return(DNNTP_SUCCESS);
}

#if MODE
/* mode_cmd -- do the MODE command (and mark servers for future reference)
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure

	(note: this is as simple as these routines can be...hum)
*/
int mode_cmd( args, no_args)
char *args[];
int no_args;
{ 
char str[NNTP_STRLEN+1];
int status, i, first_time = 1, ofunc;
	if ((no_args == 2) && (strcasecmp(args[1], "READER") == 0)){
		if (currentserver < 0) {
			currentserver = i = 0;
		} else {
			i = currentserver;
		}
		do {
			serv_array[i].misc_state = serv_array[i].connection;
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
		for(ofunc = 0; ofunc < DNNTP_OPEN_STAGES; ofunc++) {
		do {
			if (serv_array[i].misc_state < 0)
				(void) open_serv_func[ofunc](i);
				
			if (ofunc == (DNNTP_OPEN_STAGES-1)) {
				if (serv_array[i].curgroup[0])
					(void) SendGroup(i, NULL);
				if (serv_array[i].do_mode != 0) {
					if ((serv_array[i].connection >= 0)&&
					(SendToServer(i, "MODE READER") < 0)){
					int ind;
						print_couldnt_send(i);
						for(ind = i-1; ind >= 0; ind--)
						ClearServer(ind, -1);
						return (DNNTP_REMOTE_GONE_ERR);
					}
				}
			}
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
		} /*end of for loop over open stages*/
		i = currentserver;
		do {
			if (serv_array[i].do_mode != 0) {
				if (serv_array[i].connection >= 0)
					status = ReadFromServer(i, str);
				else
					status = ERR_GOODBYE;
				if ((status == OK_NOPOST) ||
					(status == OK_CANPOST)) {
					serv_array[i].do_mode = 1;
					if (first_time) {
						fputs(str, stdout);
						(void) fflush(stdout);
						first_time = 0;
					}
				} else {
					serv_array[i].do_mode = 0;
				}
			}
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
		if (first_time) {
			printf("%d none of the servers understand MODE\r\n",
				ERR_COMMAND);
			(void) fflush(stdout);
		}
	} else {
		printf("%d Usage: MODE READER\r\n", ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	return(DNNTP_SUCCESS);
}
#endif /*MODE*/

#if LISTGROUP
/* listgroup_cmd -- do the LISTGROUP command(a slight modified version of the
	group_cmd routine
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int listgroup_cmd( args, no_args)
char *args[];
int no_args;
{
int server;
int status;
char str[NNTP_STRLEN+1];
	if ((no_args == 2) || (no_args == 1)){
		/*figure out which connection this should be on*/
		if (no_args == 2) {
			Group2ServerPrep();
			server = Group2Server(args[1]);
			if (server < 0) {
				printf("%d Invalid group (failed group to \
server lookup)\r\n", ERR_NOGROUP);
				(void) fflush(stdout);
				return(DNNTP_SIMPLE_ERR);
			}
			currentserver = server;
		} else {	/*no_args == 1*/
			if ((currentserver < 0) ||
				(!serv_array[currentserver].curgroup[0])) {
				printf("%d no newsgroup selected\r\n",
					ERR_NCING);
				(void) fflush(stdout);
				return(DNNTP_NO_SETUP_ERR);
			}
			server = currentserver;
		}
			/* ERR_ACCESS not really the correct error number */
		if ((status = OpenServer(server)) < 0) {
			printf("%d group not avalible since %s is down\r\n",
				ERR_ACCESS, ServerName(server));
			(void) fflush(stdout);
			return(status);
		}
			/*send listgroup command....*/
		if (no_args == 2) {
			if ((status = SendToServer(server, "LISTGROUP %s",
				args[1])) < 0){
				if (serv_array[server].connection >= 0)
					printf("%d failure sending group \
command %s\r\n" ,ERR_FAULT, ServerName(server));
				else 
					printf("%d group not avalible since %s \
is down\r\n", ERR_ACCESS, ServerName(server));
				(void) fflush(stdout);
				return(status);
			}
			SetGroup(server, args[1], str, 0);
		} else {	/*no_args ==1*/
			if ((status = SendToServer(server, "LISTGROUP")) < 0){
				if (serv_array[server].connection >= 0)
					printf("%d failure sending group \
command %s\r\n" ,ERR_FAULT, ServerName(server));
				else 
					printf("%d group not avalible since %s \
is down\r\n", ERR_ACCESS, ServerName(server));
				(void) fflush(stdout);
				return(status);
			}
		}
		if ((status = ReadFromServer(server, str)) < 0){
			if (serv_array[server].connection >= 0)
				printf("%d failure sending group \
command %s\r\n" ,ERR_FAULT, ServerName(server));
			else 
				printf("%d group not avalible since %s \
is down\r\n", ERR_ACCESS, ServerName(server));
			(void) fflush(stdout);
			return(status);
		}
		fputs(str, stdout);
		if (status == OK_GROUP){
			while (ContReadFromServer(server, str) > 0)
				fputs(str, stdout);
				/*ContReadFromServer returned 0 on
				the . so it wasn't sent...*/
			fputs(".\r\n", stdout);
		}
		(void) fflush(stdout);
	} else {
		printf("%d Usage: LISTGROUP [<groupname>].\r\n", ERR_CMDSYN);
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	return(DNNTP_SUCCESS);
}
#endif /*LISTGROUP*/

/* print_nosend -- print message when sending stuff to server (done
	cause damn code appeared to many times.
    parameters
	servernumber	number of server where error happened
    return
	none
*/
void print_couldnt_send(servernumber)
int servernumber;
{
	printf("%d trouble sending command to real server %s \r\n",
		ERR_FAULT, ServerName(servernumber));
	(void) fflush(stdout);
}
