/* filter.c */
/* code to filter articles as we pass them along (right now, just twiddles
Xref: line), also code to do the same twidling to xover info*/
/*
 * HISTORY
 * $Log:	filter.c,v $
 * Revision 2.12  95/10/30  22:15:10  wbardwel
 * 	fix bug in handling of xovers where final feild isn't a tagged xref...
 * 	(not sure that this is really happening...but it ends up like this at some
 * 	point when it loops...and it is a minor hole anyway...)
 * 	[95/10/30  22:13:09  wbardwel]
 * 
 * Revision 2.11  94/12/15  00:44:58  wbardwel
 * 	fix xhdr xref deleting lines, and dump a tab at end of
 * 	non-xref overviews
 * 	[94/12/15  00:33:00  wbardwel]
 * 
 * Revision 2.10  94/07/12  17:26:48  wbardwel
 * 	update copyright, fix crashing on mal-formed overviews
 * 	[94/07/12  17:22:24  wbardwel]
 * 
 * Revision 2.9  94/05/27  17:21:17  wbardwel
 * 	fix??? long overviews code (and make it less likely to freak on
 * 	screwedup overviews)
 * 	[94/05/27  17:19:38  wbardwel]
 * 
 * Revision 2.8  94/03/29  17:07:26  wbardwel
 * 	fix xover xref filtering
 * 	[94/03/29  17:06:08  wbardwel]
 * 
 * Revision 2.7  94/02/21  22:56:15  wbardwel
 * 	try to catch totally trashed xover lines
 * 	[94/02/21  22:54:53  wbardwel]
 * 
 * Revision 2.6  94/02/16  16:47:09  wbardwel
 * 	fix stupid xover mangling bug (causing crashes)
 * 	[94/02/16  16:45:31  wbardwel]
 * 
 * Revision 2.5  94/01/06  04:39:31  wbardwel
 * 	allow turning on and off XREF filtering
 * 	[94/01/06  04:31:17  wbardwel]
 * 
 * Revision 2.4  93/11/02  22:34:45  wbardwel
 * 	include fixes
 * 	[93/11/02  22:01:24  wbardwel]
 * 
 * Revision 2.3  93/08/16  14:16:22  wbardwel
 * 	added support for xover lines longer than NNTP_STRLEN
 * 	[93/08/16  14:14:53  wbardwel]
 * 
 * Revision 2.2  93/08/13  21:18:43  wbardwel
 * 	[93/08/13  21:17:17  wbardwel]
 * 
 * 12-Aug-93  William Bardwell (wbardwel) at Carnegie-Mellon University
 *	Created.
 *
 * 
 */
/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

#if DNNTP_POSSIBLE_XREF_ODDNESS
/* send_art_filtered -- rip out bogus stuff from xref: line
    parameter
	serverno: server number that command was issued on
    return
	<0 if fails
*/
int send_art_filtered(serverno)
int serverno;
{
char str[NNTP_STRLEN+1];
int status, first_time;
extern char servername[];
char *tmp, *startgroup, *tmp2;
	do {
		first_time = 1;
		status = ContReadFromServer(serverno, str);
		if (status < 0)
			return (status);
		else if (status == 0)
			return (DNNTP_SUCCESS);
		if (!strncasecmp(str, "xref:", 5)) {
			tmp = index(str+6, ' ');
			if (tmp++) {
			    while ((*tmp != '\r') && *tmp != '\n') {
			    	startgroup = tmp;
				tmp = index(tmp, ':');
				if (tmp) {
				    *tmp = '\0';
				    tmp2 = index(++tmp, ' ');
				    if (!tmp2)
					tmp2 = index(tmp, '\r');
				    if (!tmp2)
					tmp2 = index(tmp, '\n');
				    if (!tmp2)
					break;
				    *tmp2 = '\0';
				    if (Group2Server(startgroup) ==
					serverno) {
					if (first_time) {
					    printf("Xref: %s", servername);
					    first_time = 0;
					}
					printf(" %s:%s", startgroup, tmp);
				    }
				    tmp = tmp2+1;
				} else
					break;
			    }
			}
			if (!first_time)
				fputs("\r\n", stdout);
		} else
			fputs(str, stdout);
	} while (first_time && strcmp(str, "\r\n") && strcmp(str, "\n"));
	while (ContReadFromServer(serverno, str) > 0)
		fputs(str, stdout);
	return (DNNTP_SUCCESS);
}

/* xover_filtered -- rip out bogus stuff from xref: field (which may or may
	not be present)
    parameter
	serverno: server number that command was issued on
    return
	<0 if fails
*/
#define NUM_NOV_TABS 7
int xover_filtered(serverno)
{
char *tmp, *tmp2, tmp3, *startgroup;
char *str, *tmpstr;
int str_len, tmpstrlen;
int status, first_time, i;
extern char servername[];
	str_len = NNTP_STRLEN+1;
	str = malloc(str_len);
	if (!str) {
#if SYSLOG
		syslog(LOG_ERR, "xover_filtered: ran out of memory");
#endif /*SYSLOG*/
		ClearServer(serverno);
		return(DNNTP_INTERNAL_ERR);
	}
	while((status = ContReadFromServerBufNum(serverno, str, str_len))
		> 0) {
		tmpstr = str;
		while ((*(tmpstr+(tmpstrlen = strlen(tmpstr))-1) != '\n') &&
			(*(tmpstr+tmpstrlen-1) != '\n')) { 
			str_len += NNTP_STRLEN;
			str = realloc(str, str_len);
			if (!str) {
#if SYSLOG
				syslog(LOG_ERR, "xover_filtered: ran out of \
memory");
#endif /*SYSLOG*/
				ClearServer(serverno);
				return(DNNTP_INTERNAL_ERR);
			}
			tmpstr = str + strlen(str);
			status = ContReadFromServerBufNum(serverno,
				tmpstr, str_len - (tmpstr-str) );
			if (status < 0) {
				free(str);
				ClearServer(serverno);
				return status;
			} else if (status == 0) {
				/*crap!!!things are broken at this point!!!*/
#if SYSLOG
				syslog(LOG_ERR, "xover_filtered: hosed xover \
filtering, got a 0 on ContRead, but we are continueing");
#endif /*SYSLOG*/
				break;
			}
		}
		first_time = 1;
		for (i = 0, (tmp = index(str, '\t'));
			(i < NUM_NOV_TABS) && tmp; i++)
			tmp = index(tmp+1, '\t');
		if(!(tmp++))
			fputs(str, stdout);
		else {
			*(tmp-1) = '\0';
			fputs(str, stdout);
			while(tmp && *tmp && (*tmp != '\r') && (*tmp != '\n')) {
				if (!strncasecmp(tmp, "xref:", 5)) {
				tmp = index(tmp+6, ' ');
				if (tmp++) {
				while ((*tmp != '\t') && (*tmp != '\r') &&
					(*tmp != '\n')) {
				    	startgroup = tmp;
					tmp = index(tmp, ':');
					if (tmp) {
					    *tmp = '\0';
					    tmp2 = index(++tmp, ' ');
					    if (!tmp2)
						tmp2 = index(tmp, '\t');
					    if (!tmp2)
						tmp2 = index(tmp, '\r');
					    if (!tmp2)
						tmp2 = index(tmp, '\n');
					    if (!tmp2)
						break;
					    tmp3 = *tmp2;
					    *tmp2 = '\0';
					    if (Group2Server(startgroup) ==
						serverno) {
						if (first_time) {
						    printf("\tXref: %s",
							servername);
						    first_time = 0;
						}
						printf(" %s:%s", startgroup,
							tmp);
					    }
					    tmp = tmp2+1;
					} else {
					    tmp2 = startgroup+strlen(startgroup) - 2;
					    break;
					}
			    	}
				tmp = tmp2+1;
			    	} else /* if tmp */
					tmp = index(tmp+5, '\t');
				} else { /*not a xref: item*/
					tmp2 = index(tmp, '\t');
					if (tmp2)
						*tmp2 = '\0';
					fputs(tmp, stdout);
					if (tmp2)
						tmp = tmp2+1;
					else /*nothing after this, exit
						the loop*/
						*tmp='\0';
				}
			}
			if (first_time)
				fputs("\t\r\n", stdout);
			else
				fputs("\r\n", stdout);
		}
	}
	free(str);
	if (status < 0)
		return (status);
	/*else if (status == 0)*/
	return (DNNTP_SUCCESS);
}

/* xhdr_xref_filtered -- rip out bogus stuff from xref info (ie each line)
    parameter
	serverno: server number that command was issued on
    return
	<0 if fails
*/
int xhdr_xref_filtered(serverno)
int serverno;
{
char *tmp, *tmp2, *startgroup;
char str[NNTP_STRLEN+1];
int status, first_time;
extern char servername[];

	while((status = ContReadFromServer(serverno, str)) > 0) {
		first_time = 1;
		tmp = index(str+1, ' ');
		if (!tmp) {
			fputs(str, stdout);
			continue;
		}
		*tmp = '\0';
		tmp = index(tmp+1, ' ');
		if (tmp++) {
			while ((*tmp != '\r') && *tmp != '\n') {
			    	startgroup = tmp;
				tmp = index(tmp, ':');
				if (tmp) {
				    *tmp = '\0';
				    tmp2 = index(++tmp, ' ');
				    if (!tmp2)
					tmp2 = index(tmp, '\r');
				    if (!tmp2)
					tmp2 = index(tmp, '\n');
				    if (!tmp2)
					break;
				    *tmp2 = '\0';
				    if (Group2Server(startgroup) ==
					serverno) {
					if (first_time) {
					    printf("%s %s", str, servername);
					    first_time = 0;
					}
					printf(" %s:%s", startgroup, tmp);
				    }
				    tmp = tmp2+1;
				} else
					break;
			}
			if (!first_time)
				fputs("\r\n", stdout);
		} else {
			printf("%s %s", str, str+strlen(str)+1);
		}
	}
	if (status < 0)
		return (status);
	/*else if (status == 0)*/
	return (DNNTP_SUCCESS);
}
#endif /*DNNTP_POSSIBLE_XREF_ODDNESS*/
