/* group.c */
/* code to manage group mappings to servers */
/* Mon Feb  8 15:27:31 EST 1993 wbardwel : created*/
/*
 * HISTORY
 * $Log:	group.c,v $
 * Revision 2.46  95/02/21  16:40:13  wbardwel
 * 	fix possible eof's in a list command...
 * 	[95/02/21  16:33:58  wbardwel]
 * 
 * Revision 2.45  95/01/09  22:43:30  wbardwel
 * 	kill lamo copied write (of list response)
 * 	[95/01/09  22:41:31  wbardwel]
 * 
 * Revision 2.44  95/01/09  03:00:31  wbardwel
 * 	clean up caching some...
 * 	[95/01/09  02:57:35  wbardwel]
 * 
 * Revision 2.43  95/01/09  00:45:52  wbardwel
 * 	minor cleanups to Prep
 * 	make Expire not call StartAdd
 * 	make EndAdd not do a dump
 * 	[95/01/09  00:43:11  wbardwel]
 * 
 * Revision 2.42  95/01/05  13:58:16  wbardwel
 * 	add G2SItem
 * 	[95/01/05  13:28:31  wbardwel]
 * 
 * Revision 2.41  95/01/01  23:20:44  wbardwel
 * 	make G2SLCFresh return time of youngest file
 * 	and make Prep install that as the time
 * 	[95/01/01  23:18:04  wbardwel]
 * 
 * Revision 2.40  94/12/15  12:07:24  wbardwel
 * 	fix so we only malloc exactly what we need after zero suppression
 * 	in the "stuff" part of an active entry
 * 	[94/12/15  12:06:35  wbardwel]
 * 
 * Revision 2.39  94/12/15  00:45:05  wbardwel
 * 	support alternate timeouts on disk cache for Prep
 * 	remove extra zeros...
 * 	[94/12/15  00:34:26  wbardwel]
 * 
 * Revision 2.38  94/10/25  18:16:48  wbardwel
 * 	generate syslog messages when can't create cache files
 * 	[94/10/25  18:15:12  wbardwel]
 * 
 * Revision 2.37  94/09/06  18:43:29  wbardwel
 * 	fix bug when just one server (or no dup groups)
 * 	[94/09/06  18:42:24  wbardwel]
 * 
 * Revision 2.36  94/08/24  01:17:40  wbardwel
 * 	some fixes on flock compat stuff
 * 	[94/08/24  01:17:14  wbardwel]
 * 
 * Revision 2.35  94/08/15  17:14:21  wbardwel
 * 	add including sys/select.h for AIX
 * 	[94/08/15  16:46:53  wbardwel]
 * 
 * Revision 2.34  94/07/30  01:48:26  wbardwel
 * 	fix nasty bit using memory that was freed (broke on linux...)
 * 	[94/07/30  01:42:40  wbardwel]
 * 
 * Revision 2.33  94/07/12  17:26:54  wbardwel
 * 	update copyright, fix for svr4 (and other sysv)
 * 	[94/07/12  17:23:02  wbardwel]
 * 
 * Revision 2.32  94/06/14  18:02:00  wbardwel
 * 	pull in fixes from list_cmd where code was copied from...
 * 	(fixes hangs)
 * 	[94/06/14  17:56:38  wbardwel]
 * 
 * Revision 2.31  94/05/26  04:35:14  wbardwel
 * 	fix selects (should fix old dead dnntpd...)
 * 	[94/05/26  04:33:16  wbardwel]
 * 
 * Revision 2.30  94/05/06  12:43:20  wbardwel
 * 	make sure pid and the string used for cache files are always
 * 	consistantly maintained
 * 	[94/05/06  12:42:59  wbardwel]
 * 
 * Revision 2.29  94/04/21  01:49:00  wbardwel
 * 	rewrite of duplicate elimination in group lists (was broken)
 * 	[94/04/21  01:46:26  wbardwel]
 * 
 * Revision 2.28  94/03/02  16:24:15  wbardwel
 * 	fixes for -c option in main.c
 * 	[94/03/02  16:23:04  wbardwel]
 * 
 * Revision 2.27  94/03/02  16:09:57  wbardwel
 * 	changes to make -c work in main.c
 * 	[94/03/02  16:08:37  wbardwel]
 * 
 * Revision 2.26  94/02/21  22:56:19  wbardwel
 * 	possible clean up for mal-formed LIST results
 * 	[94/02/21  22:55:24  wbardwel]
 * 
 * Revision 2.25  94/02/16  17:20:48  wbardwel
 * 	change all calles to salloc to be to strdup
 * 	[94/02/16  17:20:15  wbardwel]
 * 
 * Revision 2.24  94/02/16  16:47:16  wbardwel
 * 	make lowest duplicate group listing survive (determinalistically)
 * 	[94/02/16  16:46:10  wbardwel]
 * 
 * Revision 2.23  94/01/25  04:44:06  wbardwel
 * 	strdup/salloc fix
 * 	[94/01/25  04:43:09  wbardwel]
 * 
 * Revision 2.22  94/01/25  03:42:04  wbardwel
 * 	supply strdup where needed
 * 	[94/01/25  03:40:39  wbardwel]
 * 
 * Revision 2.21  94/01/21  17:43:30  wbardwel
 * 	deal with OS's missing flock()
 * 	[94/01/21  17:40:51  wbardwel]
 * 
 * Revision 2.20  94/01/06  04:39:39  wbardwel
 * 	fix bug in DupElim (and remove dead variable)
 * 	[94/01/06  04:32:39  wbardwel]
 * 
 * Revision 2.19  93/12/06  03:18:14  wbardwel
 * 	fixes for timeout code (see list_cmd.c)
 * 	[93/12/06  03:16:07  wbardwel]
 * 
 * Revision 2.18  93/12/02  03:28:57  wbardwel
 * 	Nasty bug...where it wouldn't do list right
 * 	[93/12/02  03:25:50  wbardwel]
 * 
 * Revision 2.17  93/12/02  00:45:31  wbardwel
 * 	misc cleanup, misc bug fixes see list_cmd.c for details
 * 	[93/12/02  00:30:01  wbardwel]
 * 
 * Revision 2.16  93/11/02  22:34:57  wbardwel
 * 	include fixes
 * 	[93/11/02  22:01:40  wbardwel]
 * 
 * Revision 2.15  93/10/27  16:02:45  wbardwel
 * 	misc cleanup, modify RealGroup2ServerPrep to match new list_cmd() code.
 * 	[93/10/27  15:23:40  wbardwel]
 * 
 * Revision 2.14  93/08/18  19:41:36  wbardwel
 * 	bug fixes for when a machine is down.
 * 	[93/08/18  19:35:34  wbardwel]
 * 
 * Revision 2.13  93/08/17  17:35:13  wbardwel
 * 	some very modest speed ups. (and kill bogus puts)
 * 	[93/08/17  17:33:24  wbardwel]
 * 
 * Revision 2.12  93/08/10  19:19:40  wbardwel
 * 	bug fixes and performance enhancements to ListCache system
 * 	[93/08/10  19:16:05  wbardwel]
 * 
 * Revision 2.11  93/08/05  01:23:31  wbardwel
 * 	new style Group2Server{EndAdd,Dump} to hopefully speed up list_cmd alittle
 * 	[93/08/05  01:22:40  wbardwel]
 * 
 * Revision 2.10  93/08/05  00:26:13  wbardwel
 * 	remove STRTOLOWER
 * 	[93/08/05  00:23:55  wbardwel]
 * 
 * Revision 2.9  93/08/04  16:37:23  wbardwel
 * 	various performance hack/fixes
 * 	including new style AddGroup2Server
 * 	[93/08/04  00:11:50  wbardwel]
 * 
 * Revision 2.8  93/07/28  16:19:01  wbardwel
 * 	fixed typo in error message that was causing a gcc warning
 * 	[93/07/28  16:17:38  wbardwel]
 * 
 * Revision 2.7  93/07/28  13:25:29  wbardwel
 * 	fiddled types for groupfile_array
 * 	added code to Prep can load list info from disc cache...
 * 	paralellized Prep (just like list_cmd)
 * 	made AddGroup2Server (and friends) so you don't need to add
 * 	entries in blocks of one servers entries.
 * 	bug fixes in disc cache code
 * 	[93/07/28  13:06:24  wbardwel]
 * 
 * Revision 2.6  93/07/21  22:41:23  wbardwel
 * 	bug fixes (to make OSF happier)
 * 	[93/07/21  22:34:21  wbardwel]
 * 
 * Revision 2.5  93/07/01  19:49:07  wbardwel
 * 	bug fixes for list disk cache
 * 	[93/07/01  19:14:32  wbardwel]
 * 
 * Revision 2.4  93/06/23  00:50:49  wbardwel
 * 	added code for disk list cache
 * 	[93/06/23  00:26:40  wbardwel]
 * 
 * 	fix minor bug causing multiple LIST commands.
 * 	fixed psuedo bug where if a new list of a group came by,
 * 	the misc LIST info wasn't updated in the cache.
 * 		(I don't think this can every really happen...but
 * 		that might change)
 * 	[93/06/15  17:35:39  wbardwel]
 * 
 * Revision 2.3  93/06/02  15:13:07  wbardwel
 * 	added since RTs don't seem to get bserch def otherwise
 * 	#include <stdlib.h>
 * 	[93/06/02  15:12:55  wbardwel]
 * 
 * Revision 2.2  93/06/02  14:52:23  wbardwel
 * 	Added fields to group to server info, so it can reconstruct the results
 * 	of the LIST command.  Added code so you can check if the cache is stale,
 * 	and so you can dump the cache to stdout.
 * 	[93/05/27            wbardwel]
 * 
 * 	Made it possible to use a static group file (that is compiled in)
 * 	see group_static.h
 * 	[93/05/20            wbardwel]
 * 
 * 	change to use the groups-conf file as a supliment (temporarily)
 * 	and do basic lookup on list of groups.
 * 	[93/05/10            wbardwel]
 * 
 * 	Created.
 * 	[93/02/08            wbardwel]
 * 
*/
/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"
#include <errno.h>
#if DNNTP_DISK_LIST_CACHE
#include <pwd.h>
#include <sys/stat.h>
#include <sys/file.h>
#ifdef USG
#include <sys/fcntl.h>
#endif /*USG*/
#ifndef __STDC__
#include <varargs.h>    /*for concat*/
#else /*__STDC__*/
#include <stdarg.h>
#endif /*__STDC__*/
#ifdef AIX
#include <sys/select.h>
#endif /*AIX*/

#if !defined(LOCK_EX) && !defined(NO_FLOCK)
/* this only works if you have the CMUCS libc.h...*/
#include <libc.h>
#endif /*!defined(LOCK_EX) && !defined(NO_FLOCK)*/

#ifndef LOCK_EX
static int flock();
#define NO_FLOCK 1	/*say we need an flock routine*/
#define LOCK_SH         1       /* shared lock */
#define LOCK_EX         2       /* exclusive lock */
#define LOCK_UN         8       /* unlock */
#endif /*!defined(LOCK_EX) after loading libc.h*/

#endif /*DNNTP_DISK_LIST_CACHE*/

#include <sys/time.h>

#ifndef SEEK_END
#ifdef L_XTND
#define SEEK_END L_XTND
#else /*!L_XTND*/
#define SEEK_END 2
#endif /*!L_XTND*/
#endif /*!SEEK_END*/

#if !SYSTEM_STRDUP && !CMUCS
char *strdup();
#endif /*!SYSTEM_STRDUP && !CMUCS*/ 

#if DNNTP_USE_GROUP_INFO

#if !DNNTP_USE_STATIC_GROUP
fgroup_t *groupfile_array = NULL;
int groupfile_array_len = 0;
#else /*DNNTP_USE_STATIC_GROUP*/
/*will get info from group_static.h*/
#include "group_static.h"
#endif /*DNNTP_USE_STATIC_GROUP*/

#if !DNNTP_USE_STATIC_GROUP
/* LoadGroupFile -- load structure with groups (from config file)
	note: must be called after LoadServers for checks on server numbers
	to work
    parameters
	filename: name of configuration file to read from
		(file is in format:
		"cmu   0
		alt.fan.mike-jittlov    1
		alt    2
		comp   0
		...
		"
		where ending in a . means partial group name, and
		# means a comment.)
		(note: must be an entry for every highest level group)
		also note: order matters...
    return
	>=0 on success, <0 on failure
*/
/*note: odd similarity with LoadServers...*/
int LoadGroupFile(filename)
char *filename;	/*name of file to load from*/
{
FILE *fp;
	/*holds line with group name and server (or comment)*/
/*make sure we get enough space(cause we are reading lines into static
buffer, sortof each component must fit in static buffer) (tough luck...)*/
char line[DNNTP_GROUP_STRLEN];
char group[DNNTP_GROUP_STRLEN];
int server;
	
	if ((fp = fopen(filename, "r")) == NULL) {
#if LOG
#if SYSLOG
		syslog(LOG_DEBUG, "LoadGroupFile: unable to open groups file: %m");
#endif /*SYSLOG*/
#endif /*LOG*/
		return(DNNTP_SUCCESS);
	}
		/*skip initial spaces on a line(of multiple lines)*/
	clear_spaces(fp);
	while (fgets(line, sizeof(line), fp)) {
		if (line[0] =='#'){
				/*go until last fgets on a line*/
			while (line[strlen(line)-1] != '\n'){
				if (!fgets(line, sizeof(line), fp))
					break;
			}
			continue;
		}
		if (sscanf(line, "%s", group) < 1){
#if SYSLOG
			syslog(LOG_ERR, "LoadGroupFile: unable to parse groups \
file: %m");
#endif /*SYSLOG*/
			return(DNNTP_NO_SETUP_ERR);
		}
			/*if we didn't get whole line*/
		if (line[strlen(line)-1] != '\n'){
		int i;	/*misc index*/
				/*move stuff we haven't looked at to
				begining of buffer*/
			for (i=strlen(group); isspace(line[i]); i++)
				;
			if (line[i])
				strcpy(line, line+i);
			else
				line[0] = '\0';
			clear_spaces(fp);
			i = strlen(line);
			if (!fgets(line+i, sizeof(line)-i, fp))
				break;
			if (sscanf(line, "%d", &server) < 1) {
#if SYSLOG
				syslog(LOG_ERR, "LoadGroupFile: unable to parse \
groups file: %m");
#endif /*SYSLOG*/
				return(DNNTP_NO_SETUP_ERR);
			}
		} else if (sscanf(line+strlen(group), "%d", &server) < 1) {
#if SYSLOG
				syslog(LOG_ERR, "LoadGroupFile: unable to parse \
groups file: %m");
#endif /*SYSLOG*/
				return(DNNTP_NO_SETUP_ERR);
		}

		/*now we have a group and server, lets do something with it*/
			/*expand/make groupfile_array big enough...*/
		if (groupfile_array_len)
			groupfile_array = (dgroup_t *)realloc(groupfile_array,
				(++groupfile_array_len)*sizeof(dgroup_t));
		else
			groupfile_array = (dgroup_t *)malloc(
				(++groupfile_array_len)*sizeof(dgroup_t));
		if (!groupfile_array){
#if SYSLOG
			syslog(LOG_CRIT, "LoadGroupFile: ran out of memory");
#endif /*SYSLOG*/
			return(DNNTP_INTERNAL_ERR);
		}
			/*copy stuff to dgroup_t structure*/
		groupfile_array[groupfile_array_len-1].group = strdup(group);
		if (!groupfile_array[groupfile_array_len-1].group){
#if SYSLOG
			syslog(LOG_CRIT, "LoadGroupFile: ran out of memory");
#endif /*SYSLOG*/
			return(DNNTP_INTERNAL_ERR);
		}

		if (!ValidServer(server)) {
#if SYSLOG
			syslog(LOG_ERR, "LoadGroupFile: illegal server number: %d",
				server);
#endif /*SYSLOG*/
			return(DNNTP_NO_SETUP_ERR);
		} else
			groupfile_array[groupfile_array_len-1].server_num = server;

		clear_spaces(fp);
	}
	return(groupfile_array_len);
}
#endif /*!DNNTP_USE_STATIC_GROUP*/

/* Group2ServerFileLookup(groupname) -- returns what server (number) should
	have given group
	(finds the first group name in the groupfile_array where last
	matching letter was the final '.' in groupfile_array.group or
	it was exact match)
    parameters
	groupname: name of group to lookup
    return
	server number if succeeds (which is >=0) or <0 if not
*/
int Group2ServerFileLookup(groupname)
char *groupname;
{
register char *tmp_grp;	/*tmp pointer into groupname*/
register char *tmp_chk;	/*tmp pointer into group (from groupfile_array)*/
register int group_index;	/*how far along we are in*/
	for (group_index=0; group_index < groupfile_array_len; group_index++) {
		for ( tmp_grp = groupname,
			tmp_chk = groupfile_array[group_index].group;
			((*tmp_grp == *tmp_chk) && (*tmp_grp) &&
			(*tmp_chk));
			tmp_grp++, tmp_chk++);
			/*if both strings are done or last matched charater
			was a . and the groupfile_array one has ended, then this
			is a match*/
		if ((*tmp_grp == *tmp_chk) || ( (*tmp_chk == '\0') &&
			(*tmp_grp == '.')))
			return (groupfile_array[group_index].server_num);
	}
	return(DNNTP_SIMPLE_ERR);
}
#endif /*DNNTP_USE_GROUP_INFO*/


	/* items to hold primary group info*/
dgroup_t *group_array = NULL;
int group_array_len = 0;
static int group_array_sortedlen = 0;
static int group_array_memlen = 0;
time_t group_array_secs = 0;
#if DNNTP_DISK_LIST_CACHE
static FILE **group_array_diskcache = NULL;
#endif /*DNNTP_DISK_LIST_CACHE*/

	/* function to use with bsearch's to fine stuff*/
	/*manual, brutal inlining of strcmp....*/
static int cmp_group_n_str(key, a)
char *key;
dgroup_t *a;
{
/*return(strcmp(key, a->group));*/
register char *s2 = a->group;
        while (*key == *s2++) {
                if (*key++=='\0')
                        return(0);
	}
        return(*key - *--s2);
}
	/* function to use with qsort to sort...*/
static int cmp_groups(a, b)
dgroup_t *a, *b;
{
int ret;
	if ((ret = strcmp(a->group, b->group)) == 0)
		return(a->server_num - b->server_num);
	return(ret);
}

/* Group2ServerClear() -- clears all state in the group to server lookup stuff
    parameters
	none
    return
	none
*/
void Group2ServerClear() {
int i;
	if (group_array) {
		for(i = 0; i < group_array_len; i++) {
			if(group_array[i].group)
				free(group_array[i].group);
			if(group_array[i].stuff)
				free(group_array[i].stuff);
		}
		free(group_array);
		group_array = NULL;
	}
	group_array_len = group_array_memlen = group_array_sortedlen = 0;
	group_array_secs = 0;
}

/* Group2Server(groupname) -- returns what server (number) should
	have given group (bsearches in group_array)
	people had better call Group2ServerPrep before they call
	Group2Server (and when they aren't in the middle of a
	server command) or this may not work out...
    parameters
	groupname: name of group to lookup
    return
	server number if succeeds (which is >=0) or <0 if not
*/
int Group2Server(groupname)
char *groupname;
{
dgroup_t *tmp;
	/*first make sure we are initialized*/
	if (group_array == NULL)
		return (DNNTP_INTERNAL_ERR);
	tmp = ansi_bsearch(groupname, group_array, group_array_sortedlen,
		sizeof(dgroup_t), cmp_group_n_str);
	if (tmp == NULL)
		return (DNNTP_SIMPLE_ERR);
	return (tmp->server_num);
}
/* Group2ServerItem(groupname) -- returns array index for group
	people had better call Group2ServerPrep before they call
	Group2Server (and when they aren't in the middle of a
	server command) or this may not work out...
    parameters
	groupname: name of group to lookup
    return
	index into array (or <0 if not found)
*/
int Group2ServerItem(groupname)
char *groupname;
{
dgroup_t *tmp;
	/*first make sure we are initialized*/
	if (group_array == NULL)
		return (DNNTP_INTERNAL_ERR);
	tmp = ansi_bsearch(groupname, group_array, group_array_sortedlen,
		sizeof(dgroup_t), cmp_group_n_str);
	if (tmp == NULL)
		return (DNNTP_SIMPLE_ERR);
	return (tmp-group_array);
}


/*from list_cmd.c*/
void free_stuff();
long find_eol();

/* see dnntp.h for Group2ServerPrep*/
/* RealGroup2ServerPrep -- says that routine may call Group2Server sometime soon
	(and that it is not in the middle of a command with a server)
	this makes sure that the group_array info is properly loaded.
    parameters
	RealGroup2ServerPrepTimeout - how old a file can be and still
		be prefered over running commands
	always_load - always ok to fetch fresh files?
		(if not, only do so if not already run LIST once...)
		1 means always ok
		0 means only if list hasn't been done
		-1 means never(except server outages)
    return
	>=0 if succeds <0 if fails
*/
int RealGroup2ServerPrep(timeout, always_load)
int timeout, always_load;
{
int i, status;
int ofunc;
static int done_list = 0;
time_t file_time;
		/*bogus, cause nothing in group_array yet, but this sets the
		time stamp too...*/
	/*run list commands everywhere*/ /*loosely copied from list_cmd()*/
	if (currentserver < 0)
		currentserver = i = 0;
	else
		i = currentserver;
#if DNNTP_DISK_LIST_CACHE
	if ((always_load >= 0) && (always_load || !done_list) &&
		(file_time = Group2ServerListCacheFresh(timeout))) {
		Group2ServerListCacheKillAll();
		/* nb. better not be any data in memory cache!*/
		do {
			if (Group2ServerListCacheFetch(i) < 0) {
#if SYSLOG
				syslog(LOG_CRIT, "RealGroup2ServerPrep: \
couldn't read 'Fresh' caches");
#endif /*SYSLOG*/
				printf("%d couldn't read list caches???\r\n",
					ERR_GOODBYE);
				(void) fflush(stdout);
				QuitAllServers();
				exit (1);
			}
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
		Group2ServerEndAdd();
			/*force a read, this may be old data*/
		group_array_secs = file_time;
		return(DNNTP_SUCCESS);
	} else {
#endif /*DNNTP_DISK_LIST_CACHE*/
		Group2ServerClear();
		Group2ServerStartAdd();
	do {
		serv_array[i].misc_state = serv_array[i].connection;
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
	for(ofunc = 0; ofunc < DNNTP_OPEN_STAGES; ofunc++) {
	do {
		if (serv_array[i].misc_state < 0) {
			if (open_serv_func[ofunc](i) < 0) {
#if DNNTP_DISK_LIST_CACHE
				if (Group2ServerListCacheFetch(i) < 0) {
#endif /*DNNTP_DISK_LIST_CACHE*/
					printf("%d couldn't open server %s\r\n",
						ERR_GOODBYE, ServerName(i));
					(void) fflush(stdout);
					QuitAllServers();
					Group2ServerListCacheKillAll();
					exit (1);
#if DNNTP_DISK_LIST_CACHE
				}
				i = (i+1)%serv_array_len;
				group_array_secs = 0;
				continue;
#endif /*DNNTP_DISK_LIST_CACHE*/
			}
		}
		if ((ofunc == (DNNTP_OPEN_STAGES-1)) &&
			(serv_array[i].connection >= 0)){
			if (serv_array[i].curgroup[0])
				SendGroup(i, NULL);
			status = SendToServer(i, "LIST");
			if ( status < 0) {
				CloseServer(i);
#if DNNTP_DISK_LIST_CACHE
				if (Group2ServerListCacheFetch(i) < 0) {
#endif /*DNNTP_DISK_LIST_CACHE*/
					printf("%d couldn't run LIST command on server \
%s\r\n", ERR_GOODBYE, ServerName(i));
					(void) fflush(stdout);
					QuitAllServers();
					Group2ServerListCacheKillAll();
					exit (1);
#if DNNTP_DISK_LIST_CACHE
				}
				i = (i+1)%serv_array_len;
				group_array_secs = 0;
				continue;
#endif /*DNNTP_DISK_LIST_CACHE*/
			}
				/* have send a list command...*/
			done_list = 1;
		}
		i = (i+1)%serv_array_len;
	} while (i != currentserver);
	} /*end of for loop on stages of open*/
		/*from list_cmd, select/read based LIST command*/
	{
	char **buffers = NULL, **read_args;
	long *bufremains = NULL;
	fd_set readready;
	int i;	/*misc index*/
	int maxfd;
	long readret, next_eol, prev_eol;

		/*create buffers for servers*/
		buffers = (char **)calloc(serv_array_len, sizeof(char *));
		bufremains = (long *)calloc(serv_array_len, sizeof(long));
		if (!buffers || !bufremains) {
#if SYSLOG
			syslog(LOG_CRIT, "Group2ServerPrep: ran out of memory");
#endif /*SYSLOG*/
			printf("%d server out of memory\r\n", ERR_GOODBYE);
			QuitAllServers();
			exit(1);
		}
		FD_ZERO(&readready);
		for (i=0, maxfd = 0, ofunc = 0; i < serv_array_len; i++) {
		    if (serv_array[i].connection >= 0) {
#if DNNTP_LIST_MEMBUF_PAGEALIGN
			buffers[i] = malloc(DNNTP_LIST_MEMBUFSIZE+
				DNNTP_LIST_MEMPAGESIZE);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
			buffers[i] = malloc(DNNTP_LIST_MEMBUFSIZE);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
			if (!buffers[i]) {
#if SYSLOG
			    syslog(LOG_CRIT, "Group2ServerPrep: ran out of memory");
#endif /*SYSLOG*/
			    printf("%d server out of memory\r\n",
				ERR_GOODBYE);
			    QuitAllServers();
#if DNNTP_DISK_LIST_CACHE
			    Group2ServerListCacheKillAll();
#endif /*DNNTP_DISK_LIST_CACHE*/
			    exit(1);
			}
#if DNNTP_LIST_MEMBUF_PAGEALIGN
			buffers[i] += DNNTP_LIST_MEMPAGESIZE;
#endif /*DNNTP_LIST_MEMBUF_PAGEALIGN*/
			maxfd = MAX(serv_array[i].connection, maxfd);
			FD_SET(serv_array[i].connection, &readready);
			ofunc++;
			serv_array[i].misc_state = 1;
		    } else
			serv_array[i].misc_state = 0;
		}
			
		/*loop*/
		do {
		    if (ofunc > 1) { /*if more than one fd left*/
				/*select for reads, no timeout*/
				/* doesn't catch other errors...*/
			while (((readret = select(maxfd+1, &readready,
				NULL, NULL, NULL)) < 0) &&
				(errno == EINTR))
				errno = 0;
				if (readret < 0) {
#if SYSLOG
					syslog(LOG_CRIT, "Group2ServerPrep: select %m");
#endif /*SYSLOG*/
					printf("%d server internal error\r\n",
						ERR_GOODBYE);
					QuitAllServers();
#if DNNTP_DISK_LIST_CACHE
					Group2ServerListCacheKillAll();
#endif /*DNNTP_DISK_LIST_CACHE*/
					exit(1);
				}
		    }
			/*for all servers' fds' which have mask bit set*/
			for (i = 0; i < serv_array_len; i++) {
			    if (ofunc > 1) {
				if ((buffers[i] == NULL) ||
				    !FD_ISSET(serv_array[i].connection,
				    &readready))
				    continue;
			    } else {
				if ((buffers[i] == NULL) ||
				    (serv_array[i].connection < 0))
				    continue;
			    }
				/*read a buffers(-prev amount) worth*/
				/* doesn't catch other errors...*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
			    while (((readret = read(serv_array[i].connection,
				buffers[i], DNNTP_LIST_MEMBUFSIZE)) < 0) && 
				(errno == EINTR))
				errno = 0;
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
			    while (((readret = read(serv_array[i].connection,
				buffers[i]+bufremains[i],
				DNNTP_LIST_MEMBUFSIZE-bufremains[i])) < 0) &&
				(errno == EINTR))
				errno = 0;
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				/*do some real error handling here...*/
			    if (readret <= 0) {
#if SYSLOG
				syslog(LOG_ERR, "Group2ServerPrep: error read server '%s' %m", ServerName(i));
#endif /*SYSLOG*/
				if (serv_array[i].misc_state) {
					CloseServer(i);
					ofunc--;
#if DNNTP_LIST_MEMBUF_PAGEALIGN
					free(buffers[i]-DNNTP_LIST_MEMPAGESIZE);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					free(buffers[i]);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						/*mark as gone*/
					buffers[i] = NULL;
#if DNNTP_DISK_LIST_CACHE
					if (Group2ServerListCacheFetch(i) <
						0 ) {
						Group2ServerListCacheKillAll();
#endif /*DNNTP_DISK_LIST_CACHE*/
						print_couldnt_send(i);
						free_stuff(&buffers,
							&bufremains);
						return(DNNTP_REMOTE_GONE_ERR);
#if DNNTP_DISK_LIST_CACHE
					}
					i = (i+1)%serv_array_len;
					continue;
#endif /*DNNTP_DISK_LIST_CACHE*/
				} else { /*partway into reply*/
					QuitAllServers();
#if DNNTP_DISK_LIST_CACHE
					Group2ServerListCacheKillAll();
#endif /*DNNTP_DISK_LIST_CACHE*/
					exit(1);
				}
			    }
			    prev_eol = 0;
				/*while find eol(buf), and !=.\r\n*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
			    while ((( next_eol = find_eol(buffers[i]+
				prev_eol-bufremains[i], (bufremains[i]
				+ readret)- prev_eol)) >= 0) &&
				strncmp(buffers[i]+prev_eol-
				bufremains[i], ".\r\n", MIN(4, next_eol+1)))
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
			    while ((( next_eol = find_eol(buffers[i]+
				prev_eol, (bufremains[i] + readret)-
				prev_eol)) >= 0) &&
				strncmp(buffers[i]+prev_eol, ".\r\n",
				MIN(4, next_eol+1)))
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
							{
				/*if first time for server, do
				status message, if first time period,
				do print out of status*/
				if (serv_array[i].misc_state) {
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				    if (*(buffers[i]+prev_eol-
					bufremains[i]) != CHAR_OK)
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				    if (*(buffers[i]+prev_eol) !=
					CHAR_OK)
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					{
#if DNNTP_LIST_MEMBUF_PAGEALIGN
						if (!strncasecmp(buffers[i]+
							prev_eol-bufremains[i],
							DNNTP_TIMEOUT_STR,
							DNNTP_TIMEOUT_STRLEN))
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						if (!strncasecmp(buffers[i]+
							prev_eol,
							DNNTP_TIMEOUT_STR,
							DNNTP_TIMEOUT_STRLEN))
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						{
						/*mark as gone*/
						FD_CLR(serv_array[i].connection,
							&readready);
						    CloseServer(i);
						    OpenServer(i);
						FD_SET(serv_array[i].connection,
							&readready);
						    if (serv_array[i].
							curgroup[0])
							SendGroup(i, NULL);
						    status = ReSendToServer(i);
						    bufremains[i] = 0;
/*copied from above SendToServer*/
					if ( status < 0) {
						CloseServer(i);
#if DNNTP_DISK_LIST_CACHE
					if (Group2ServerListCacheFetch(i) < 0) {
						Group2ServerListCacheKillAll();
#endif /*DNNTP_DISK_LIST_CACHE*/
						print_couldnt_send(i);
						free_stuff(&buffers,
							&bufremains);
						return(DNNTP_REMOTE_GONE_ERR);
#if DNNTP_DISK_LIST_CACHE
					}
/*stolen from below*/
						/*free stuff*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
					free(buffers[i]-DNNTP_LIST_MEMPAGESIZE);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					free(buffers[i]);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						/*mark as gone*/
					buffers[i] = NULL;
					FD_CLR(serv_array[i].connection,
						&readready);
					ofunc--;
						/*hope there isn't any more data
						read...(if so close!)*/
					if ((prev_eol+next_eol+1) !=
						(bufremains[i]+readret)) {
						CloseServer(i);
					}
					break;
#endif /*DNNTP_DISK_LIST_CACHE*/
					}
					serv_array[i].misc_state = 1;
					break;
					} else { /*error wasn't a timeout...*/
#if DNNTP_DISK_LIST_CACHE
						Group2ServerListCacheKillAll();
#endif /*DNNTP_DISK_LIST_CACHE*/
						print_couldnt_send(i);
						free_stuff(&buffers,
							&bufremains);
						return(DNNTP_REMOTE_GONE_ERR);
					}
				} /*!= CHAR_OK*/
				serv_array[i].misc_state = 0;
				} else { /*in the middle of list results*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				*(buffers[i]+prev_eol+next_eol-
					bufremains[i]) = '\0';
			        read_args = break_firstarg(buffers[i]+
					prev_eol-bufremains[i]);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				*(buffers[i]+prev_eol+next_eol) = '\0';
			        read_args = break_firstarg(buffers[i]+
					prev_eol);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				/*check if this should have
				come from this server (and toss
				if not*/
				AddGroup2Server(i, read_args[0],
					read_args[1]);
				free_arg(read_args);
			    }
			    prev_eol += next_eol + 1;
			}
			if ((!serv_array[i].misc_state) && (next_eol < 0)) {
				if (prev_eol < bufremains[i]+readret) {
					/*copy remainders into place*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				    bcopy(buffers[i]+prev_eol-
					bufremains[i], buffers[i]-
					((bufremains[i]+readret)-
					prev_eol), (bufremains[i]+
					readret)-prev_eol);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				    bcopy(buffers[i]+prev_eol,
					buffers[i], (bufremains[i]+
					readret)-prev_eol);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				}
					/*save new remaining data num*/
				bufremains[i] = (bufremains[i]+ readret)-
					prev_eol;
			} else if (!serv_array[i].misc_state) {
					/*free stuff*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				free(buffers[i]-DNNTP_LIST_MEMPAGESIZE);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				free(buffers[i]);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					/*mark as gone*/
				buffers[i] = NULL;
				FD_CLR(serv_array[i].connection,
					&readready);
				ofunc--;
					/*hope there isn't any more data
					read...(if so close!)*/
				if ((prev_eol+next_eol+1) !=
					(bufremains[i]+readret)) {
					CloseServer(i);
				}
			}
		    }
		    if (ofunc > 1) {
			/*create select mask*/
			FD_ZERO(&readready);
			for(i = 0; i < serv_array_len; i++)
			    if (buffers[i] &&
				(serv_array[i].connection >= 0))
				FD_SET(serv_array[i].connection,
					&readready);
		    }
		} while (ofunc > 0);
		free(buffers);
		free(bufremains);
		{
		struct timeval tv;
			if (gettimeofday(&tv, NULL) >=0)
				group_array_secs = tv.tv_sec;
		}
	}
#if DNNTP_DISK_LIST_CACHE
	}
#endif /*DNNTP_DISK_LIST_CACHE*/
	Group2ServerEndAdd();
	return (DNNTP_SUCCESS);
}

/*Group2ServerSort() -- sort Group2Server info (after done AddGroup2Server
	stuff...) (and record length of data that is sorted)
    parameters
	none
    return
	>=0 if succeds <0 if fails
*/
int Group2ServerSort() {
	if(group_array != NULL) {
		qsort(group_array, group_array_len, sizeof(dgroup_t),
			cmp_groups);
		group_array_sortedlen = group_array_len;
	} else
		return (DNNTP_SIMPLE_ERR);
	return (DNNTP_SUCCESS);
}

/*Group2ServerDupElim() -- eliminate any duplicate entries in the list
	database (which is sorted, so we can tell...)
    parameters
	none
    return
	>=0 if succeds <0 if fails
*/
int Group2ServerDupElim() {
int dst_start = -1, src_start, src_end;

if (group_array != NULL) {
	for (src_end = 0, src_start = 0; src_end < (group_array_sortedlen-1);) {
		if (strcmp(group_array[src_end].group,
			group_array[src_end+1].group) == 0) {
			if (dst_start > 0) {
				bcopy(group_array+src_start,
					group_array+dst_start,
					(src_end-src_start+1)*sizeof(dgroup_t));
				dst_start += (src_end-src_start+1);
			} else
				dst_start = src_end+1;
				/*clean up (and skip dups)*/
			src_start = src_end+1;
			while (((src_start < group_array_sortedlen) &&
				(strcmp(group_array[src_end].group,
				group_array[src_start].group) == 0))) {
				free(group_array[src_start].group);
				free(group_array[src_start].stuff);
				src_start++;
			}
			src_end = src_start;
		} else
			src_end++;
	}
	if ((dst_start > 0) && (src_start < group_array_sortedlen)) {
			bcopy(group_array+src_start, group_array+dst_start,
				(src_end-src_start+1)*sizeof(dgroup_t));
			group_array_len = group_array_sortedlen =
				dst_start+(src_end-src_start+1);
	} else if (dst_start > 0) {
		group_array_len = group_array_sortedlen = dst_start;
	}
} else
	return (DNNTP_SIMPLE_ERR);
return (DNNTP_SUCCESS);
}

/* AddGroup2Server -- add an entry to group to server lookup info (with lots
	of checks before really added)
	note: that info is now screwed up still someone
	sorts it...
    parameters
	server_no: 	server this group might be on
	group: 		name of group
	hi_art, lo_art, status:	fields of the same name from a "LIST" command
    return
	>=0 if succeds <0 if fails( note: failure is not a big deal...
	just means wasn't added)
*/
int AddGroup2Server(server_no, group, list_stuff)
int server_no;
char *group, *list_stuff;
{
int ret_val = DNNTP_SUCCESS, cur_item;
char *crlf;
#if DNNTP_USE_GROUP_INFO
int tmp;
	/* if needed check against file*/
	if (groupfile_array&&groupfile_array_len) {	/* ie was loaded in*/
		tmp = Group2ServerFileLookup(group);
		if ((tmp < 0) || (tmp != server_no))
			return (DNNTP_SIMPLE_ERR);
	}
#endif /*DNNTP_USE_GROUP_INFO*/
	cur_item = group_array_len;
	group_array_len++;
	if (group_array == NULL) {
		/*add it in*/
		group_array_memlen = DNNTP_GROUP_INIT_ALLOC;
		group_array = (dgroup_t *)malloc(sizeof(dgroup_t)*
			group_array_memlen);
	} else if (group_array_len > group_array_memlen) {
			group_array_memlen += DNNTP_GROUP_INC_ALLOC;
			group_array = (dgroup_t *)realloc(group_array,
				sizeof(dgroup_t)*group_array_memlen);
	}
	if (!group_array) {
#if SYSLOG
		syslog(LOG_CRIT, "AddGroup2Server: ran out of memory");
#endif /*SYSLOG*/
		printf("%d nntp server ran out of memory\r\n", ERR_GOODBYE);
		(void) fflush(stdout);
		QuitAllServers();
		exit(1);
	}
	group_array[cur_item].server_num = server_no;
	group_array[cur_item].group = strdup(group);
	if (!group_array[cur_item].group) {
#if SYSLOG
		syslog(LOG_CRIT, "AddGroup2Server: ran out of memory");
#endif /*SYSLOG*/
		printf("%d nntp server ran out of memory\r\n", ERR_GOODBYE);
		(void) fflush(stdout);
		QuitAllServers();
		exit(1);
	}
	group_array[cur_item].stuff = NULL;
do_list_stuff:
	if (group_array[cur_item].stuff)
		free(group_array[cur_item].stuff);
		/*kill 0's before upper bound*/
	while ((*list_stuff == '0') && (*(list_stuff+1) != ' '))
		list_stuff++;
	if (((crlf = strchr(list_stuff, '\r')) != NULL) ||
		((crlf = strchr(list_stuff, '\n')) != NULL))
		*crlf = '\0';
	else
		crlf = list_stuff+strlen(list_stuff);
	{
	char *space_loc, *zero_loc;
	space_loc = strchr(list_stuff, ' ');
	if (!space_loc) {
#if SYSLOG
		syslog(LOG_CRIT, "AddGroup2Server: syntax errors in active");
#endif /*SYSLOG*/
		printf("%d syntax errors in active file\r\n", ERR_GOODBYE);
		(void) fflush(stdout);
		QuitAllServers();
		exit(1);
	}
	zero_loc = space_loc+1;
	while ((*zero_loc == '0') && (*(zero_loc+1) != ' '))
		zero_loc++;
		/* this is same as strdup, except we know length*/
		/*ahhhh will break on memory overruns...*/
	group_array[cur_item].stuff = malloc(space_loc-list_stuff+1 +
			crlf-zero_loc+1);
	if (!group_array[cur_item].stuff) {
#if SYSLOG
		syslog(LOG_CRIT, "AddGroup2Server: ran out of memory");
#endif /*SYSLOG*/
		printf("%d nntp server ran out of memory\r\n", ERR_GOODBYE);
		(void) fflush(stdout);
		QuitAllServers();
		exit(1);
	}
	bcopy(list_stuff, group_array[cur_item].stuff,
		space_loc-list_stuff+1);
	bcopy(zero_loc, group_array[cur_item].stuff+
		(space_loc-list_stuff+1), crlf-zero_loc+1);
	}
#if DNNTP_DISK_LIST_CACHE
	if ((serv_array[server_no].connection >= 0) &&
		group_array_diskcache && group_array_diskcache[server_no]) {
		fputs( group_array[cur_item].group,
			group_array_diskcache[server_no]);
		putc(' ', group_array_diskcache[server_no]);
		fputs( group_array[cur_item].stuff,
			group_array_diskcache[server_no]);
		putc('\r', group_array_diskcache[server_no]);
		putc('\n', group_array_diskcache[server_no]);
	}
#endif /*DNNTP_DISK_LIST_CACHE*/
	return (ret_val);
}

/* Group2ServerDump -- dumps the contents of the Group2Server info to stdout
	in a format suitable for answering a LIST command (minus .\r\n)
	(also no fflush...)
    parameters
	file: a FILE * to dump to fflushed includes .\r\n
    return
	none
*/
void Group2ServerDump(file)
FILE *file;
{
int i;
	if (file == NULL) {
#if SYSLOG
		syslog(LOG_ERR, "Group2ServerDump: passed NULL file pointer");
#endif /*SYSLOG*/
		return;
	}
	for (i = 0; i < group_array_len; i++) {
		fputs(group_array[i].group, file);
		putc(' ', file);
		fputs(group_array[i].stuff, file);
		putc('\r', file);
		putc('\n', file);
		/*fprintf(file, "%s %s\r\n", group_array[i].group,
			group_array[i].stuff);*/
	}
	fputs(".\r\n", file);
	fflush(file);
}

/* Group2ServerPrint -- print active line for provided group
    parameters
	group - group to print out line for
    return
	>=0 if successful, <0 if group not found
*/
Group2ServerPrint(groupname, file)
char *groupname;
FILE *file;
{
dgroup_t *tmp;
	/*first make sure we are initialized*/
	if (group_array == NULL)
		return (DNNTP_INTERNAL_ERR);
	tmp = ansi_bsearch(groupname, group_array, group_array_sortedlen,
		sizeof(dgroup_t), cmp_group_n_str);
	if (tmp == NULL)
		return (DNNTP_SIMPLE_ERR);
	fputs(tmp->group, file);
	putc(' ', file);
	fputs(tmp->stuff, file);
	putc('\r', file);
	putc('\n', file);
	return (DNNTP_SUCCESS);
}

/* Group2ServerExpire -- kill LIST cache (and group2server info) if
	older than a certain time.  You better be prepared to refill it
	if you use this.
    parameters
	none
    return
	0 if didn't expire it, !=0 if we did
*/
int Group2ServerExpire()
{
struct timeval tv;
	if (gettimeofday(&tv, NULL) < 0) {
#if SYSLOG
		syslog(LOG_ERR, "Group2ServerExpire: gettimeofday failed: %m");
#endif
		printf("%d unable to get time to check LIST cache\r\n",
			ERR_GOODBYE);
		(void) fflush(stdout);
		QuitAllServers();
		exit(1);
	}
	if (((tv.tv_sec - group_array_secs) > DNNTP_LIST_CACHE_TIME) ||
		(group_array == NULL)) {
		if (group_array)
			Group2ServerClear();
		group_array_secs = tv.tv_sec;
		return(1);	/* we did an expire */
	}
	return(0);	/* we didn't do an expire */
}

char tmpfilestr[MAXPATHLEN+1] = "";
size_t tmpfile_strlen;
#ifndef CMUCS
static pid_t my_pid;
#else /*CMUCS*/
static int my_pid;
#endif /*CMUCS*/
/* Group2ServerStartAdd -- do any preperations for starting to add stuff to
	G2S database (currently just open files to do on disk caching)
	Note: you damn well better have an empty database when this is called
	(ie called Group2ServerClear or first time for everything)
    parameters
	none
    return
	0 if successful <0 if not
*/
int Group2ServerStartAdd() {
#if DNNTP_DISK_LIST_CACHE
int i;
	if (tmpfilestr[0] == '\0') {
	struct passwd *pw;
		pw = getpwuid(getuid());
		if (concat(tmpfilestr, sizeof(tmpfilestr), pw->pw_dir, "/",
			DNNTP_DISK_LIST_CACHE_NAME, ".", NULL) == NULL) {
#if SYSLOG
			syslog(LOG_ERR, "Group2ServerStartAdd: list cache file \
pathname too long");
#endif /*SYSLOG*/
			return (DNNTP_NO_SETUP_ERR);
		}
		tmpfile_strlen = strlen(tmpfilestr);
	}
	my_pid = getpid();
	if (!group_array_diskcache) {
		group_array_diskcache = (FILE **)malloc(sizeof(FILE *)*
			serv_array_len);
		if (!group_array_diskcache){
#if SYSLOG
			syslog(LOG_CRIT, "Group2ServerStartAdd: ran out of memory");
#endif /*SYSLOG*/
			printf("%d nntp server ran out of memory\r\n", ERR_GOODBYE);
			(void) fflush(stdout);
			QuitAllServers();
			exit(1);
		}
	}

	for (i = 0; i < serv_array_len; i++) {
		sprintf(tmpfilestr+tmpfile_strlen, "%x.%x", i, my_pid);
		group_array_diskcache[i] = fopen(tmpfilestr, "w");
		if (!group_array_diskcache[i]){	/*damn no disk cache*/
		int j;
#if LOG
#if SYSLOG
			if (i == 0)
				syslog(LOG_DEBUG, "Group2ServerStartAdd: \
unable to open new cachefile %s", tmpfilestr);
#endif /*LOG*/
#endif /*SYSLOG*/
			/*clean up and pretend it never happened*/
			for (j = 0 ; j < i; j++)
				fclose(group_array_diskcache[j]);
			free(group_array_diskcache);
			group_array_diskcache = NULL;
			return (DNNTP_NO_SETUP_ERR);
		}
	}
	tmpfilestr[tmpfile_strlen] = '\0';
#endif /*DNNTP_DISK_LIST_CACHE*/
	return (DNNTP_SUCCESS);
}

/* Group2ServerEndAdd -- finish up adding stuff to the G2S database
	(currently just close disk cache files, and install them if
	they are newer than what is there)
    parameters
	dump: pass a FILE* to dump to or NULL (to not dump)
    return
	0 if successful <0 if not
*/
int Group2ServerEndAdd()
{
#if DNNTP_DISK_LIST_CACHE
char realfile[MAXPATHLEN+1];
char tmpfilename[MAXPATHLEN+1];
int realfilefd;
struct stat realstat;
struct stat tmpstat;
int i;
#endif /*DNNTP_DISK_LIST_CACHE*/
	Group2ServerSort();
	Group2ServerDupElim();
		/* cut off excess...*/

	if (group_array_len < group_array_memlen) {
		group_array_memlen = group_array_len;
		group_array = (dgroup_t *)realloc(group_array,
			sizeof(dgroup_t)*group_array_memlen);
	}
#if DNNTP_DISK_LIST_CACHE
	if (group_array_diskcache) {
		my_pid = getpid();
		for (i = 0; i < serv_array_len; i++) {
			if (group_array_diskcache[i]) {
				fflush(group_array_diskcache[i]);
			strcpy(realfile, tmpfilestr);
			sprintf(realfile+tmpfile_strlen, "%x", i);
			strcpy(tmpfilename, realfile);
			sprintf(tmpfilename+strlen(tmpfilename), ".%x", my_pid);
			realfilefd = open(realfile, O_RDONLY);
			if (realfilefd >= 0) {
				flock(realfilefd, LOCK_EX);
				realstat.st_ctime = 1;
				fstat(realfilefd, &realstat);
			}
			if (group_array_diskcache[i]) {
				tmpstat.st_ctime = 0;
				fstat(fileno(group_array_diskcache[i]),
					&tmpstat);
			}
					/* is ours newer?*/
			if (group_array_diskcache[i] && (((realfilefd < 0) ||
				(tmpstat.st_ctime > realstat.st_ctime)) &&
				tmpstat.st_size)) {
				if (rename (tmpfilename, realfile) < 0)
					unlink(tmpfilename);
			} else {
				unlink(tmpfilename);
			}
			if (realfilefd >= 0) {
				flock(realfilefd, LOCK_UN);
				close(realfilefd);
			}
			if (group_array_diskcache[i]) {
				/* note we don't end up deleteing it here...*/
				fclose(group_array_diskcache[i]);
				group_array_diskcache[i] = NULL;
			}
			realfile[tmpfile_strlen] = '\0';
			} /* if diskcache[i]*/
		}
	}
#endif /*DNNTP_DISK_LIST_CACHE*/
	return(DNNTP_SUCCESS);
}

#if DNNTP_DISK_LIST_CACHE
/* Group2ServerListCacheKillAll -- kill off (temporary) cache files
	Note: things in an inconsistant state for Group2ServerEndAdd
	(ie exit now or call Group2ServerStartAdd again!)
    parameters
	none
    return
	0 if successful <0 if not
*/
int Group2ServerListCacheKillAll()
{
int i;
char killfile[MAXPATHLEN+1];
	if (group_array_diskcache) {
		strcpy(killfile, tmpfilestr);
		my_pid = getpid();
		for (i = 0; i < serv_array_len; i++) {
			if (group_array_diskcache[i]) {
				fclose(group_array_diskcache[i]);
				group_array_diskcache[i] = NULL;
				sprintf(killfile+tmpfile_strlen, "%x.%x", i, my_pid);
				unlink(killfile);
			}
		}
		free(group_array_diskcache);
		group_array_diskcache = NULL;

	}
	return (DNNTP_SUCCESS);
}

/* Group2ServerListCacheFetch -- fetch disk copy of an old list cache
	must be called after an Group2ServerExpire, or
	Group2ServerListCacheFresh
    parameters
	server_no:	number of server who needs to be loaded from disk
    return
	0 if successful <0 if not
*/
int Group2ServerListCacheFetch(server_no)
int server_no;
{
char cachefile[MAXPATHLEN+1];
int cachefd;
char **read_args;
long file_size;
char *buffer, *bufptr, *nextline;
#if DNNTP_USE_GROUP_INFO
int tmp_groupfile_array_len;
#endif /*DNNTP_USE_GROUP_INFO*/
	if (ValidServer(server_no)) {
		strcpy(cachefile, tmpfilestr);
		sprintf(cachefile+tmpfile_strlen, "%x", server_no);
		cachefd = open(cachefile, O_RDONLY);
		if (cachefd < 0)
			return(DNNTP_NO_SETUP_ERR);
#if DNNTP_USE_GROUP_INFO
			/*hiddious kludge to turn off groupfile lookup,
			since any rubish in the disc cache already went
			through this when it was writen*/
		tmp_groupfile_array_len = groupfile_array_len;
		groupfile_array_len = 0;
#endif /*DNNTP_USE_GROUP_INFO*/
		if (group_array_diskcache && group_array_diskcache[server_no]) {
			fclose(group_array_diskcache[server_no]);
			group_array_diskcache[server_no] = NULL;
		}
		my_pid = getpid();
		sprintf(cachefile+strlen(cachefile), ".%x",
			my_pid);
		unlink(cachefile);
		flock(cachefd, LOCK_SH);
		file_size = lseek(cachefd, 0, SEEK_END);
		if (file_size < 0) {
			flock(cachefd, LOCK_UN);
			close (cachefd);
			return(DNNTP_NO_SETUP_ERR);
		}
		lseek(cachefd, 0, L_SET);
		buffer = malloc(file_size+1);
		if (!buffer) {
			flock(cachefd, LOCK_UN);
			close (cachefd);
			return(DNNTP_NO_SETUP_ERR);
		}
		if (read(cachefd, buffer, file_size) < 0) {
			flock(cachefd, LOCK_UN);
			close (cachefd);
			free(buffer);
			return(DNNTP_NO_SETUP_ERR);
		}
		*(buffer+file_size) = '\0';
		bufptr = buffer;
		do {
			nextline = index(bufptr, '\n');
			if (nextline) {
				*nextline = '\0';
				read_args = break_firstarg(bufptr);
					/*trust result, we wrote it*/
				AddGroup2Server(server_no, read_args[0],
					read_args[1]);
				bufptr = nextline+1;
				free_arg(read_args);
			}
		} while (nextline);

		flock(cachefd, LOCK_UN);
		close(cachefd);
		free(buffer);
#if DNNTP_USE_GROUP_INFO
		groupfile_array_len = tmp_groupfile_array_len;
#endif /*DNNTP_USE_GROUP_INFO*/
	} else {
#if SYSLOG
		syslog(LOG_ERR, "Group2ServerListCacheFetch: illegal server number: %d",
			server_no);
#endif /*SYSLOG*/
		return(DNNTP_INTERNAL_ERR);
	}
	return (DNNTP_SUCCESS);
}

/* Group2ServerListCacheFresh -- is the Disk List Cache accessable and
	of an apropriate age?
    parameters
	none
    return
	>0 if "fresh" ==0 if needs to be re-fetched (or re-LISTed anyway)
*/
time_t Group2ServerListCacheFresh(how_old)
long how_old;
{
struct stat tmpstat;
int i;
struct timeval tv;
time_t min_time = ~0;

	if (tmpfilestr[0] == '\0') {
	struct passwd *pw;
		pw = getpwuid(getuid());
		if (concat(tmpfilestr, sizeof(tmpfilestr), pw->pw_dir, "/",
			DNNTP_DISK_LIST_CACHE_NAME, ".", NULL) == NULL) {
#if SYSLOG
			syslog(LOG_ERR, "Group2ServerListCacheFresh: list \
cache file pathname too long");
#endif /*SYSLOG*/
			return (DNNTP_NO_SETUP_ERR);
		}
		tmpfile_strlen = strlen(tmpfilestr);
	}
	my_pid = getpid();
	if (gettimeofday(&tv, NULL) < 0) {
#if SYSLOG
		syslog(LOG_ERR, "Group2ServerListCacheFresh: couldn't get \
the time of day %m");
#endif /*SYSLOG*/
		return (0);
	}
	for (i = 0; i < serv_array_len; i++) {
		sprintf(tmpfilestr+tmpfile_strlen, "%x", i);
		if (stat(tmpfilestr, &tmpstat)<0)
			goto done_false;
		min_time = MIN(min_time, tmpstat.st_ctime);
		if ((tv.tv_sec-tmpstat.st_ctime) > how_old )
			goto done_false;
		if (!(((tmpstat.st_uid == getuid())&&(tmpstat.st_mode&S_IREAD))||
		((tmpstat.st_gid == getgid())&&(tmpstat.st_mode&(S_IREAD>>3)))||
		(tmpstat.st_mode&(S_IREAD>>6))))
			goto done_false;
	}
	tmpfilestr[tmpfile_strlen] = '\0';
	if (min_time == ~0)
		min_time = 1;
	return (min_time);
done_false:
	tmpfilestr[tmpfile_strlen] = '\0';
	return (0);

}

#if NO_FLOCK
static int flock(fd, opr)
int fd, opr;
{
struct flock fl;
	switch (opr) {
		case LOCK_SH:
			fl.l_type = F_RDLCK;
		break;
		case LOCK_EX:
			fl.l_type = F_WRLCK;
		break;
		case LOCK_UN:
			fl.l_type = F_UNLCK;
		break;
		default:
#if SYSLOG
			syslog(LOG_ERR, "flock got unsupported operation");
#endif /*SYSLOG*/
			return (-1);
	}
	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 0;
	return(fcntl(fd, F_SETLKW, &fl));
}
#endif /*NO_FLOCK*/
#endif /*DNNTP_DISK_LIST_CACHE*/
#if !SYSTEM_STRDUP && !CMUCS
char *strdup(s)
char *s;
{
register size_t len;
register char *ret;
	if (s) {
		len = strlen(s);
		ret = malloc(len+1);
		if (ret)
			bcopy(s, ret, len+1);
		return(ret);
	}
	return (NULL);
}
#endif /*!SYSTEM_STRDUP && !CMUCS*/ 
