/* list_cmd.c */
/* code to actually execute (well distribute) list command*/
/*
 * HISTORY
 * $Log:	list_cmd.c,v $
 * Revision 2.22  95/01/09  22:43:55  wbardwel
 * 	minor cleanup
 * 	[95/01/09  22:42:47  wbardwel]
 * 
 * Revision 2.21  95/01/09  03:00:46  wbardwel
 * 	clean up caching...
 * 	[95/01/09  02:58:53  wbardwel]
 * 
 * Revision 2.20  95/01/09  00:46:14  wbardwel
 * 	delete active file stuff...(done by prep now)
 * 	[95/01/09  00:44:30  wbardwel]
 * 
 * Revision 2.19  95/01/01  23:20:58  wbardwel
 * 	don't zero cache time, when we fetch files, use files' time
 * 	[95/01/01  23:19:53  wbardwel]
 * 
 * Revision 2.18  94/12/15  00:45:23  wbardwel
 * 	add "LIST ACTIVE <grouppat>" support
 * 	[94/12/15  00:35:15  wbardwel]
 * 
 * Revision 2.17  94/08/15  17:14:39  wbardwel
 * 	include sys/select.h for AIX
 * 	[94/08/15  17:13:31  wbardwel]
 * 
 * Revision 2.16  94/07/30  01:49:47  wbardwel
 * 	add sys/time.h (for select) Linux wants it..
 * 	[94/07/30  01:43:17  wbardwel]
 * 
 * Revision 2.15  94/07/12  17:27:01  wbardwel
 * 	update copyright
 * 	[94/07/12  17:23:21  wbardwel]
 * 
 * Revision 2.14  94/06/14  18:02:08  wbardwel
 * 	fix double copy of buffer freeing...
 * 	[94/06/14  17:58:07  wbardwel]
 * 
 * Revision 2.13  94/06/03  17:07:34  wbardwel
 * 	fix malformed if (which broken list hang fixes)
 * 	[94/06/03  17:07:17  wbardwel]
 * 
 * Revision 2.12  94/06/03  16:20:17  wbardwel
 * 	fix hangs when only one server left
 * 	[94/06/03  16:19:38  wbardwel]
 * 
 * Revision 2.11  94/05/26  04:35:20  wbardwel
 * 	fix selects (should fix old dnntpd`s hanging around...)
 * 	[94/05/26  04:34:09  wbardwel]
 * 
 * Revision 2.10  94/05/05  04:19:35  wbardwel
 * 	make failures on LIST NEWSGROUPS and like much gentler
 * 	[94/05/05  04:19:00  wbardwel]
 * 
 * Revision 2.9  94/04/21  01:49:09  wbardwel
 * 	make dangerous LIST caching configureable
 * 	[94/04/21  01:47:10  wbardwel]
 * 
 * Revision 2.8  94/04/13  13:24:58  wbardwel
 * 	fix wee little bug in some server outage code...
 * 	[94/04/13  13:24:13  wbardwel]
 * 
 * Revision 2.7  94/03/22  17:43:20  wbardwel
 * 	lots of medeocre fixes for what to do on real server outages...
 * 	do proper clean up of closed sockets and recurse on errors...
 * 	[94/03/22  17:40:34  wbardwel]
 * 
 * Revision 2.6  93/12/06  03:18:21  wbardwel
 * 	lots of fixes for state mistakes during timeouts 
 * 	[93/12/06  03:17:21  wbardwel]
 * 
 * Revision 2.5  93/12/02  03:29:07  wbardwel
 * 	Nasty bug...where it wouldn't do list right
 * 	[93/12/02  03:26:17  wbardwel]
 * 
 * Revision 2.4  93/12/02  00:45:53  wbardwel
 * 	fix bugs that screwed up things when real servers went down.  add
 * 	code to deal with Timeouts right.  fix bug that occasionally caused
 * 	core dumps in find_eol
 * 	[93/12/02  00:35:31  wbardwel]
 * 
 * Revision 2.3  93/11/02  22:35:11  wbardwel
 * 	include fixes
 * 	[93/11/02  22:01:55  wbardwel]
 * 
 * Revision 2.2  93/10/27  16:02:52  wbardwel
 * 	Created.
 * 	[93/10/19            wbardwel]
 * 
*/
/*
 * Copyright (c) 1993, 1994, 1995 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

#include <errno.h>
#if DNNTP_USE_GROUP_INFO
#ifndef __STDC__
#include <varargs.h>	/*for concat*/
#else /*__STDC__*/
#include <stdarg.h>
#endif /*__STDC__*/
#endif /*DNNTP_USE_GROUP_INFO*/
#include <sys/time.h> /*for select*/
#ifdef AIX
#include <sys/select.h>
#endif /*AIX*/

extern int currentserver;
extern void print_couldnt_send();

/*internal routine for list_cmd to free buffers*/
/*also used by RealGroup2ServerPrep*/
void free_stuff(buffers, bufremains)
char ***buffers;
long **bufremains;
{
int i;
	if (*buffers) {
		for(i = 0; i < serv_array_len; i++)
			if (*buffers[i]) {
				free(*buffers[i]);
				*buffers[i] = NULL;
			}
		free(*buffers);
		*buffers = NULL;
	}
	if (*bufremains) {
		free(*bufremains);
		*bufremains = NULL;
	}
}

/* finds \r\n sequence in buf, stoping at len if that comes first*/
/* return index into buf, or -1 */
long find_eol(buf, length)
char *buf;
long length;
{
register char *t=buf, *len = buf+length-1;
	if (length)
		do {
			if((*t == '\r') && (*(t+1) == '\n'))
				return (t-buf+1);
		} while(++t<len);
	return (-1);
}


/* list_cmd -- do the LIST command
    parameters
	args: argv style broken up command
	no_args: number of arguments in args
    return
	>=0 on success <0 on failure
*/
int list_cmd( args, no_args)
char *args[];
int no_args;
{
char str[NNTP_STRLEN+1];
int status;
int first_time = 1;
int ofunc;
	/* dumb optimization...*/
int dist_cmd;
time_t file_time;
	dist_cmd = ((no_args == 2) && (!strcasecmp(args[1], "DISTRIBUTIONS")));
		/*valid list command*/
 	if ((no_args == 1) || (!strcasecmp(args[1], "ACTIVE"))) {
	int is_empty = !group_array;
		Group2ServerExpire();
        	if (!group_array) {
#if DNNTP_SERVER_CONSISTANT_WO_LISTS
                	RealGroup2ServerPrep(DNNTP_LIST_CACHE_TIME, 1);
#else /*!DNNTP_SERVER_CONSISTANT_WO_LISTS*/
			/*if was empty, loading is ok, if not run list before,
			(odd...if empty, list can't have been run...but someday
			we may want this feature) otherwise don't do loading*/
                	RealGroup2ServerPrep(DNNTP_LIST_CACHE_TIME,
				(is_empty?0:-1));
#endif /*!DNNTP_SERVER_CONSISTANT_WO_LISTS*/
        	}
		printf("%d\r\n", OK_GROUPS);
#if EXTENDED_LIST_ACTIVE
		if (no_args > 2)
			return(list3_cmd(args, no_args));
#endif /*EXTENDED_LIST_ACTIVE*/
		Group2ServerDump(stdout);
		return(DNNTP_SUCCESS);
	}

	if (dist_cmd || ((no_args == 2) &&
		 (!strcasecmp(args[1], "NEWSGROUPS")) )) {
			/*poll all servers and put results all together.*/
	int i;
		if (currentserver < 0)
			currentserver = i = 0;
		else
			i = currentserver;
		if (!dist_cmd) {
			Group2ServerPrep();
		}
		do {
			serv_array[i].misc_state = serv_array[i].connection;
			i = (i+1)%serv_array_len;
		} while (i != currentserver);
		for(ofunc = 0; ofunc < DNNTP_OPEN_STAGES; ofunc++) {
		do {
			if (serv_array[i].misc_state < 0) {
				if (open_serv_func[ofunc](i) < 0) {
					printf("%d couldn't open \
server %s\r\n", ERR_GOODBYE, ServerName(i));
					(void) fflush(stdout);
					QuitAllServers();
					exit (1);
				}
			}	/* if needed to be opened*/
			if ((ofunc == (DNNTP_OPEN_STAGES-1)) &&
				(serv_array[i].connection >= 0)) {
				if (serv_array[i].curgroup[0])
					SendGroup(i, NULL);
				status = SendToServer(i, "LIST %s",
					args[1]);
				if (status < 0) {
					CloseServer(i);
					print_couldnt_send(i);
					return(DNNTP_SIMPLE_ERR);
				}
			}
			if (dist_cmd)
				break;
			i = (i+1)%serv_array_len;
		} while ((i != currentserver));
		}	/* end of for loop on open stages*/
		i = currentserver;
		/* reusing ofunc just to be that way.....*/
		if (dist_cmd) {
			status = ReadFromServer(currentserver, str);
			fputs(str, stdout);
			if (status == OK_GROUPS) {
				while (ContReadFromServer(i, str) > 0) {
					fputs(str, stdout);
				}
			}
		} else {
			/* note, timeouts here break! */
		char **buffers = NULL, **read_args;
		long *bufremains = NULL;
		fd_set readready;
		int i;	/*misc index*/
		int maxfd;
		long readret, next_eol, prev_eol;

			/*create buffers for servers*/
			buffers = (char **)calloc(serv_array_len,
				sizeof(char *));
			bufremains = (long *)calloc(serv_array_len,
				sizeof(long));
			if (!buffers || !bufremains) {
#if SYSLOG
				syslog(LOG_CRIT, "list_cmd: ran out of memory");
#endif /*SYSLOG*/
				printf("%d server out of memory\r\n",
					ERR_GOODBYE);
				QuitAllServers();
				exit(1);
			}
			FD_ZERO(&readready);
			for (i=0, maxfd = 0, ofunc = 0; i < serv_array_len;
				i++) {
			    if (serv_array[i].connection >= 0) {
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				buffers[i] = malloc(DNNTP_LIST_MEMBUFSIZE+
					DNNTP_LIST_MEMPAGESIZE);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				buffers[i] = malloc(DNNTP_LIST_MEMBUFSIZE);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				if (!buffers[i]) {
#if SYSLOG
				    syslog(LOG_CRIT, "list_cmd: ran out of \
memory");
#endif /*SYSLOG*/
				    printf("%d server out of memory\r\n",
					ERR_GOODBYE);
				    QuitAllServers();
				    exit(1);
				}
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				buffers[i] += DNNTP_LIST_MEMPAGESIZE;
#endif /*DNNTP_LIST_MEMBUF_PAGEALIGN*/
				maxfd = MAX(serv_array[i].connection, maxfd);
				FD_SET(serv_array[i].connection, &readready);
				ofunc++;
				serv_array[i].misc_state = 1;
			    } else {
				serv_array[i].misc_state = 0;
			    }

			}
				
			/*loop*/
			do {
			    if (ofunc > 1) { /*if more than one fd left*/
					/*select for reads, no timeout*/
					/* doesn't catch other errors...*/
				while (((readret = select(maxfd+1, &readready,
					NULL, NULL, NULL)) < 0) &&
					(errno == EINTR))
					errno = 0;
				if (readret < 0) {
#if SYSLOG
					syslog(LOG_CRIT, "list_cmd: select %m");
#endif /*SYSLOG*/
					printf("%d server internal error\r\n",
						ERR_GOODBYE);

					QuitAllServers();
					exit(1);
				}
			    }
				/*for all servers' fds' which have mask bit set*/
				for (i = 0; i < serv_array_len; i++) {
				    if (ofunc > 1) {
					if ((buffers[i] == NULL) ||
					    !FD_ISSET(serv_array[i].connection,
					    &readready))
					    continue;
				    } else {
					if ((buffers[i] == NULL) ||
					    (serv_array[i].connection < 0))
					    continue;
				    }

					/*read a buffers(-prev amount) worth*/
					/* doesn't catch other errors...*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				    while (((readret = read(serv_array[i].
					connection,  buffers[i],
					DNNTP_LIST_MEMBUFSIZE))
					< 0) && (errno == EINTR))
						errno = 0;
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				    while (((readret = read(serv_array[i].
					connection,  buffers[i]+bufremains[i],
					DNNTP_LIST_MEMBUFSIZE-bufremains[i]))
					< 0) && (errno == EINTR))
						errno = 0;
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					/*do some real error handling here...*/
				    if (readret < 0) {
#if SYSLOG
					syslog(LOG_ERR, "list_cmd: error \
reading from server '%s' %m", ServerName(i));
#endif /*SYSLOG*/
					if (serv_array[i].misc_state) {
						print_couldnt_send(i);
						QuitAllServers();
						exit(1);
					} else { /*partway into reply*/
						QuitAllServers();
						exit(1);
					}
				    }
				    prev_eol = 0;
					/*while find eol(buf), and !=.\r\n*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
				    while ((( next_eol = find_eol(buffers[i]+
					prev_eol-bufremains[i], (bufremains[i]
					+ readret)- prev_eol)) >= 0) &&
					((next_eol != 2) || strncmp(buffers[i]+
					prev_eol-bufremains[i], ".\r\n", MIN(4,
					next_eol+1))))
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
				    while ((( next_eol = find_eol(buffers[i]+
					prev_eol, (bufremains[i] + readret)-
					prev_eol)) >= 0) &&
					((next_eol != 2) || strncmp(buffers[i]+
					prev_eol, ".\r\n",
					MIN(4, next_eol+1))))
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
								{
					/*if first time for server, do
					status message, if first time period,
					do print out of status*/
					if (serv_array[i].misc_state) {
#if DNNTP_LIST_MEMBUF_PAGEALIGN
					    if (((*(buffers[i]+prev_eol-
						bufremains[i]) == CHAR_OK) ||
						(ofunc == 1)) && first_time) {
						fwrite(buffers[i]+prev_eol-
						    bufremains[i], sizeof(char),
 						    next_eol-prev_eol+1,
						    stdout);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					    if (((*(buffers[i]+prev_eol) ==
						CHAR_OK) || (ofunc == 1)) &&
						first_time) {
						fwrite(buffers[i]+prev_eol,
						    sizeof(char), next_eol-
						    prev_eol+1, stdout);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						first_time = 0;
#if DNNTP_LIST_MEMBUF_PAGEALIGN
					    } else if (*(buffers[i]+prev_eol-
						bufremains[i]) != CHAR_OK)
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					    } else if (*(buffers[i]+prev_eol) !=
						CHAR_OK)
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					    {
#if DNNTP_LIST_MEMBUF_PAGEALIGN
						if (!strncasecmp(buffers[i]+
							prev_eol-bufremains[i],
							DNNTP_TIMEOUT_STR,
							DNNTP_TIMEOUT_STRLEN))
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						if (!strncasecmp(buffers[i]+
							prev_eol,
							DNNTP_TIMEOUT_STR,
							DNNTP_TIMEOUT_STRLEN))
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						{
						/*mark as gone*/
						FD_CLR(serv_array[i].connection,
							&readready);
						    CloseServer(i);
						    OpenServer(i);
						FD_SET(serv_array[i].connection,
							&readready);
						    if (serv_array[i].
							curgroup[0])
							SendGroup(i, NULL);
						    status = ReSendToServer(i);
						    bufremains[i] = 0;
/*copied from above SendToServer*/
				if (status < 0) {
					CloseServer(i);
					print_couldnt_send(i);
					QuitAllServers();
					exit(1);
				}
				serv_array[i].misc_state = 1;
				break;
/*end steal*/

					} else { /*error wasn't a timeout...*/
						print_couldnt_send(i);
						free_stuff(&buffers,
							&bufremains);
						return(DNNTP_REMOTE_GONE_ERR);
					}
					} /*!= CHAR_OK*/
					serv_array[i].misc_state = 0;
					} else { /*in the middle of list results*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
					*(buffers[i]+prev_eol+next_eol-
						bufremains[i]) = '\0';
				        read_args = break_firstarg(buffers[i]+
						prev_eol-bufremains[i]);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					*(buffers[i]+prev_eol+next_eol) = '\0';
				        read_args = break_firstarg(buffers[i]+
						prev_eol);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					/*check if this should have
					come from this server (and toss
					if not*/
					if (Group2Server( read_args[0]) == i)
					    printf("%s %s\n", read_args[0],
						read_args[1]);
					free_arg(read_args);
				    }
				    prev_eol += next_eol + 1;
				}
				if ((!serv_array[i].misc_state) &&
					(next_eol < 0)) {
					if (prev_eol < bufremains[i]+readret) {
						/*copy remainders into place*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
					    bcopy(buffers[i]+prev_eol-
						bufremains[i], buffers[i]-
						((bufremains[i]+readret)-
						prev_eol),
						(bufremains[i]+readret)
						-prev_eol);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					    bcopy(buffers[i]+prev_eol,
						buffers[i], (bufremains[i]+
						readret)-prev_eol);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					}
						/*save new remaining data num*/
					bufremains[i] = (bufremains[i]+
						readret)-prev_eol;
				} else if (!serv_array[i].misc_state) {
						/*free stuff*/
#if DNNTP_LIST_MEMBUF_PAGEALIGN
					free(buffers[i]-DNNTP_LIST_MEMPAGESIZE);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
					free(buffers[i]);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						/*mark as gone*/
					buffers[i] = NULL;
					FD_CLR(serv_array[i].connection,
						&readready);
					ofunc--;
						/*hope there isn't any more data
						read...(if so close!)*/
					if ((prev_eol+next_eol+1) !=
						(bufremains[i]+readret)) {
						CloseServer(i);
						ofunc--;
#if DNNTP_LIST_MEMBUF_PAGEALIGN
						free(buffers[i]-DNNTP_LIST_MEMPAGESIZE);
#else /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						free(buffers[i]);
#endif /*!DNNTP_LIST_MEMBUF_PAGEALIGN*/
						/*mark as gone*/
						buffers[i] = NULL;
					}
				}
			    }
			    if (ofunc > 1) {
				/*create select mask*/
				FD_ZERO(&readready);
				for(i = 0; i < serv_array_len; i++)
				    if (buffers[i] &&
					(serv_array[i].connection >= 0)) {
					FD_SET(serv_array[i].connection,
						&readready);
					}
			    }
			} while (ofunc > 0) ;
			free(buffers);
			free(bufremains);
		}
	} else {
			/*or garbage, send error message*/
#if EXTENDED_LIST_ACTIVE
		printf("%d Usage: LIST [ACTIVE [newsgroup]|NEWSGROUPS|DISTRIBUTIONS].\r\n",
			ERR_CMDSYN);
#else /*!EXTENDED_LIST_ACTIVE*/
		printf("%d Usage: LIST [ACTIVE [newsgroup]|NEWSGROUPS|DISTRIBUTIONS].\r\n",
			ERR_CMDSYN);
#endif /*!EXTENDED_LIST_ACTIVE*/
		(void) fflush(stdout);
		return(DNNTP_SIMPLE_ERR);
	}
	fputs(".\r\n", stdout);
	(void) fflush(stdout);
	return(DNNTP_SUCCESS);
}
