/* parse.c */
/* code to break up commands (and other parsing functions */
/* Sun Jan 31 16:02:31 EST 1993 wbardwel : created*/
/*
 * HISTORY
 * $Log:	parse.c,v $
 * Revision 2.6  93/11/02  22:35:26  wbardwel
 * 	include fixes
 * 	[93/11/02  22:02:49  wbardwel]
 * 
 * Revision 2.5  93/10/27  16:03:03  wbardwel
 * 	make break_* remeber old bit of malloced memory, and only make new memory,
 * 	if this hasn't been "free'ed".  make free_arg to also do this.
 * 	[93/10/27  15:27:20  wbardwel]
 * 
 * Revision 2.4  93/08/17  17:37:25  wbardwel
 * 	fix break_args, and make free_arg a #define...
 * 	[93/08/17  17:33:59  wbardwel]
 * 
 * Revision 2.3  93/08/04  16:38:21  wbardwel
 * 	make break_firstarg, and redo break_args so everything goes to one
 * 	routine (to implement both) (see dnntp.h for some of changes)
 * 	[93/08/04  00:13:46  wbardwel]
 * 
 * Revision 2.2  93/06/02  14:52:40  wbardwel
 * 	Created.
 * 	[93/01/31            wbardwel]
 * 
*/
/*
 * Copyright (c) 1993 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

static char **ret=NULL;
static int num_mem_args = 0;
/* break_args -- break command up based on spaces
	note: currently this will accept commands that have bogus
	white space in them... (ie things that aren't legit by RFC977
	note: returned item points into string passed in and as such
	trashes it, and will be messed up by reuse of that argument
    parameters
	command: a line (direct from client connection) with a command in it
	no_args: pointer to integer where number of args in return will be
		depositied
    return
	an argv style item (with command broken up) */
/* see dnntp.h for actual code (macro)*/
#if 0
char ** break_args(command, no_args)
char *command;	/* command line */
int *no_args;
{
	return (break_args_int(command, no_args, 0));
}
#endif /*0*/
/* break_args_int -- break command up based on spaces(see break_args for
	caveats when using)(this is the internal version)
    parameters
	command: a line (direct from client connection) with a command in it
	no_args: pointer to integer where number of args in return will be
		depositied
	one_arg_only: if 0 do what break_args wants if 1 do what break_firstarg
		wants
    return
	an argv style item (with command broken up) */
char ** break_args_int(command, no_args, one_arg_only)
char *command;	/* command line */
int *no_args, one_arg_only;
{
int num_args = 0;
char *begin_of_word;
		/*skip initial whitespace*/
	if (num_mem_args < 0) {
		num_mem_args = - num_mem_args;
	} else {
		ret = NULL;
		num_mem_args = 0;
	}
	do {
		while (isspace(*command))
			command++;
		if (*command) {
			begin_of_word = command;
			while ((*command != '\0') && !isspace(*command))
				command++;
				/*if end of command(we dump rest...(you better \
				make sure nothing valuable after this (like \
				use fgets)*/
			if (command != begin_of_word) {
				if (*command){
					if (!one_arg_only || num_args == 0)
						*command = '\0';
					command++;
				}
				num_args++;
				if (num_args > num_mem_args) {
					if (!one_arg_only)
						num_mem_args +=
						DNNTP_PARSE_INIT_ALLOC;
					else
						num_mem_args = 2;
					if (!ret) {
						ret = (char **)malloc(
							num_mem_args*
							sizeof(char *));
					} else {
						ret = (char **)realloc(ret,
							num_mem_args*
							sizeof(char *));
					}
				}
				if (!ret) {
#if SYSLOG
					syslog(LOG_ERR, "break_args: ran \
out of memory");
#endif /*SYSLOG*/
					fprintf(stdout, "%d nntp server ran \
out of memeory\r\n", ERR_GOODBYE);
					(void) fflush(stdout);
					exit(1);
				}
				ret[num_args-1] = begin_of_word;
			}
		}
	} while (*command && (!one_arg_only || num_args == 1));
	if (no_args)
		(*no_args) = num_args;
	if (num_mem_args > num_args) {
		ret = (char **)realloc(ret, num_args*sizeof(char *));
		num_mem_args = num_args;
	}
	return(ret);
}

/* free_arg -- free argv style item (from break_args)
    parameters
	arg: pointer to a pointer to put an argv style item
    return
	none */
void free_arg(arg)
char **arg;
{
	if (arg) {
		if(arg != ret)
			free(arg);
		else
			num_mem_args = -num_mem_args;
	}
}

#if 0
/* break_firstarg -- break command up based on spaces, but just break off
	first and second argument.
	note: currently this will accept commands that have bogus
	white space in them... (ie things that aren't legit by RFC977
	note: returned item points into string passed in and as such
	trashes it, and will be messed up by reuse of that argument
    parameters
	command: a line (direct from client connection) with a command in it
    return
	argv style list with just first two args in it.
*/
/* look in dnntp.h for real code (macro)*/
char ** break_firstarg(command)
char *command;	/* command line */
{
	return (break_args_int(command, NULL, 1));
}
#endif /*0*/
