/* post.c */
/* code to figure out who to send a posted article to (which servers)
   expects post to be waiting on stdin(as in after an IHAVE or POST
   which we have said yes too.*/
/* May 05 1993 -wbardwel created*/
/*
 * HISTORY
 * $Log:	post.c,v $
 * Revision 2.9  94/07/12  17:27:24  wbardwel
 * 	update copyright
 * 	[94/07/12  17:24:16  wbardwel]
 * 
 * Revision 2.8  94/05/17  04:37:15  wbardwel
 * 	re-write Newsgroups: parsing, (also if no lines returned, error
 * 	message already done) much cleaner, but does it work?
 * 	[94/05/17  04:35:27  wbardwel]
 * 
 * Revision 2.7  94/03/09  23:31:57  wbardwel
 * 	posting fixes of unknown value
 * 	[94/03/09  23:31:34  wbardwel]
 * 
 * Revision 2.6  94/01/06  04:39:49  wbardwel
 * 	allow any length lines in posts.
 * 	[94/01/06  04:35:07  wbardwel]
 * 
 * Revision 2.5  93/11/02  22:35:33  wbardwel
 * 	include fixes
 * 	[93/11/02  22:03:06  wbardwel]
 * 
 * Revision 2.4  93/08/13  16:36:38  wbardwel
 * 	silly hacks to accept \n in addition to \r\n to accomidate GNUS
 * 	[93/08/13  16:29:33  wbardwel]
 * 
 * Revision 2.3  93/07/21  22:41:44  wbardwel
 * 	bug fix (for OSF)
 * 	[93/07/21  22:36:39  wbardwel]
 * 
 * Revision 2.2  93/06/02  14:52:44  wbardwel
 * 	changed code for newsgroups line, so it reoders line to have
 * 	valid group at from, and returns server for that group (and thus
 * 	the article)
 * 	[93/05/21            wbardwel]
 * 
 * 	post was crashing if first line was the newsgroups line due
 * 	to while() accessing art_curline-1 (where art_curline == 0)
 * 	fixed klugy way
 * 	[93/05/18            wbardwel]
 * 
 * 	Created.
 * 	[93/05/05            wbardwel]
 * 
 *
*/
/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

/* check if malloc/realloc worked, if not printf stuff and quit*/
static void do_post_check_malloc();

/* do_post -- read a post(from client aka stdin) and send to all the
	right servers.
    parameters
	ret_servs: the array of newsgroups for the article for each server
	ret_nolines: number of lines in the array with the article in it.
	ret_lines: the array with the article in it.
    return
	>=0 on success <0 on failure
*/
int do_post( ret_nolines, ret_lines)
int *ret_nolines;
char ***ret_lines;
{
char **art;	/* where we will be holding the article */
int art_mem_size=0;	/*current memory for article array (*art)*/
int art_curline = 0;	/*what line we are on*/
size_t curline_len;
int curline_pos;
char *colon_loc;
int code = DNNTP_SUCCESS;
int dont_check = 0;	/*start out in checking headers mode*/
int serv_no=DNNTP_SIMPLE_ERR;
char *start_ngrps, *cur_ngrp, *end_ngrp, tmp_ch;

	Group2ServerPrep();
	do {
			/* allocate a new block of lines*/
		if (art_curline >= art_mem_size) {
			if (art_mem_size == 0) {
				art = (char **)malloc(DO_POST_INC_LINES*
					sizeof(char *));
				art_mem_size = DO_POST_INC_LINES;
			} else {
				art_mem_size += DO_POST_INC_LINES;
				art = (char **)realloc(art, art_mem_size*
					sizeof(char *));
			}
			do_post_check_malloc(art);
			bzero(art+art_curline,
 				(art_mem_size-art_curline)*sizeof(char *));
		}
		curline_len = NNTP_STRLEN+1;
		curline_pos = 0;
		art[art_curline] = (char *)malloc(NNTP_STRLEN+1);
		do_post_check_malloc(art[art_curline]);
		do {
			/*hum....what do we do if fgets fails?*/
		    if (fgets(art[art_curline]+curline_pos,
			curline_len-curline_pos, stdin) == NULL) {
#if LOG
#if SYSLOG
			syslog(LOG_INFO, "do_post: got EOF from client?");
#endif /*SYSLOG*/
#endif /*LOG*/
			printf("%d huh? I got an EOF from you\r\n",
				ERR_GOODBYE);
			(void) fflush(stdout);
			QuitAllServers();
			exit(1);
		    }
		    curline_pos += strlen(art[art_curline]+curline_pos);
		    if (*(art[art_curline]+curline_pos-1) != '\n') {
			curline_len += NNTP_STRLEN;
			art[art_curline]  = realloc(art[art_curline],
				curline_len);
			do_post_check_malloc(art[art_curline]);
		    } else {
			art[art_curline] = (char *)realloc(art[art_curline],
				curline_pos+1);
			do_post_check_malloc(art[art_curline]);
			curline_pos = -1;
		    }
		} while(curline_pos >= 0);

		if (dont_check) {
			art_curline++;
			continue;
		}
		if ((!strcmp(art[art_curline], "\r\n"))||
			(!strcmp(art[art_curline], "\n"))) {
			dont_check = 1;
			art_curline++;
			continue;
		} else {
		long tmp_str_len = strlen(art[art_curline])+1;
			colon_loc = index(art[art_curline], ':');
				/*note: don't do spaces in front of headers*/
			if (colon_loc && (!strncasecmp(art[art_curline],
				DO_POST_NEWSGROUP_HDR,
				(colon_loc-art[art_curline]+1)))) {
				start_ngrps = colon_loc+2;
				while (*start_ngrps == ' ')
					start_ngrps++;
				cur_ngrp = start_ngrps;
			do {
				end_ngrp = index(cur_ngrp, ',');
				if (!end_ngrp) {
					end_ngrp = index(cur_ngrp, '\r');
					if (!end_ngrp)
						end_ngrp = index(cur_ngrp,
							'\n');
				}
				if (!end_ngrp) {
					/*bogus stuff near end of line*/
#if LOG
#if SYSLOG
					syslog(LOG_INFO, "do_post: got \
indecipherable end of Newsgroups: line");
#endif /*SYSLOG*/
#endif /*LOG*/
					printf("%d indecipherable end of \
Newsgroups: line\r\n", ERR_POSTFAIL);
					(void) fflush(stdout);
					code = DNNTP_SIMPLE_ERR;
					goto quit;
				}
				tmp_ch = *end_ngrp;
				*end_ngrp = '\0';
				serv_no = Group2Server(cur_ngrp);
				if (serv_no < 0) {
					if ((tmp_ch == '\r') ||
						(tmp_ch == '\n')) {
						/* no good groups*/
#if LOG
#if SYSLOG
						syslog(LOG_INFO, "do_post: no \
good groups");
#endif /*SYSLOG*/
#endif /*LOG*/
						printf("%d no legitimate \
newsgroups\r\n", ERR_POSTFAIL);
						(void) fflush(stdout);
						code = DNNTP_SIMPLE_ERR;
						goto quit;
					}
					cur_ngrp = end_ngrp+1;
				}
				if ((tmp_ch != '\r') && (tmp_ch != '\n'))
					*end_ngrp = tmp_ch;
			} while ((serv_no < 0) && (tmp_ch != '\r') &&
				(tmp_ch != '\n'));
			if (serv_no >= 0) {
				/* +1 in case they didn't have a \r and \n*/
			char *tmp_str = malloc(tmp_str_len+1);
				do_post_check_malloc(art);
				end_ngrp = index(cur_ngrp, '\r');
				if (!end_ngrp)
					end_ngrp = index(cur_ngrp, '\n');
				if (end_ngrp)
					*end_ngrp = '\0';
				strcpy(tmp_str, "Newsgroups: ");
				if (cur_ngrp != start_ngrps) {
					strcat(tmp_str, cur_ngrp);
					strcat(tmp_str, ",");
					*(cur_ngrp-1) = '\0';
					strcat(tmp_str, start_ngrps);
					strcat(tmp_str, "\r\n");
				} else { 
					strcat(tmp_str, start_ngrps);
					strcat(tmp_str, "\r\n");
				}
				free(art[art_curline]);
				art[art_curline] = tmp_str;
				dont_check = 1;
			} else {
				/*internal error, this was supposed to be
				handled earlier*/
#if LOG
#if SYSLOG
				syslog(LOG_INFO, "do_post: no \
good groups???(possible internal error)");
#endif /*SYSLOG*/
#endif /*LOG*/
				printf("%d no legitimate newsgroups???\r\n",
					ERR_POSTFAIL);
				(void) fflush(stdout);
				code = DNNTP_INTERNAL_ERR;
				goto quit;
			}
			} /*if Newsgroups: header*/
		}
		art_curline++;
	} while (strcmp(art[art_curline-1], ".\r\n") &&
		strcmp(art[art_curline-1], ".\n"));

		/*got whole article, return everything*/
	if (art_curline < art_mem_size) {
		art = (char **)realloc(art, art_curline*sizeof(char *));
		do_post_check_malloc(art);
	}
	(*ret_nolines) = art_curline;	/*note returning number of lines*/
	(*ret_lines) = art;
	if ((serv_no < 0) && (art_curline <= 0))
		art_curline = 1;
	return (serv_no);
quit:
		/*if error, clear out what needs it...*/
	{
	char tmp_buf[NNTP_STRLEN+1];
	int i;
		if (strcmp(art[art_curline], ".\r\n") &&
			strcmp(art[art_curline], ".\n"))
			while (fgets(tmp_buf, sizeof(tmp_buf), stdin) &&
				strcmp(tmp_buf, ".\r\n") &&
				strcmp(tmp_buf, ".\n"));
		for (i=0; (i < art_mem_size); i++)
			if (art[i])
				free(art[i]);
	}
	free(art);
	(*ret_nolines) = 0;
	return (code);	/*up to caller to do error message*/
}

/* do_post_free_ret -- free returned stuff from do_post
    parameters
	servs: the array of boolean (ints) which had servers in it
	nolines: number of lines in the array with the article in it.
	lines: the array with the article in it.
    return
	>=0 on success (no such thing as failure...)
*/
int do_post_free_ret(nolines, lines)
int nolines;
char **lines;
{
int i;
	if (lines) {
		for(i = 0; i < nolines; i++)
			if (lines[i])
				free(lines[i]);
		free(lines);
	}
	return (DNNTP_SUCCESS);
}

	/* do check for failed allocation....*/
	/*I just got sick of this damn code being everywhere....*/
static void do_post_check_malloc(ptr)
void *ptr;
{
	if (!ptr) {
#if SYSLOG
		syslog(LOG_ERR, "do_post: ran out of memory");
#endif /*SYSLOG*/
		printf("%d dnntp server ran out of memory\r\n", ERR_GOODBYE);
		(void) fflush(stdout);
		QuitAllServers();
		exit(1);
	}
}
