/* serve.c */
/* code to do main serving loop (after connection)*/
/* Sat Jan 30 19:38:43 EST 1993 wbardwel : created*/
/*note: small pieces stolen from nntp/server/serve.c*/
/*
 * HISTORY
 * $Log:	serve.c,v $
 * Revision 2.15  95/02/21  16:40:24  wbardwel
 * 	IDLE_CLOSE stuff from noyd@u.washington.edu
 * 	[95/02/21  16:36:11  wbardwel]
 * 
 * Revision 2.14  95/02/08  13:08:55  wbardwel
 * 	[95/02/08  13:02:59  wbardwel]
 * 
 * Revision 2.13  94/12/15  00:45:40  wbardwel
 * 	fix typo in intro message
 * 	[94/12/15  00:35:48  wbardwel]
 * 
 * Revision 2.12  94/10/25  18:17:16  wbardwel
 * 	fix typo around bogus command syslog message (so it can be turned off
 * 	in the normal way
 * 	[94/10/25  18:16:16  wbardwel]
 * 
 * Revision 2.11  94/07/12  17:27:29  wbardwel
 * 	update copyright, only drop SIGPIPE, if there is a SIGPIPE
 * 	[94/07/12  17:24:38  wbardwel]
 * 
 * Revision 2.10  94/02/09  21:12:07  wbardwel
 * 	fix type casting on ctime call
 * 	[94/02/09  21:10:42  wbardwel]
 * 
 * Revision 2.9  94/01/06  04:39:59  wbardwel
 * 	fix bug where a blank line crashed.
 * 	[94/01/06  04:35:46  wbardwel]
 * 
 * Revision 2.8  93/11/02  22:35:41  wbardwel
 * 	misc fixes
 * 	[93/11/02  22:03:49  wbardwel]
 * 
 * Revision 2.7  93/10/27  16:03:07  wbardwel
 * 	fix a comment.
 * 	[93/10/27  15:27:46  wbardwel]
 * 
 * Revision 2.6  93/08/13  16:36:49  wbardwel
 * 	convert some syslog(LOG_INFO to syslog(LOG_DEBUG as they ought to be
 * 	[93/08/13  16:30:10  wbardwel]
 * 
 * Revision 2.5  93/08/05  00:26:25  wbardwel
 * 	down grade strcasecmp to a STRTOUPPER and strcmp
 * 	[93/08/05  00:24:58  wbardwel]
 * 
 * Revision 2.4  93/08/04  16:38:47  wbardwel
 * 	use servername instead of looking it up again
 * 	[93/08/04  16:28:17  wbardwel]
 * 
 * Revision 2.3  93/07/16  16:21:38  wbardwel
 * 	add MODE, DATE, and LISTGROUP commands, use bsearch to identify commands.
 * 	[93/07/16  16:12:44  wbardwel]
 * 
 * Revision 2.2  93/06/02  14:52:51  wbardwel
 * 	removed command_no code
 * 	[93/05/12            wbardwel]
 * 
 * 	Created.
 * 	[93/01/30            wbardwel]
 * 
*/
/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#ifdef AIX
#include <sys/select.h>
#endif /*AIX*/

extern gethostname();

static int cmd_cmp();

/* done in dnntp.h */
/*struct cmdent_t {
	char	*cmd_name;
	int	(*cmd_fn)();
}*/
struct cmdent_t cmdtable[] = {	/*should be ordered by usage % */
	{ "ARTICLE",	art_cmd },
#if AUTH
	{ "AUTHCAP",	auth_cmd },
	{ "AUTHINFO",	auth_cmd },
	{ "AUTHSYS",	auth_cmd },
#endif /*AUTH*/
	{ "BODY",	art_cmd },
	{ "DATE",	date_cmd },
	{ "GROUP",	group_cmd },
	{ "HEAD",	art_cmd },
	{ "HELP",	help_cmd },
	{ "IHAVE",	ihave_cmd },
	{ "LAST",	nextlast_cmd },
	{ "LIST",	list_cmd },
#if LISTGROUP
	{ "LISTGROUP",	listgroup_cmd },
#endif /*LISTGROUP*/
#if MODE
	{ "MODE",	mode_cmd },
#endif /*MODE*/
	{ "NEWGROUPS",	newgroups_cmd },
	{ "NEWNEWS",	newnews_cmd },
	{ "NEXT",	nextlast_cmd },
	{ "POST",	post_cmd },
	{ "SLAVE",	slave_cmd },
	{ "STAT",	art_cmd }
#if XHDR
	,{ "XHDR",	xhdr_cmd }
#endif /*XHDR*/
#if XOVER
	,{ "XOVER",	xover_cmd }
#endif /*XOVER*/
#if XTHREAD
	,{ "XTHREAD",	xthread_cmd }
#endif /*XTHREAD*/
};

#define QUIT_STR "QUIT"
/* server_loop -- main loop to serve a client connection
    parameters
	none
    return
	none */
void server_loop()
{
char command_line[NNTP_STRLEN+1];	/* array we read commands into*/
struct timeval connecttime;
extern char servername[];		/* filled in by remote_auth*/
		/*we will detect broken connections with read*/
#ifdef SIGPIPE
	(void) signal(SIGPIPE, SIG_IGN);
#endif /*SIGPIPE*/
		/*do any authentication checks*/
#if !LOCAL_ONLY
	if (remote_auth(0) < 0){
			/*syslog already done...*/
                fprintf(stdout, "%d %s NNTP server doesn't trust you\r\n",
                        ERR_ACCESS, servername);
                (void) fflush(stdout);
                exit(1);
	} else
#endif /*!LOCAL_ONLY*/
	{
	char *connecttimestr;
	char *tmpstr;
		gettimeofday(&connecttime, (struct timezone *)0);
		connecttimestr = ctime((time_t *)&(connecttime.tv_sec));
		tmpstr = index(connecttimestr, '\n');
		if(!tmpstr)
			tmpstr = "Unknown date";
		else
			*tmpstr = '\0';
		/*note: that we lie, and maybe it will be true later...*/
		/*(ie maybe they can't post, and maybe we really can do auth*/
		fprintf(stdout, "%d %s NNTP[auth] server version %s ready \
at %s (canpost)\r\n", OK_CANPOST, servername, DNNTP_VERSION_STR, connecttimestr);
		(void) fflush(stdout);
	}
		/*do command stuff*/
		/* quiting will be done by code when we see a quit */
	while (1) {
	char **args;
	int no_args;
	struct cmdent_t *current_command;
#if DNNTP_IDLE_CLOSE
	/*   This DNNTP_IDLE_CLOSE may look complicated, but it impliments a simple policy.
	       If client connection is idle for DNNTP_IDLE_CLIENT seconds then
                  Close server connections which haven't received commands in
		  DNNTP_IDLE_CLOSE seconds;
	*/
	struct timeval TimeOut;
	time_t last,now;
	int i,open_servers;
	fd_set readready,exceptions;
	long readret;
	int  infd = fileno(stdin);
        	FD_ZERO(&readready);
		FD_SET(infd,&readready);
		FD_ZERO(&exceptions);
		FD_SET(infd,&exceptions);
	        if ( DNNTP_IDLE_CLIENT ) {  /* loose enforcement of DNNTP_IDLE_CLOSE */
			TimeOut.tv_sec = DNNTP_IDLE_CLIENT;
			TimeOut.tv_usec = 0;
		} else {  /* strict enforcement of DNNTP_IDLE_CLOSE */
			/* calculate next check */
			now = time(NULL);
			last = now;
			open_servers = 0;
        		for (i = 0; i < serv_array_len; i++) {
				if (serv_array[i].connection >= 0 ) {
                		  if ( serv_array[i].lastsent < last )
					last =  serv_array[i].lastsent;
				  open_servers++;
				}
			}
			if ( open_servers ) {
			  if (last <= now - DNNTP_IDLE_CLOSE) { 	
#if DEBUG_CMD_PRINT
#if SYSLOG
                            if (debug_level >= DEBUG_CMD_PRINT)
                              syslog(LOG_DEBUG, "Timeout, closing idle servers\n");
#endif /*SYSLOG*/
#endif /*DEBUG_CMD_REP_PRINT*/
                            CloseIdleServers();
                            continue;   /* while(1) */
			  } else {
			    TimeOut.tv_sec = last + DNNTP_IDLE_CLOSE - now;
			    TimeOut.tv_usec = 0;
			  }
			} else {
			  TimeOut.tv_sec = 0;
			  TimeOut.tv_usec = 0;
			}
                }

		readret = select(infd+1,&readready,NULL,&exceptions,&TimeOut);
		if ( readret < 1 ) {
#if DEBUG_CMD_PRINT
#if SYSLOG
                        if (debug_level >= DEBUG_CMD_PRINT)
                                syslog(LOG_DEBUG, "Timeout, closing idle servers\n");
#endif /*SYSLOG*/
#endif /*DEBUG_CMD_REP_PRINT*/
			CloseIdleServers();
	  		continue;   /* while(1) */
		}
#endif /*DNNTP_IDLE_CLOSE*/

		if (!fgets(command_line, sizeof(command_line), stdin)) {
#if DEBUG_CMD_REP_PRINT		/*lousey way to enable, but...*/
#if SYSLOG
			if (debug_level >= DEBUG_CMD_REP_PRINT)
				syslog(LOG_DEBUG, "connection closed?: %m\n");
#endif /*SYSLOG*/
#endif /*DEBUG_CMD_REP_PRINT*/
			QuitAllServers();
			exit(0);
		}
			/*print commands on demand*/
#if DEBUG_CMD_PRINT
			/*really ought to go somewhere better than syslog*/
#if SYSLOG
		if (debug_level >= DEBUG_CMD_PRINT)
			syslog(LOG_DEBUG, "recved: \'%s\'", command_line);
#endif /*SYSLOG*/
#endif /*DEBUG_CMD_PRINT*/
			
		args = break_args(command_line, &no_args);
		if (args && no_args) {
				/*identify command*/
			STRTOUPPER(args[0]);
			current_command = ansi_bsearch(args[0], cmdtable,
				NUM_COMMANDS, sizeof(struct cmdent_t), cmd_cmp);
			if (current_command)
				(current_command->cmd_fn)(args, no_args
					/*,cmdtable[i].cmd_no*/);
			else {
				if (strcasecmp(args[0], QUIT_STR) == 0) {
					break;
				} else {
#if DEBUG_CMD_PRINT
#if SYSLOG
				/* don't check level, cause you want to know!*/
				/*if (debug_level >= DEBUG_CMD_PRINT)*/
					syslog(LOG_ERR, "unrecognized command \
%s", args[0]);
#endif	/*SYSLOG*/
#endif /*DEBUG_CMD_PRINT*/
					printf("%d Command unrecognized.\r\n",
						ERR_COMMAND);
					(void) fflush(stdout);
				}
			}
			free_arg(args);
		}
	}
	printf("%d closing connection\r\n", OK_GOODBYE);
	(void)fflush(stdout);
	QuitAllServers();
}

/* cmd_cmp -- used to test if a command string matches
	a given entry in the command table (used through bsearch)
    parameters
	cmd: current command string
	ent: command entity(in command table) to check against
    return
	<0 if ent is after cmd in table, 0 if ent is the same, and >0
	if ent is before cmd in the table(ie just like strcmp)
*/
static int cmd_cmp( cmd, ent )
void *cmd;
void *ent;
{
return (strcmp((char *)cmd, ((struct cmdent_t *)ent)->cmd_name));
}
