/* servers.c */
/* code to manage set of servers */
/* Thu Feb  4 12:57:41 EST 1993 wbardwel : created*/
/*
 * HISTORY
 * $Log:	servers.c,v $
 * Revision 2.26  95/02/21  16:41:21  wbardwel
 * 	IDLE_CLOSE stuff from noyd@u.washington.edu
 * 	[95/02/21  16:36:53  wbardwel]
 * 
 * Revision 2.25  94/07/12  17:27:35  wbardwel
 * 	update copyright, fix for svr4
 * 	[94/07/12  17:24:57  wbardwel]
 * 
 * Revision 2.24  94/06/14  18:02:19  wbardwel
 * 	remove redundant inclusion of in.h
 * 	[94/06/14  17:58:31  wbardwel]
 * 
 * Revision 2.23  94/06/02  23:27:59  wbardwel
 * 	actually close server so we don't get more errors later
 * 	when get a failure string from opening a server
 * 	[94/06/02  23:27:35  wbardwel]
 * 
 * Revision 2.22  94/05/26  04:35:28  wbardwel
 * 	fix mis-use of time
 * 	[94/05/26  04:34:45  wbardwel]
 * 
 * Revision 2.21  94/04/21  01:49:59  wbardwel
 * 	make syslog debug messages clearer, rearrange some open code
 * 	[94/04/21  01:48:03  wbardwel]
 * 
 * Revision 2.20  93/12/02  03:29:14  wbardwel
 * 	removed unused variable
 * 	[93/12/02  03:26:52  wbardwel]
 * 
 * Revision 2.19  93/12/02  01:08:52  wbardwel
 * 	fix so we use length of curent group in server_t
 * 	[93/12/02  01:08:08  wbardwel]
 * 
 * Revision 2.18  93/12/02  00:46:17  wbardwel
 * 	don't keep looking servers up.  remember when server went down, and don't
 * 	retry until a while afterwards.  detect send failures better.
 * 	[93/12/02  00:37:40  wbardwel]
 * 
 * Revision 2.17  93/11/02  22:35:47  wbardwel
 * 	allow for not having vsnprintf
 * 	[93/11/02  22:04:19  wbardwel]
 * 
 * Revision 2.16  93/10/27  16:03:16  wbardwel
 * 	misc cleanup.  Add another stage to opening, so it now sends initial commands
 * 	before it reads anything on the connection.
 * 	[93/10/27  15:36:27  wbardwel]
 * 
 * Revision 2.15  93/08/18  19:41:43  wbardwel
 * 	fixes for when a server is down
 * 	[93/08/18  19:37:30  wbardwel]
 * 
 * Revision 2.14  93/08/17  17:38:52  wbardwel
 * 	modest speed fix
 * 	[93/08/17  17:34:21  wbardwel]
 * 
 * Revision 2.13  93/08/16  14:16:28  wbardwel
 * 	change ContReadFromServer to ContReadFromServerBufNum 
 * 	(see dnntp.h for more details)
 * 	[93/08/16  14:15:36  wbardwel]
 * 
 * Revision 2.12  93/08/13  21:18:49  wbardwel
 * 	added return code to SuppressXREALLY
 * 	[93/08/13  21:16:41  wbardwel]
 * 
 * Revision 2.11  93/08/13  16:37:07  wbardwel
 * 	bug fixes (inperticular to when a server goes down)
 * 	[93/08/13  16:30:44  wbardwel]
 * 
 * Revision 2.10  93/08/10  19:19:51  wbardwel
 * 	fix bugs in: OpenStage1, SendGroup (timeout problem), and cleanup syslog trouble
 * 	[93/08/10  19:17:26  wbardwel]
 * 
 * Revision 2.9  93/08/04  16:39:14  wbardwel
 * 	more CMU_ANDREW_KLUDGE (speed ups)
 * 	[93/08/04  16:28:46  wbardwel]
 * 
 * 	performance fixes and kludges
 * 	[93/08/04  00:15:30  wbardwel]
 * 
 * Revision 2.8  93/07/30  15:53:20  wbardwel
 * 	added some missing debuging print outs
 * 	[93/07/30  15:52:20  wbardwel]
 * 
 * Revision 2.7  93/07/28  13:25:50  wbardwel
 * 	broke up OpenServer into multiple stages
 * 	[93/07/28  13:21:50  wbardwel]
 * 
 * Revision 2.6  93/07/21  22:41:50  wbardwel
 * 	bug fix to make things work right under OSF
 * 	added ClearServer!
 * 	[93/07/21  22:37:43  wbardwel]
 * 
 * Revision 2.5  93/07/16  16:21:46  wbardwel
 * 	bug fixes and code to wedge MODE resends into RealOpenServer
 * 	[93/07/16  16:13:21  wbardwel]
 * 
 * Revision 2.4  93/07/01  19:49:16  wbardwel
 * 	fix bug reguarding forcable expiration in RealOpenServer
 * 	[93/07/01  19:15:47  wbardwel]
 * 
 * Revision 2.3  93/06/23  00:50:59  wbardwel
 * 	added changes for disk list cache (ie change Send* so that they won't
 * 	crash if server not open)
 * 	[93/06/23  00:27:48  wbardwel]
 * 
 * 	made RealOpenServer such that it won't exit unless it is really
 * 	hosed (so we can take other measures if nessicary)
 * 	[93/06/15  17:36:34  wbardwel]
 * 
 * Revision 2.2  93/06/02  14:52:57  wbardwel
 * 	Added use of XREALLY command on server opens.
 * 	[93/05/24            wbardwel]
 * 
 * 	Created.
 * 	[93/02/04            wbardwel]
 *
*/
/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software_Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */


#include "dnntpconf.h"
#include "dnntp.h"

#include <errno.h>
/*#include <netdb.h> done in dnntp.h*/
#include <arpa/inet.h>
#include <varargs.h>

dservers_t *serv_array = 0;
int serv_array_len = 0;

/* LoadServers -- load structure with servers (from config file)
    parameters
	filename: name of configuration file to read from
		(files of the format
		"netnews1.cs.cmu.edu 119
		anotherserver.here.org <correct portnumber>
		...
		" with # meaning comments start there)
    return
	number of servers on success, -1 on failure
*/
int LoadServers(filename)
char *filename;	/*name of file to load from*/
{
FILE *fp;
	/*holds line with hostname and port (or comment)*/
/*make sure we get enough space(cause we are reading lines into static
buffer, sortof each component must fit in static buffer) (tough luck...)*/
char line[MAXHOSTNAMELEN+1];
char host[MAXHOSTNAMELEN+1];
int port;
	
	if ((fp = fopen(filename, "r")) == NULL) {
#if SYSLOG
		syslog(LOG_ERR, "LoadServers: unable to open servers file: %m");
#endif /*SYSLOG*/
		return(DNNTP_NO_SETUP_ERR);
	}
		/*skip initial spaces on a line(of multiple lines)*/
	clear_spaces(fp);
	while (fgets(line, sizeof(line), fp)) {
		if (line[0] =='#'){
				/*go until last fgets on a line*/
			while(line[strlen(line)-1] != '\n'){
				if(!fgets(line, sizeof(line), fp))
					break;
			}
			continue;
		}
		if (sscanf(line, "%s", host) < 1){
#if SYSLOG
			syslog(LOG_ERR, "LoadServers: unable to parse servers \
file: %m");
#endif /*SYSLOG*/
			return(DNNTP_NO_SETUP_ERR);
		}
			/*if we didn't get whole line*/
		if (line[strlen(line)-1] != '\n'){
		int i;  /*misc index*/
				/*move stuff we haven't looked at to
				begining of buffer*/
			for(i=strlen(host); isspace(line[i]); i++)
				;
			if(line[i])
				strcpy(line, line+i);
			else
				line[0] = '\0';
			clear_spaces(fp);
			i = strlen(line);
			if(!fgets(line+i, sizeof(line)-i, fp))
				break;
			if (sscanf(line, "%d", &port) < 1){
#if SYSLOG
				syslog(LOG_ERR, "LoadServers: unable to parse \
servers file: %m");
#endif /*SYSLOG*/
				return(DNNTP_NO_SETUP_ERR);
			}
		} else if (sscanf(line+strlen(host), "%d", &port) < 1){
#if SYSLOG
			syslog(LOG_ERR, "LoadServers: unable to parse \
servers file: %m");
#endif /*SYSLOG*/
			return(DNNTP_NO_SETUP_ERR);
		}

		/*now we have a host and port, lets do something with it*/
			/*expand/make serv_array big enough...*/
		if (serv_array_len)
			serv_array = (dservers_t *)realloc(serv_array,
				(++serv_array_len)*sizeof(dservers_t));
		else
			serv_array = (dservers_t *)malloc(
				(++serv_array_len)*sizeof(dservers_t));
		if(!serv_array){
#if SYSLOG
			syslog(LOG_ERR, "LoadServer: ran out of memory");
#endif /*SYSLOG*/
			return(DNNTP_INTERNAL_ERR);
		}
			/*clear and copy stuff to dservers_t structure*/
		strcpy(serv_array[serv_array_len-1].hostname, host);
		if (port <0) {
#if SYSLOG
			syslog(LOG_ERR, "LoadServer: illegal port number");
#endif /*SYSLOG*/
			return(DNNTP_NO_SETUP_ERR);
		}else
			serv_array[serv_array_len-1].port = htons(port);
		serv_array[serv_array_len-1].connection = -1;
		serv_array[serv_array_len-1].curgroup[0] = 
			serv_array[serv_array_len-1].lastcommand[0] = '\0';
		serv_array[serv_array_len-1].connectionfpr = NULL;
		serv_array[serv_array_len-1].connectionfpw = NULL;
		serv_array[serv_array_len-1].sentgroup = 0;
		serv_array[serv_array_len-1].connection_addr.sin_addr.s_addr = 0;
#if DNNTP_DISK_LIST_CACHE
		serv_array[serv_array_len-1].server_down_timeout = 0;
#endif /*DNNTP_DISK_LIST_CACHE*/
#if DNNTP_IDLE_CLOSE
		serv_array[serv_array_len-1].lastsent = 0;
#endif /*DNNTP_IDLE_CLOSE*/
#if DO_XREALLY
		serv_array[serv_array_len-1].do_xreally = 1;
#endif /*DO_XREALLY*/
#if MODE
		serv_array[serv_array_len-1].do_mode = DNNTP_INIT_MODE_VAL;
#endif /*MODE*/
		bzero((char *) &serv_array[serv_array_len-1].connection_addr,
			sizeof(serv_array[serv_array_len-1].connection_addr));

		clear_spaces(fp);
	}
	return(serv_array_len);
}

int (*open_serv_func[DNNTP_OPEN_STAGES])() = { OpenStage1, OpenStage2, OpenStage3, OpenStage4 };
/* OpenServer -- open connection to a server (if needed)*/
/*see dnntp.h*/
/* RealOpenServer -- really open a connection (ie it is needed)
    parameters
	servernumber: number of server to open to (ie offset into servers
		structure.
    return
	>=0 on success, <0 on failure
*/
int RealOpenServer(serverno)
int serverno;
{
int i, status;
	for (i = 0 ; i < DNNTP_OPEN_STAGES; i++)
		if ((status = open_serv_func[i](serverno)) != DNNTP_SUCCESS)
			return (status);
	return (DNNTP_SUCCESS);
}

/* stage 1 opens socket, connects and sets up stdio versions of things*/
int OpenStage1(serverno)
int serverno;
{
	if (ValidServer(serverno)) {
		/*look up hostname*/
		if (*serv_array[serverno].hostname) {
		   if (serv_array[serverno].connection_addr.sin_addr.s_addr == 0) {
			if(isdigit(*serv_array[serverno].hostname)) {
				serv_array[serverno].connection_addr.sin_addr.
					s_addr = inet_addr(serv_array[serverno].hostname);
			} else {
				struct hostent *he;
				if (((he = gethostbyname(serv_array[serverno].
					hostname)) == NULL) ||
					(he->h_addrtype != AF_INET) ||
					(he->h_length != 4)) {
#if SYSLOG
					syslog(LOG_ERR, "OpenStage1: \
unable to look up server: %s", serv_array[serverno].hostname);
#endif /*SYSLOG*/
					serv_array[serverno].connection = -1;
					serv_array[serverno].misc_state = 0;
					return(DNNTP_REMOTE_GONE_ERR);
				}

				bcopy (he->h_addr,
					&(serv_array[serverno].connection_addr.
					sin_addr), sizeof(struct in_addr));
				/*note: we really ought to try all the
				addresses instead of the first one from the
				list*/
			}
			serv_array[serverno].connection_addr.sin_family =
				AF_INET;
			serv_array[serverno].connection_addr.sin_port =
				serv_array[serverno].port;
		    }	/* need to do lookup*/
#if DNNTP_DISK_LIST_CACHE
			if (serv_array[serverno].server_down_timeout &&
				((time(NULL) - serv_array[serverno].
				server_down_timeout) <
				DNNTP_SERVER_DOWN_TIMEOUT)) {
				serv_array[serverno].connection = -1;
				serv_array[serverno].misc_state = 0;
				return (DNNTP_REMOTE_GONE_ERR);
			}
#endif /*DNNTP_DISK_LIST_CACHE*/
			if ((serv_array[serverno].connection =
				socket(AF_INET, SOCK_STREAM, 0)) < 0) {
#if SYSLOG
				syslog(LOG_ERR, "OpenStage1: socket() \
failed: %m");
#endif /*SYSLOG*/
				serv_array[serverno].misc_state = 0;
				return(DNNTP_REMOTE_GONE_ERR);
			}
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
				syslog(LOG_DEBUG, "open connecting to %d",
					serverno);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
#if DNNTP_DISK_LIST_CACHE
			serv_array[serverno].server_down_timeout = 0;
#endif /*DNNTP_DISK_LIST_CACHE*/
			while (connect(serv_array[serverno].connection,
				(struct sockaddr *) (&serv_array[serverno].
				connection_addr), sizeof(serv_array[serverno].
				connection_addr)) < 0) {
				if (errno == EINTR) {
					errno = 0;
					continue;
				}
#if SYSLOG
				syslog(LOG_ERR, "OpenStage1: connect() \
to %s:%d failed: %m", ServerName(serverno), ntohs(serv_array[serverno].port));
#endif /*SYSLOG*/
#if 0
				printf("%d couldn't connect to real server %s.\
Goodbye\r\n", ERR_GOODBYE, serv_array[serverno].hostname);
				goto quit_err;
#else	/*!0*/
				close(serv_array[serverno].connection);
				serv_array[serverno].connection = -1;
				serv_array[serverno].misc_state = 0;
#if DNNTP_DISK_LIST_CACHE
				serv_array[serverno].server_down_timeout =
					time(NULL);
#endif /*DNNTP_DISK_LIST_CACHE*/

				return (DNNTP_REMOTE_GONE_ERR);
#endif	/*!0*/
			}
			serv_array[serverno].sentgroup = 0;
	 	} else {
#if SYSLOG
			syslog(LOG_ERR, "internal error: OpenStage1 bogus \
server name");
#endif /*SYSLOG*/
			serv_array[serverno].connection = -1;
			serv_array[serverno].misc_state = 0;
			return(DNNTP_INTERNAL_ERR);
		}
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: OpenStage1 bogus server \
number");
#endif /*SYSLOG*/
#endif /*LOG*/
		serv_array[serverno].connection = -1;
		serv_array[serverno].misc_state = 0;
		return(DNNTP_INTERNAL_ERR);
	}
	return (DNNTP_SUCCESS);
quit_err:
	(void) fflush(stdout);
	QuitAllServers();
	serv_array[serverno].connection = -1;
	serv_array[serverno].misc_state = 0;
	return(DNNTP_SIMPLE_ERR);
}

/*stage 2 sends MODE command (if applicable), send XREALLY command
(if applicable)*/
int OpenStage2(serverno)
int serverno;
{
/*char str[NNTP_STRLEN+1];*/
	if ((serv_array[serverno].connectionfpw =
		fdopen( serv_array[serverno].connection, "w")) == NULL){
#if SYSLOG
		syslog(LOG_ERR, "OpenStage2: fdopen(%d, \"w\") failed: %m", serverno);
#endif /*SYSLOG*/
#if 0
		printf("%d couldn't open write channel to real \
server %s.  Goodbye\r\n", ERR_GOODBYE, serv_array[serverno].hostname);
		goto quit_err;
#else	/*!0*/
		CloseServer (serverno);
#if DNNTP_DISK_LIST_CACHE
		serv_array[serverno].server_down_timeout = time(NULL);
#endif /*DNNTP_DISK_LIST_CACHE*/
		return (DNNTP_REMOTE_GONE_ERR);
#endif	/*!0*/
	}
	if ((serv_array[serverno].connectionfpr =
		fdopen( serv_array[serverno].connection,
		"r")) == NULL){
#if SYSLOG
		syslog(LOG_ERR, "OpenStage2: fdopen(\"r\") \
failed: %m");
#endif /*SYSLOG*/
#if 0
		printf("%d couldn't open read channel to real \
dserver %s.  Goodbye\r\n", ERR_GOODBYE, serv_array[serverno].hostname);
		goto quit_err;
#else	/*!0*/
		CloseServer (serverno);
#if DNNTP_DISK_LIST_CACHE
		serv_array[serverno].server_down_timeout = time(NULL);
#endif /*DNNTP_DISK_LIST_CACHE*/
		return (DNNTP_REMOTE_GONE_ERR);
#endif	/*!0*/
	}
#if MODE
	if (serv_array[serverno].do_mode == 1){
		fputs("MODE READER\r\n", serv_array[serverno].connectionfpw);
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
			syslog(LOG_DEBUG, "open sending: 'MODE READER' to %d",
				serverno);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
		(void) fflush(serv_array[serverno].connectionfpw);
	}
#endif /*MODE*/
#if DO_XREALLY
	if (serv_array[serverno].do_xreally){
#if DNNTP_CMU_ANDREW_KLUDGE
		/*send xreally*/
		extern char servername[];
		if (strcasecmp(clientname, servername) == 0) {
			fprintf(serv_array[serverno].connectionfpw,
				"XREALLY %s\r\n", clientname);
		} else
#endif /*DNNTP_CMU_ANDREW_KLUDGE*/
		fprintf(serv_array[serverno].connectionfpw,
			"XREALLY %s\r\n", clientaddrstr);
		(void) fflush(serv_array[serverno].connectionfpw);
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
#if DNNTP_CMU_ANDREW_KLUDGE
			if (strcasecmp(clientname, servername) == 0)
			syslog(LOG_DEBUG, "open sending: 'XREALLY %s' \
to %d", clientname, serverno);
			else
#endif /*DNNTP_CMU_ANDREW_KLUDGE*/
			syslog(LOG_DEBUG, "open sending: 'XREALLY %s' \
to %d", clientaddrstr, serverno);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
	}
#endif /*DO_XREALLY*/
#if DNNTP_IDLE_CLOSE
	/*Set stamp now, regardless of whether we've done MODE or XREALLY*/
	serv_array[serverno].lastsent = time(NULL);  
#endif /*DNNTP_IDLE_CLOSE*/

	return (DNNTP_SUCCESS);
quit_err:
	(void) fflush(stdout);
	QuitAllServers();
	return(DNNTP_SIMPLE_ERR);
}

/*stage 3 reads initial string*/
int OpenStage3(serverno)
int serverno;
{
char str[NNTP_STRLEN+1];
	if ((!fgets(str, sizeof(str), serv_array[serverno].connectionfpr)) ||
		(strncmp(str, "200", 3) && strncmp(str, "201", 3))) {
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_REP_PRINT)
			syslog(LOG_DEBUG, "open read (error)from %d: '%s'",
				serverno, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
#if SYSLOG
		syslog(LOG_ERR, "OpenStage3: didn't get valid initial \
string");
#endif /*SYSLOG*/
		CloseServer (serverno);
#if DNNTP_DISK_LIST_CACHE
		serv_array[serverno].server_down_timeout = 0;
#endif /*DNNTP_DISK_LIST_CACHE*/
		return (DNNTP_REMOTE_GONE_ERR);
	}
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_REP_PRINT)
			syslog(LOG_DEBUG, "open read from %d: '%s'",
				serverno, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
		/*kill the cache...(performance hit...) but we must
		keep it intact since we may be in the middle of
		something that needs it*/ /*could be done in Close,
		but this is better so it won't get flushed just cause
		one server died...(ie for DISK_LIST_CACHE stuff*/
		/* only do this if there is something to kill, otherwise
		we may erase time mark from Group2ServerExpire)*/
	if (group_array)
		group_array_secs = 0;
	return(DNNTP_SUCCESS);
}

/*stage 4 reads the response from MODE and XREALLY (if applicable) and
trashes cache*/
int OpenStage4(serverno)
int serverno;
{
char str[NNTP_STRLEN+1];
#if MODE
	if (serv_array[serverno].do_mode == 1){
		if (!fgets(str, sizeof(str),
			serv_array[serverno].connectionfpr)||
			(!strncmp(str, "500", 3))) {
				/*hum, shouldn't happen...*/
			serv_array[serverno].do_mode = 0;
		}
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_REP_PRINT)
			syslog(LOG_DEBUG, "open read MODE from %d: '%s'",
				serverno, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
	}
#endif /*MODE*/
#if DO_XREALLY
	if (serv_array[serverno].do_xreally) {
		/*check reply*/
		str[0] = '\0';
		if ((!fgets(str, sizeof(str),
			serv_array[serverno].connectionfpr)) ||
			(strncmp(str, "200", 3) &&
			strncmp(str, "201", 3))) {
			if (!strncmp(str, "500", 3)) {
				/*mark it if they just didn't understand*/
				serv_array[serverno].do_xreally = 0;
			} else {
#if SYSLOG
				syslog(LOG_ERR, "unauthorized connection from \
%s, real server %s didn't approve", clientaddrstr,
				serv_array[serverno].hostname);
#endif /*SYSLOG*/
				printf("%d real server %s doesn't trust you.  \
Goodbye\r\n", ERR_GOODBYE, serv_array[serverno].hostname);
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
				if (debug_level >= DEBUG_REMOTE_REP_PRINT)
					syslog(LOG_DEBUG, "open read from %d: \
'%s'", serverno, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
				goto quit_err;
			}
		}
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_REP_PRINT)
			syslog(LOG_DEBUG, "open XREALLY read from %d: '%s'",
				serverno, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
		}
#endif /*DO_XREALLY*/
	return(DNNTP_SUCCESS);
quit_err:
	(void) fflush(stdout);
	QuitAllServers();
	exit(1);
}

/* CloseServer -- close connection to server (and mark as closed)
    parameters
	servernumber: number of server to close (ie offset into servers
		structure.
    return
	none
*/
void CloseServer(servernumber)
int servernumber;
{
	if (ValidServer(servernumber)) {
		if (serv_array[servernumber].connectionfpr)
			fclose(serv_array[servernumber].connectionfpr);
		else if (!serv_array[servernumber].connectionfpw)
				/* if fclose isn't gonna catch it */
			if (serv_array[servernumber].connection >= 0)
				close(serv_array[servernumber].connection);
		if (serv_array[servernumber].connectionfpw)
			fclose(serv_array[servernumber].connectionfpw);
		serv_array[servernumber].connectionfpr = NULL;
		serv_array[servernumber].connectionfpw = NULL;
		serv_array[servernumber].misc_state = 0;
		serv_array[servernumber].connection = -1;
		serv_array[servernumber].sentgroup = 0;
#if DNNTP_IDLE_CLOSE
		serv_array[servernumber].lastsent = 0;
#endif /*DNNTP_IDLE_CLOSE*/
	}
#if LOG
#if SYSLOG
	else
		syslog(LOG_ERR, "internal error: closed bogus server number");
#endif /*SYSLOG*/
#endif /*LOG*/
}

/* SendGroup -- send any commands to make sure server is in proper group
    parameters
	servernumber: number of server send group commands to (actual
		group to enter is in servers structure.)
    return
	>=0 if succeeds <0 if fails (if fails from server not liking command
	returns -error_code;
*/
int SendGroup(servernumber, str)
int servernumber;
char *str;
{
int status;
	if (ValidServer(servernumber)) {
		if (!serv_array[servernumber].sentgroup) {
		if (serv_array[servernumber].connection < 0)
			return (DNNTP_REMOTE_GONE_ERR);
		if (serv_array[servernumber].curgroup[0]) {
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
				syslog(LOG_DEBUG, "sending: 'GROUP %s' to %d",
					serv_array[servernumber].curgroup,
					servernumber);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
			if ((fprintf(serv_array[servernumber].connectionfpw,
				"GROUP %s\r\n",
				serv_array[servernumber].curgroup) == EOF) ||
				(fflush(serv_array[servernumber].connectionfpw)
				== EOF)) {
resend_group:
				CloseServer(servernumber);
				OpenServer(servernumber);
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
				if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
					syslog(LOG_DEBUG, "(error)resending: \
'GROUP %s' to %d", serv_array[servernumber].curgroup, servernumber);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
				if ((serv_array[servernumber].connection<0)||
					(fprintf(serv_array[servernumber].
					connectionfpw, "GROUP %s\r\n",
					serv_array[servernumber].curgroup)
					== EOF) || (fflush(serv_array[
					servernumber].connectionfpw) == EOF)) {
#if LOG
#if SYSLOG
					syslog(LOG_ERR, "odd error: SendGroup \
unable to send to server %d", servernumber);
#endif /*SYSLOG*/
#endif /*LOG*/
#if DNNTP_DISK_LIST_CACHE
					serv_array[servernumber].
						server_down_timeout =
						time(NULL);
#endif /*DNNTP_DISK_LIST_CACHE*/
					return(DNNTP_REMOTE_DIS_ERR);
				}
			}
			{
			char tmp_str[NNTP_STRLEN+1];
				if (str == NULL)
					str = tmp_str;
				if ((!fgets(str, NNTP_STRLEN+1,
					serv_array[servernumber].
					connectionfpr))||
					(!strncasecmp(str, DNNTP_TIMEOUT_STR,
					DNNTP_TIMEOUT_STRLEN))) {
					str[0] = '\0';
					goto resend_group;
				}

#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
				if (debug_level >= DEBUG_REMOTE_REP_PRINT)
					syslog(LOG_DEBUG, "read from %d: '%s'",
						servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
					/* this is a cheat, but should work*/
				status = atoi(str);
				if (status != OK_GROUP)
					return(-status);
			}
		} else	/*!serv_array[servernumber.curgroup[0]*/
			return(DNNTP_INTERNAL_ERR);
		}	/*!serv_array[servernumber].sentgroup*/
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: SendGroup to bogus server \
number");
#endif /*SYSLOG*/
#endif /*LOG*/
		return(DNNTP_INTERNAL_ERR);
	}
#if DNNTP_IDLE_CLOSE
        serv_array[servernumber].lastsent = time(NULL);
#endif /*DNNTP_IDLE_CLOSE*/
	serv_array[servernumber].sentgroup = 1;
	return(DNNTP_SUCCESS);
}

/* SetGroup -- send any commands to make server go to group (first time)
	and record which group (on for this server)
    parameters
	servernumber: number of server send group commands to.
	group: name of group
	str: dump return string from server here (for caller) see SendGroup
		for caveats
	reallysend: should we send the actual group command? (or just mark
		groupname for future reference(and assume it was sent!))

    return
	>=0 if succeeds <0 if fails
*/
int SetGroup(servernumber, group, str, reallysend)
int servernumber;
char *group;
char *str;
int reallysend;
{
	if (ValidServer(servernumber)) {
		strncpy(serv_array[servernumber].curgroup, group, NNTP_STRLEN);
		serv_array[servernumber].curgroup[NNTP_STRLEN-1] = '\0';
		if (reallysend) {
				/*haven't sent this one yet*/
			serv_array[servernumber].sentgroup = 0;
			return(SendGroup(servernumber, str));
		} else {
			serv_array[servernumber].sentgroup = 1;
		}
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: SetGroup to bogus server \
number");
#endif /*SYSLOG*/
#endif /*LOG*/
		return(DNNTP_INTERNAL_ERR);
	}
	/*NOTREACHED*/
}

/* SendToServer -- send command to a server (and save command to be
	resend later if needed (in ReadFromServer))
    parameters
	servernumber: number of server send commands to.
	formatstr: printf style format string
	args: stuff to filling printf item, in parens...
    return
	>=0 if succeeds <0 if fails
*/
int SendToServer(servernumber, formatstr, va_alist)
int servernumber;
char *formatstr;
va_dcl
{
va_list printfargs;
int code = DNNTP_SUCCESS;
	va_start(printfargs);
	if (ValidServer(servernumber)) {
#ifdef CMUCS
		if (vsnprintf(serv_array[servernumber].lastcommand,
			sizeof(serv_array[servernumber].lastcommand), formatstr,
			printfargs) == EOF) {
			code = DNNTP_SIMPLE_ERR;
			goto quit;
		}
#else /*!CMUCS*/	/*not as safe!!!*/
		vsprintf(serv_array[servernumber].lastcommand,
			formatstr, printfargs);
#endif /*!CMUCS*/
			/*make sure it succeded*/
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
			syslog(LOG_DEBUG, "sending: '%s' to %d",
				serv_array[servernumber].lastcommand,
				servernumber);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
		if (serv_array[servernumber].connection < 0)
			return(DNNTP_REMOTE_GONE_ERR);
		if ((fprintf(serv_array[servernumber].connectionfpw, "%s\r\n",
			serv_array[servernumber].lastcommand) == EOF) ||
	 		(fflush(serv_array[servernumber].connectionfpw) == EOF)) {
			CloseServer(servernumber);
			if ((code = OpenServer(servernumber)) < 0)
				return(code);
			if (serv_array[servernumber].curgroup[0])
				SendGroup(servernumber, NULL);
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
				syslog(LOG_DEBUG, "(error)sending: '%s' to %d",
					serv_array[servernumber].lastcommand,
					servernumber);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
			if ((fprintf(serv_array[servernumber].connectionfpw,
				"%s\r\n",
				serv_array[servernumber].lastcommand) == EOF) ||
	 		(fflush(serv_array[servernumber].connectionfpw) == EOF)) {
#if LOG
#if SYSLOG
				syslog(LOG_ERR, "odd error: SendToServer \
unable to send to server %d", servernumber);
#endif /*SYSLOG*/
#endif /*LOG*/
				code = DNNTP_REMOTE_GONE_ERR;
#if DNNTP_DISK_LIST_CACHE
				serv_array[servernumber].
					server_down_timeout = time(0);
#endif /*DNNTP_DISK_LIST_CACHE*/
				goto quit;
			}
		}
#if DNNTP_IDLE_CLOSE
		serv_array[servernumber].lastsent = time(NULL);
#endif /*DNNTP_IDLE_CLOSE*/
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: SendToServer to bogus server \
number");
#endif /*SYSLOG*/
#endif /*LOG*/
		code = DNNTP_INTERNAL_ERR;
	}
quit:
	(void) fflush(serv_array[servernumber].connectionfpw);
	va_end(printfargs);
	return(code);
}

/* ReSendToServer -- resend last command to server
    parameters
	servernumber: number of server send commands to.
    return
	>=0 if succeeds <0 if fails
*/
int ReSendToServer(servernumber)
int servernumber;
{
	if (ValidServer(servernumber)) {
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
			syslog(LOG_DEBUG, "resending: '%s' to %d",
				serv_array[servernumber].lastcommand,
				servernumber);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
		if ((fprintf(serv_array[servernumber].connectionfpw, "%s\r\n",
			serv_array[servernumber].lastcommand) == EOF) ||
			(fflush(serv_array[servernumber].connectionfpw) == EOF)){
			CloseServer(servernumber);
			OpenServer(servernumber);
			SendGroup(servernumber, NULL);
#if DEBUG_REMOTE_CMD_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_CMD_PRINT)
				syslog(LOG_DEBUG, "(error)resending: '%s' to %d"
					,serv_array[servernumber].lastcommand,
					servernumber);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_CMD_PRINT*/
			if ((fprintf(serv_array[servernumber].connectionfpw,
				"%s\r\n",
				serv_array[servernumber].lastcommand) == EOF) ||
			(fflush(serv_array[servernumber].connectionfpw) == EOF)){
#if LOG
#if SYSLOG
				syslog(LOG_ERR, "odd error: ReSendToServer \
unable to send to server %d", servernumber);
#endif /*SYSLOG*/
#endif /*LOG*/
				CloseServer(servernumber);
#if DNNTP_DISK_LIST_CACHE
				serv_array[servernumber].
					server_down_timeout = time(0);
#endif /*DNNTP_DISK_LIST_CACHE*/
				return(DNNTP_REMOTE_DIS_ERR);
			}
		}
#if DNNTP_IDLE_CLOSE
		serv_array[servernumber].lastsent = time(NULL);
#endif /*DNNTP_IDLE_CLOSE*/
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: ReSendToServer to bogus \
server number");
#endif /*SYSLOG*/
#endif /*LOG*/
		return (DNNTP_INTERNAL_ERR);
	}
	return(DNNTP_SUCCESS);
}

/* ReadFromServer -- read results from a command
	and resend command if needed
    parameters
	servernumber: number of server send commands to.
	str: string to put result in (this had better be NNTP_STRLEN+1)
		(note: str may be trashed even on failure)
    return
	>=0 if succeeds (and result code if succeeds)<0 if fails
*/
int ReadFromServer(servernumber, str)
int servernumber;
char *str;
{
char bogo_str[NNTP_STRLEN+1];
int status;
	if (!str)
		str = bogo_str;
	if (ValidServer(servernumber)) {
		if ((!fgets(str, NNTP_STRLEN+1,
			serv_array[servernumber].connectionfpr)) ||
			(!strncasecmp(str, DNNTP_TIMEOUT_STR,
			DNNTP_TIMEOUT_STRLEN))) {
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_REP_PRINT)
				syslog(LOG_DEBUG, "read(error) from %d: '%s'",
					servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
			CloseServer(servernumber);
			OpenServer(servernumber);
			if (serv_array[servernumber].curgroup[0])
				if ((status = SendGroup(servernumber, NULL))
					< 0)
					return(status);
			if ((status = ReSendToServer(servernumber)) < 0)
				return(status);
			if (!fgets(str, NNTP_STRLEN+1,
				serv_array[servernumber].connectionfpr)){
#if LOG
#if SYSLOG
				syslog(LOG_ERR, "odd error: ReadFromServer \
couldn't talk to server %d",
					servernumber);
#endif /*SYSLOG*/
#endif /*LOG*/
#if DNNTP_DISK_LIST_CACHE
				serv_array[servernumber].
					server_down_timeout = time(0);
#endif /*DNNTP_DISK_LIST_CACHE*/
				return(DNNTP_REMOTE_DIS_ERR);
			}
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_REP_PRINT)
				syslog(LOG_DEBUG, "read from %d: '%s'",
					servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
		}
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_REP_PRINT)
				syslog(LOG_DEBUG, "read from %d: '%s'",
					servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
			/*we have got a line or we have returned by here*/
				/*get number thing from beginning of line*/
		return (atoi(str));
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: ReadFromServer to bogus \
server number");
#endif /*SYSLOG*/
#endif /*LOG*/
		return(DNNTP_INTERNAL_ERR);
	}
	/*NOTREACHED*/
}

/* ContReadFromServerBufNum -- read data from command (like . terminated lists)
	
    parameters
	servernumber: number of server send commands to.
	read: string to put result in (this had better be NNTP_STRLEN+1)
    return
	>0 if succeeds and more data to come (0 if last item(ie . line)
		<0 if fails
*/
int ContReadFromServerBufNum(servernumber, str, strl)
int servernumber;
char *str;
size_t strl;
{
	if (ValidServer(servernumber)) {
	/*failed fgets here is fatal, since in the middle of a command*/
		if (!fgets(str, strl,
			serv_array[servernumber].connectionfpr)) {
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_REP_PRINT)
				syslog(LOG_DEBUG, "cread(error) from %d: '%s'",
					servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
#if LOG
#if SYSLOG
			syslog(LOG_ERR, "odd error: ContReadFromServer \
couldn't talk to server %d",
				servernumber);
#endif /*SYSLOG*/
#endif /*LOG*/
			return(DNNTP_REMOTE_DIS_ERR);
		}
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_REP_PRINT)
			syslog(LOG_DEBUG, "contread from %d: '%s'",
				servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
		return(!((*str == '.')&&(*(str+1) == '\r')&&
			(*(str+2) == '\n')));
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: ContReadFromServer to bogus \
server number");
#endif /*SYSLOG*/
#endif /*LOG*/
		return(DNNTP_INTERNAL_ERR);
	}
	/*NOTREACHED*/
}

#if XTHREAD
/* ContBinReadFromServer -- read data from command (like . terminated lists)
    parameters
	servernumber: number of server send commands to.
	read: string to put result in (this had better be NNTP_STRLEN+1)
    return
	>0 if succeeds and more data to come (0 if last item(ie . line)
		<0 if fails
*/
int ContBinReadFromServer(servernumber, no_bytes, str)
int servernumber;
int no_bytes;
char *str;
{
	if (ValidServer(servernumber)) {
	/*failed fgets here is fatal, since in the middle of a command*/
		
		if (!fread(str, 1, no_bytes,
			serv_array[servernumber].connectionfpr)) {
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
			if (debug_level >= DEBUG_REMOTE_REP_PRINT)
				syslog(LOG_DEBUG, "bread(error) from %d: '%s'",
					servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
#if LOG
#if SYSLOG
			syslog(LOG_ERR, "odd error: ContBinReadFromServer \
couldn't talk to server %d",
				servernumber);
#endif /*SYSLOG*/
#endif /*LOG*/
			return(DNNTP_REMOTE_DIS_ERR);
		}
#if DEBUG_REMOTE_REP_PRINT
#if SYSLOG
		if (debug_level >= DEBUG_REMOTE_REP_PRINT)
			syslog(LOG_DEBUG, "bread from %d: '%s'",
				servernumber, str);
#endif /*SYSLOG*/
#endif /*DEBUG_REMOTE_REP_PRINT*/
	} else {
#if LOG
#if SYSLOG
		syslog(LOG_ERR, "internal error: ContBinReadFromServer to \
bogus server number");
#endif /*SYSLOG*/
#endif /*LOG*/
		return(DNNTP_INTERNAL_ERR);
	}
	return(DNNTP_SUCCESS);
}
#endif /*XTHREAD*/

/* QuitAllServers -- send quit to all servers, and close them all
    parameters
	none
    return
	>=0 if succeeds (no such thing as failure...)
*/
int QuitAllServers()
{
int i;
	for (i = 0; i < serv_array_len; i++)
		if (serv_array[i].connectionfpw) {
			fputs(DNNTP_CLOSE_SERVER_CMD,
				serv_array[i].connectionfpw);
			fflush(serv_array[i].connectionfpw);
			CloseServer(i);
		}
	return (DNNTP_SUCCESS);
}


void clear_spaces(fp)
FILE *fp;
{
int ch;
	do{
		ch = fgetc(fp);
	}while(isspace(ch));
	if(ch != EOF)
		ungetc(ch, fp);
}

#if DO_XREALLY
	/* ServersSuppressXREALLY -- this turns off xreally use
	to all real servers (ie this is a local connection so this is
	unneeded (ie connection will come from same host as client...*/
int ServersSuppressXREALLY()
{
int i;
	for(i=0; i < serv_array_len; i++)
		serv_array[i].do_xreally = 0;
	return (DNNTP_SUCCESS);
}
#endif /*DO_XREALLY*/

/* ClearServer -- clear extra data from a serveR
    parameters
	servernum: server to clear
	code: if ReadFromServer gets this return code then expect a .\r\n line
    return
	>=0 if succeeds <0 if fails
*/
int ClearServer(servernum, code)
int servernum, code;
{
int status;
char str[NNTP_STRLEN+1];
	if ((status = ReadFromServer(servernum, str)) < 0)
		return (status);
	if (status == code)
		while ((status = ContReadFromServer(servernum, str))>0);
	if (status < 0)
		return (status);
	return (DNNTP_SUCCESS);
}
 
#if DNNTP_IDLE_CLOSE
/* CloseIdleServers --  Close connections to servers which have
        not receieved a command in DNNTP_IDLE_CLOSE seconds.
   parameters
	none
   returns
	number of connections closed
*/
int CloseIdleServers()
{
int i;
time_t idle;
int status;

	status = 0;
	idle = time(NULL) - DNNTP_IDLE_CLOSE;
        for (i = 0; i < serv_array_len; i++) {
		if ( serv_array[i].lastsent > idle ) continue;
                if (serv_array[i].connection >= 0 ) {
                        CloseServer(i);
			status++;
                }
	}
        return (status);
}
#endif /*DNNTP_IDLE_CLOSE*/
